/*********************************************************************************/
/* Copyright (c) 2001, Analog Devices Inc., All Rights Reserved			 */
/*										 */
/* Buffers.h									 */
/*										 */
/* Various types of "buffers" that will be used throughout the driver		 */
/*										 */
/* This file is part of the "ADI USB ADSL Driver for Linux".			 */
/* 										 */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it     */
/* and/or modify it under the terms of the GNU General Public License as         */
/* published by the Free Software Foundation; either version 2 of the License,   */
/* or (at your option) any later version.					 */
/*										 */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be    */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		 */
/* GNU General Public License for more details.					 */
/*										 */	
/* You should have received a copy of the GNU General Public License		 */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software*/
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     */
/*********************************************************************************/

#ifndef BUFFERS_H
#define BUFFERS_H

/*********************************************************************************/
/* Some buffer-related macros							 */
/*********************************************************************************/

#define GET_DSP_MEMORY(_phw_, _dst_, _type_, _memsize_) \
    if (_phw_->DspAllocCount<kMAX_DSP_ALLOCS) \
    { \
        _dst_ = _type_ vmalloc(_memsize_); \
        if (_dst_) \
        { \
            ZAP(DEBUG_MEMORY, DEBUG_INFO, (kTAB"alloced %x bytes at %p for dsp\n", _memsize_, _dst_)); \
            _phw_->DspAllocs[_phw_->DspAllocCount].AllocedSize = _memsize_; \
            _phw_->DspAllocs[_phw_->DspAllocCount++].pData     = (UInt8 *)_dst_; \
        } \
    }

#define FREE_DSP_MEMORY(_buf_) \
    { \
        if (_buf_.pData) \
        { \
            ZAP(DEBUG_MEMORY, DEBUG_INFO, (kTAB"freeing %p for dsp\n", _buf_.pData)); \
            vfree(_buf_.pData); \
            _buf_.pData = NULL; \
        } \
    }

#define GET_KBUFFER(_memsize_) kmalloc(_memsize_, in_interrupt() ? GFP_ATOMIC : GFP_KERNEL)
#define GET_VBUFFER(_memsize_) vmalloc(_memsize_)

#define FREE_KBUFFER(_ptr_) \
    if (_ptr_) \
    { \
        kfree(_ptr_); \
        _ptr_ = NULL; \
    }

#define FREE_VBUFFER(_ptr_) \
    if (_ptr_) \
    { \
        vfree(_ptr_); \
        _ptr_ = NULL; \
    }
    

/*********************************************************************************/
/* Structs that will be held in queues (manipulated with Enqueue and Dequeue)    */
/* In order to simplify and buffer management, each type of buffer will begin    */
/* with a generic buffer and follow with specific stuff. This will allow generic */
/* buffer manipulation routines to operate on any buffer. Note that if the       */
/* buffer needs associated with a Pb stored in the buffer (so you can get back   */ 
/* to the buffer pointer in a USB callback), the AdiUSBPB structure that the     */
/* pQueueEntry member will reside in needs to immediately follow the GB member.  */
/*********************************************************************************/

/* Compatibility with 2.4.19 and higher */
#include "linux/version.h"
#if LINUX_VERSION_CODE > 132115
typedef struct urb urb_t;
typedef struct
{
        __u8  requesttype;
        __u8  request;
        __u16 value;
        __u16 index;
        __u16 length;
} devrequest __attribute__ ((packed));
#endif

typedef struct
{
    UInt8        *pData;
    UInt32       AllocedSize;
} BUFFER;
                            
typedef struct
{
    QHdr         List;
    UInt8        *pData;
    UInt32       AllocedSize;
} GENERIC_BUFFER;

#define CTRL_URB_Q_SIZE 16

typedef struct
{
    GENERIC_BUFFER GB;
    void        *pHw;
    urb_t       *urb;
} QUEUED_URB_BUFFER;
                     
/*********************************************************************************/
/* This will hold network packet data that is waiting to be sent back to         */
/* the higher-level stack. Two queues of these will be maintained - a free       */
/* queue and a waiting queue							 */
/*********************************************************************************/
#define ETHERNET_PACKET_Q_SIZE    64
#define ETHERNET_PACKET_DATA_SIZE 1536
typedef struct
{
    GENERIC_BUFFER GB;
    UInt32         EncapSkipSize;
    UInt32         PacketSize;
    Boolean        DropThisPacket;
} ETHERNET_PACKET_BUFFER;

/************************************************************************************/
/* This will be the receiver buffer for incoming network data. We will queue        */
/* up a bunch of reads with USB with these buffers, then process them into          */
/* real ethernet packets as they come in. This enables us to support situations     */
/* where the entire ethernet packet does not come in in one USBBulkRead. We	    */
/* do require that the data received into these buffers is an exact number          */
/* of ATM cells									    */	
/* LINUX NOTE :									    */
/* In kernel 2.4.7-10 (the stock kernel with RedHat 7.2), the usb-uhci module	    */
/* has a problem queuing multiple read bulk urbs). Therefore, we set this to one so */
/* we'll work with uhci or ohci. However, higher performance can be seen on ohci    */
/* if you allow multiple read urbs. So, if you're on an ohci-based machine,	    */
/* set INCOMING_Q_SIZE to > 1 to achieve greater performance.			    */
/************************************************************************************/
#ifdef USBULK
#define INCOMING_Q_SIZE 1
#else
/*We'll set it to 6 when we're dealing with ISOCHRONOUS data.*/
#define INCOMING_Q_SIZE 6
#endif

/************************************************************************************/
/* Here's how I got this - Max ethernet   = 1500 bytes + 			    */	
/*          possible Ethernet FCS         = 4 bytes +				    */
/*          possible MPOA encapsulation   = 10 bytes +				    */
/*          CPCS PDU trailer              = 8 bytes +				    */
/*          at most 2 ATM payloads of pad = 96 bytes				    */
/*                                       ----------				    */
/*                                         1618 bytes				    */
/*          Divided among ATM cells = 34 ATM cells * 53 bytes			    */	
/*                      Grand total = 1802					    */
/* NOTE NOTE NOTE NOTE								    */	
/* This logic all makes sense IF the device received data in ATM cell sized chunks, */
/* however, it doesn't. The USB pipes receive 64bytes at a time, so we have         */
/* to have space for a size that is the LCM of 64 and 53, which is 64*53            */
/************************************************************************************/
#define INCOMING_DATA_SIZE (64 * 53)
typedef struct
{
    GENERIC_BUFFER GB;
    void          *pHw;
    urb_t          Urb;
} INCOMING_BUFFER;

/************************************************************************************/
/* This will be where we build our segmentation array of ATM cells to be sent	    */
/* out over the ADSL link.  OUTGOING_Q_SIZE used to be set at 16 - but when	    */
/* using WhatRoute to do large pings - we'd get too many ethernet packets too	    */
/* fast and we'd run out of outgoing buffers. So, I've upped it to 64 - we	    */
/* should never run out now.							    */
/************************************************************************************/
#define OUTGOING_Q_SIZE    64
#define OUTGOING_DATA_SIZE 1802 // same logic as for INCOMING_DATA_SIZE
typedef struct
{
    GENERIC_BUFFER GB;
    urb_t          Urb;
    UInt32         TotalBytes;
} OUTGOING_BUFFER;

#endif // BUFFERS_H
