/***********************************************************************************/
/* Copyright (c) 2001, Analog Devices Inc., All Rights Reserved			   */
/*										   */	
/* Dsp.c									   */
/*										   */
/* Code for load and processing of Dsp code					   */
/*										   */
/* This file is part of the "ADI USB ADSL Driver for Linux".			   */
/*										   */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it 	   */	
/* and/or modify it under the terms of the GNU General Public License as 	   */	
/* published by the Free Software Foundation; either version 2 of the License, 	   */	
/* or (at your option) any later version.					   */
/*										   */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be      */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	   */	
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		   */
/* GNU General Public License for more details.					   */
/*										   */	
/* You should have received a copy of the GNU General Public License		   */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software  */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA       */
/***********************************************************************************/

#include "Adiutil.h"

/***********************************************************************************/
/* Local only prototypes ...							   */
/***********************************************************************************/
static UInt32 itoi(UInt8 *h, UInt32 l);

/***********************************************************************************/
/* DspFoodProcessor								   */
/*										   */
/* The DSP code is in motorolla S-record format. We need to process it into	   */
/* blocks and pages so we can intelligently send it to the DSP.			   */
/***********************************************************************************/
#define IFNULL_JUMP(ptr,msg) \
    if (!ptr) \
    { \
        ZAP(DEBUG_DSP, DEBUG_ERRORS, msg); \
        ret = false; \
        goto Food; \
    }

Boolean DspFoodProcessor(Hardware *pHw)
{
    Boolean ret = true;
    UInt32 i, Flag;
    UInt32 MainPageFileOffset;
    UInt32 SwapPageCount, MainBlockCount;
    UInt8 *p;
    UInt8 *pBuf = pHw->pDsp;
    IDMAPage *pPages = NULL;
    IDMABlock *pBlocks = NULL;

    ZAP(DEBUG_DSP, DEBUG_ENTRY, ("+++ProcessADSLFirmware\n"));

    /*********************************************************/
    /* If we get here, then we know all of the firwmare files*/
    /* were downloaded successfully. We need to process the  */
    /* ADSL firmware data into MAIN and SWAP page records    */
    /*********************************************************/
    ZAP(DEBUG_DSP, DEBUG_INFO, (kTAB"Begin DSP processing. IDMAStart=%#X\n", pHw->IDMAStart));

    /**************************************************************************/
    /* Now the real processing begins ...... 				      */	
    /* Note that I do not completely understand all that is happening here.   */
    /* This code is dealing with the processing of the S records from the .BNM*/
    /* files. If it doesn't work - I'll dig into understanding it - otherwise,*/
    /* it will stay a mystery until someone reads up on the S record format.  */
    /**************************************************************************/
    p = pBuf + pHw->IDMAStart;
    MainPageFileOffset = itoi(p, 4); p+=4;
    SwapPageCount      = itoi(p, 1); p+=1;
    ZAP(DEBUG_DSP, DEBUG_INFO, (kTAB"MainPageFileOffset=%#X SwapPageCount=%#X\n", 
                                  MainPageFileOffset, SwapPageCount));
    pHw->SwapPageCount = SwapPageCount;

    /*Allocate space for our swap page array*/
    GET_DSP_MEMORY(pHw, pPages, (IDMAPage *), sizeof(IDMAPage)*SwapPageCount);
    IFNULL_JUMP(pPages, (kTAB"Not enough memory for swap pages.\n"));
    pHw->SwapPages = pPages;

    /*Allocate memory for swap page blocks and get their file offsets*/
    ZAP(DEBUG_MEMORY, DEBUG_INFO, (kTAB"...loop for allocating swap page blocks\n"));
    for (i=0; i<SwapPageCount; i++)
    {
        GET_DSP_MEMORY(pHw, pBlocks, (IDMABlock *),
                       sizeof(IDMABlock)*2); // swap pages are two blocks at most
        IFNULL_JUMP(pBlocks, (kTAB"Not enough memory for swap page blocks.\n"));
        pPages[i].Blocks = pBlocks;
        pPages[i].Blocks[0].FileOffset = itoi(p, 4); p+=4; 
    }

    /*Figure out the actual information about each swap page*/
    ZAP(DEBUG_MEMORY, DEBUG_INFO, (kTAB"...figure out swap page info\n"));
    for (i=0; i<SwapPageCount; i++)
    {
        UInt32 ExtSize;
        if (pPages[i].Blocks[0].FileOffset == 0)
           continue; /*make sure this page actually exists before processing*/
        
        p = pBuf + pPages[i].Blocks[0].FileOffset;
        pPages[i].BlockCount                = itoi(p, 1); p+=1;
        pPages[i].Blocks[0].DSPAddr         = itoi(p, 2); p+=2;
        pPages[i].Blocks[0].DSPSize         = itoi(p, 2); p+=2;
        pPages[i].Blocks[0].DSPExtendedSize = itoi(p, 2); p+=2;
        ExtSize = pPages[i].Blocks[0].DSPExtendedSize;
        Flag = pPages[i].Blocks[0].DSPAddr & 0x4000;
        /*If this is data memory, we can leave it alone*/
        if (Flag)
        {
           pPages[i].Blocks[0].MemOffset = p;
           pPages[i].Blocks[0].IsAlloced = FALSE;
        }
        else
        /**********************************************************************/
        /* However, if it is program memory, we have to stuff pad bytes. Since*/
        /* each PM word is 3 bytes, 21 "words" = 63 bytes. USB transfers are  */
        /* broken into 64 byte chunks by the low level drivers/hardware. In   */
        /* order to keep from sending partial words, the firmware expects each*/
        /* 64 byte chunk to end in a pad byte - he throws it away. So, we have*/
        /* to make a new PM buffer that has these pad bytes		      */	
        /**********************************************************************/
        {
            UInt32 a, b;
            UInt8 *pNewPM = NULL;

            ZAP(DEBUG_MEMORY, DEBUG_INFO, (kTAB"...swap page is program memory, reprocessing\n"));
            /*Allocate the same number of 64byte chunks as 63byte chunks we already have*/
            GET_DSP_MEMORY(pHw, pNewPM, (UInt8 *), (((ExtSize/63)+1)*64));
            IFNULL_JUMP(pNewPM, (kTAB"Not enough memory for main page program memory.\n"));
            pPages[i].Blocks[0].MemOffset = pNewPM;
            pPages[i].Blocks[0].IsAlloced = TRUE;
            /*a = source index, b = dest index*/
            for (a=0, b=0; a<ExtSize; a+=3, b+=3)
            {
                /*are we at byte 63? (which would really be byte 64)*/
                if (b && ((a%63)==0))
                {
                   pNewPM[b++] = 0xFF; // stuff in an extra byte
                   pPages[i].Blocks[0].DSPExtendedSize++; // add it to the total length
                }
                pNewPM[b  ] = p[a];
                pNewPM[b+1] = p[a+1];
                pNewPM[b+2] = p[a+2];
            }
        }

        p += ExtSize;

        if (pPages[i].BlockCount == 2)
        {
           pPages[i].Blocks[1].DSPAddr         = itoi(p, 2); p+=2;
           pPages[i].Blocks[1].DSPSize         = itoi(p, 2); p+=2; 
           pPages[i].Blocks[1].DSPExtendedSize = itoi(p, 2); p+=2;
           ExtSize = pPages[i].Blocks[1].DSPExtendedSize;
           Flag = pPages[i].Blocks[1].DSPAddr & 0x4000;
           /*If this is data memory, we can leave it alone*/
           if (Flag)
           {
              pPages[i].Blocks[1].MemOffset = p;
              pPages[i].Blocks[1].IsAlloced = FALSE;
           }
           else
           /**********************************************************************/
           /* However, if it is program memory, we have to stuff pad bytes. Since*/
           /* each PM word is 3 bytes, 21 "words" = 63 bytes. USB transfers are  */
           /* broken into 64 byte chunks by the low level drivers/hardware. In   */
           /* order to keep from sending partial words, the firmware expects each*/
           /* 64 byte chunk to end in a pad byte - he throws it away. So, we have*/
           /* to make a new PM buffer that has these pad bytes			 */
           /**********************************************************************/
           {
              UInt32 a, b;
              UInt8 *pNewPM = NULL;
        
              /*Allocate the same number of 64byte chunks as 63byte chunks we already have*/
              GET_DSP_MEMORY(pHw, pNewPM, (UInt8 *), (((ExtSize/63)+1)*64));
              IFNULL_JUMP(pNewPM, (kTAB"Not enough memory for main page program memory.\n"));
              pPages[i].Blocks[1].MemOffset = pNewPM;
              pPages[i].Blocks[1].IsAlloced = TRUE;
              /*a = source index, b = dest index*/
              for (a=0, b=0; a<ExtSize; a+=3, b+=3)
              {
                  /*are we at byte 63? (which would really be byte 64)*/
                  if (b && ((a%63)==0))
                  {
                     pNewPM[b++] = 0xFF; // stuff in an extra byte
                     pPages[i].Blocks[1].DSPExtendedSize++; // add it to the total length
                  }
                  pNewPM[b  ] = p[a];
                  pNewPM[b+1] = p[a+1];
                  pNewPM[b+2] = p[a+2];
              }
          }
        }
    }

    p = pBuf + MainPageFileOffset;
    MainBlockCount = itoi(p, 1); p+=1;

    /*Get space for main page blocks*/
    GET_DSP_MEMORY(pHw, pBlocks, (IDMABlock *), sizeof(IDMABlock)*MainBlockCount);
    IFNULL_JUMP(pBlocks, (kTAB"Not enough memory for main page blocks.\n"));
    pHw->MainPage.BlockCount = MainBlockCount;
    pHw->MainPage.Blocks     = pBlocks;

    for (i=0; i<MainBlockCount; i++)
    {
        p = pBuf + MainPageFileOffset + 1 + i*0xC00;
        pBlocks[i].FileOffset      = MainPageFileOffset + 1 + i*0xC00;
        pBlocks[i].DSPAddr         = itoi(p, 2); p+=2;
        Flag = pBlocks[i].DSPAddr & 0x4000;
        pBlocks[i].DSPSize         =(itoi(p, 2))*((Flag)?2:3); p+=2;  
        pBlocks[i].DSPExtendedSize = itoi(p, 2); p+=2;
        /*If this is data memory, we can leave it alone*/
        if (Flag)
        {
           pBlocks[i].MemOffset = p;
           pBlocks[i].IsAlloced = FALSE;
        }
        else
        /***********************************************************************/ 
        /* However, if it is program memory, we have to stuff pad bytes. Since */
        /* each PM word is 3 bytes, 21 "words" = 63 bytes. USB transfers are   */
        /* broken into 64 byte chunks by the low level drivers/hardware. In    */
        /* order to keep from sending partial words, the firmware expects each */
        /* 64 byte chunk to end in a pad byte - he throws it away. So, we have */ 
        /* to make a new PM buffer that has these pad bytes		       */	
        /***********************************************************************/ 
        {
           UInt32 a, b, Size;
           UInt8 *pNewPM = NULL;

           Size = pBlocks[i].DSPSize;
           /*Allocate the same number of 64byte chunks as 63byte chunks we already have*/
           GET_DSP_MEMORY(pHw, pNewPM, (UInt8 *), (((Size/63)+1)*64));
           IFNULL_JUMP(pNewPM, (kTAB"Not enough memory for main page program memory.\n"));
           pBlocks[i].MemOffset = pNewPM;
           pBlocks[i].IsAlloced = TRUE;
           /***********************************************************************/ 
           /*If this is the "first" block (which will be the last one sent),      */    
           /* we need to offset our padding by 3 bytes, 'cause we don't include   */
           /* the first 3 bytes in the full block send. The DSP starts up when    */
           /* offset 0 is written, so we save those 3 bytes until last.		  */
           /* (Yes, it really does make sense.)					  */
           /***********************************************************************/ 
            if (pBlocks[i].DSPAddr == 0)
            {
               pNewPM[0] = p[0];
               pNewPM[1] = p[1];
               pNewPM[2] = p[2];
               pNewPM += 3;
               p      += 3;
            }

            /*a = source index, b = dest index*/
            for (a=0, b=0; a<Size; a+=3, b+=3)
            {
                /*are we at byte 63? (which would really be byte 64)*/
                if (b && ((a%63)==0))
                {
                   pNewPM[b++] = 0xFF;     // stuff in an extra byte
                   pBlocks[i].DSPSize++;   // add it to the total length
                }
                pNewPM[b  ] = p[a];
                pNewPM[b+1] = p[a+1];
                pNewPM[b+2] = p[a+2];
            }
        }
    }

Food:
    ZAP(DEBUG_DSP, DEBUG_EXIT, ("---ProcessADSLFirmware\n"));
    return ret;
}

/****************************************************************************/
/* itoi - straight from NDIS code					    */
/*      - does nibble swapping for endian conversions			    */
/****************************************************************************/
static UInt32 itoi(UInt8 *h, UInt32 l)
{

  UInt32 num, i;

  for (num = 0, i=0; i<l ; i++)
  {
      num<<=8;
      num+=h[i];
  }

  return num;
}

/****************************************************************************/
/* FreeDspData								    */
/*									    */
/* Releases memory allocated for storage of firmware code		    */	
/****************************************************************************/
void FreeDspData(Hardware *pHw)
{
    UInt32 i;
    #ifdef ZAPS
    UInt32 total = 0;
    #endif

    ZAP(DEBUG_DSP, DEBUG_ENTRY, ("+++FreeDspData\n"));

    ZAP(DEBUG_DSP, DEBUG_INFO, (kTAB"Freeing %#X Dsp allocs\n", pHw->DspAllocCount));

    for (i=0; i<pHw->DspAllocCount; i++)
    {
#ifdef ZAPS
        total += pHw->DspAllocs[i].AllocedSize;
#endif
        FREE_DSP_MEMORY(pHw->DspAllocs[i]);
    }

    pHw->DspAllocCount = 0;
    ZAP(DEBUG_DSP, DEBUG_INFO, (kTAB"%#X bytes of memory were alloced for DSP\n", total));
    ZAP(DEBUG_DSP, DEBUG_EXIT, ("---FreeDspData\n"));
}
