/***************************************************************************************/
/* Copyright (c) 2001, Analog Devices Inc., All Rights Reserved			       */
/*										       */
/* Me.C (Management Entity)							       */	
/*										       */
/* Contains functions to initialize & refresh Modem State Machine.		       */
/* Invoked at startup, runs on secondary USB IRQ and on timer tick.		       */
/*										       */
/* Calls internally Sm.C, which contains detailed case handling.		       */
/*										       */
/*										       */  	
/* This file is part of the "ADI USB ADSL Driver for Linux".			       */
/*										       */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it 	       */
/* and/or modify it under the terms of the GNU General Public License as               */
/* published by the Free Software Foundation; either version 2 of the License,         */
/* or (at your option) any later version.					       */
/*										       */	
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be 	       */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	       */	
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		       */
/* GNU General Public License for more details.					       */	
/*										       */	
/* You should have received a copy of the GNU General Public License		       */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software      */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	       */
/***************************************************************************************/

#include "Adiutil.h"
#include <linux/sched.h>

/***************************************************************************************/
/*  ProcessIncomingCmv								       */	
/*										       */	
/*  Invoked on secondary System Task Level thread, this routine grabs 		       */
/*  new data from modem and clears TX mailbox status early on.			       */	
/*  Input data are passed immediatelly to State Machine for processing.		       */	
/*  Messages are discriminated to Replies and Alerts according to 		       */	
/*  the type in the header, Replies are always synchronous messages.		       */
/*										       */	
/*  We expect response to arrive from modem within few hundred msecs.		       */	
/*  It gets all messages - either Replies or unsolicited Alerts.		       */
/***************************************************************************************/
void ProcessIncomingCmv(Hardware *pHw, UInt16 *pMsg)
{
    UInt32 rv;
    mp_message Msg; /*decoded message struct*/
    Boolean* pFlag; /*which flag to clear at exit*/


    int UserApiReply=FALSE;					//	


    ZAP(DEBUG_MSG, DEBUG_ENTRY, (">>>ProcessIncomingCmv\n"));

    rv = MsgDecode(&Msg, pMsg);

    if ( Msg.type == MP_FUNCTION_TYPE_ADSLDIRECTIVE )
    {
       /*Alerts from modem, arrive ASYNCHRONOUSLY*/
       pHw->AdiModemSm.InboundAsyncPending = TRUE;
       pFlag = &pHw->AdiModemSm.InboundAsyncPending;
       if ( rv == MP_DECODE_OK )
       {
          ZAP(DEBUG_MSG, DEBUG_INFO, ("   ProcessIncomingCmv: EVENT_RX_ASYNC\n"));
          rv = EVENT_RX_ASYNC;
       }
       else
       {
          ZAP(DEBUG_MSG, DEBUG_ATTN, ("   ProcessIncomingCmv: EVENT_RX_ASYNC_ERROR\n"));
          rv = EVENT_RX_ASYNC_ERROR;
       }
    }
    else
    {
													//	
       if ( (pHw->AdiApiRWPending) && ( (Msg.type==MP_FUNCTION_TYPE_MEMACCESS) 					//
                                ||(Msg.type==MP_FUNCTION_TYPE_MSGDECERR) 					//
                                 ||(Msg.type==MP_FUNCTION_TYPE_MSGACCERR)					//
                                 ||(Msg.type==MP_FUNCTION_TYPE_WAITING_4_REPLY)  ) )				//
      {														//
       // If the message is error reply, set error flag								//
         pHw->AdiModemSm.InboundSyncPending = TRUE;								//
         if ( (Msg.type==MP_FUNCTION_TYPE_MSGDECERR)								//
            ||(Msg.type==MP_FUNCTION_TYPE_MSGACCERR))								//
         {													//	
            pHw->APICmvError=1;											//
         }													//
         else													//
            pHw->APICmvError=0;											//
         pFlag = &pHw->AdiModemSm.InboundSyncPending;								//
      //													//       
      // Copy data into hardware specific data (both encoded row message and 32 bit data) structure and wake up //
      // the user process											//
      //													//
         UserApiReply = TRUE;											//
         memcpy((char *)(pHw->ApiCmvRaw),(char *)pMsg,2*kCMV_DATA_WORDS);					//
         pHw->ApiCmvData = Msg.data;										//
      //													//														
         pHw->AdiApiRWPending = FALSE;										//
         													//
         wake_up_interruptible(&(pHw->wq)); 									//
       }													//
       else													//
       ///////////////////////////////////////////////////////////////////////////////////////////////////////////	
       {
         /*Replies from modem, arrive SYNCHRONOUSLY*/
         pHw->AdiModemSm.InboundSyncPending = TRUE;
         pFlag = &pHw->AdiModemSm.InboundSyncPending;
         if ( rv == MP_DECODE_OK )
         {
            ZAP(DEBUG_MSG, DEBUG_INFO, ("   ProcessIncomingCmv: EVENT_RX_SYNC\n"));
            rv = EVENT_RX_SYNC;
         }
         else
         {
            ZAP(DEBUG_MSG, DEBUG_ATTN, ("   ProcessIncomingCmv: EVENT_RX_SYNC_ERROR\n"));
            rv = EVENT_RX_SYNC_ERROR;
         }
         /* when SYNC reply arrives, we can clear outbound CMV flag*/
         /* it is set in <MsgSend> fxn, that may be called from <ModemSM>*/
      }
      pHw->AdiModemSm.OutboundPending = FALSE;
    }
    
    if ( (pHw->Queue.Pending) && ( Msg.type != MP_FUNCTION_TYPE_ADSLDIRECTIVE ) )			//
    {													//
       if (pHw->Queue.Type==1)										//
         UserMsgSend(pHw, pHw->Queue.CmvRaw);								//
       else    												//
         MsgSend(pHw, pHw->Queue.CmvRaw);								//
       pHw->Queue.Pending = FALSE;									//
       													//
    }													//	
    if  (!UserApiReply) 										 //
    ///////////////////////////////////////////////////////////////////////////////////////////////////////
     ModemSM((UInt16)rv, pHw, &Msg);
       
    *pFlag = FALSE;
    ZAP(DEBUG_MSG, DEBUG_EXIT, ("<<<ProcessIncomingCmv\n"));
}


/***************************************************************************************/
/*										       */	
/*  WatchdogTimer								       */
/*										       */
/*  This routine is of the type <SecondaryInterruptHandler2> for timer,		       */
/*  guaranteed to be called at <kSecondaryInterruptLevel>, never <kTaskLevel>	       */
/*										       */
/*  Its main purpose is to detect modem gone AWOL, and to kick-off RESET.	       */
/*  If some USB command transaction did not complete, that's failure.		       */
/*  If there is no USB command sequence going on at the moment, 		       */	
/*  the function will initiate "Poll Status" for modem periodically.		       */
/*										       */
/*  To avoid too frequent polling, "at ease" interval is set to 1 sec or so.	       */
/*										       */
/*  When this callback is invoked, pGlobals->timerSM is already released.	       */	
/*  No need to do anything w/ it, only set to NULL just in case.		       */	
/***************************************************************************************/
void WatchdogTimer(unsigned long ptr)
{
    Hardware* pHw = (Hardware*)ptr;
    AdiMSM*   pAdiSm = &(pHw->AdiModemSm);

    ZAP(DEBUG_SM, DEBUG_ENTRY, ("+++Watchdog timer entry (%lx)\n", ptr));

    if ( pAdiSm->ModemReplyExpected == MP_FUNCTION_TYPE_WAITING_4_REPLY )
    {
       ZAP(DEBUG_SM,DEBUG_ERRORS,(kTAB"Watchdog timer found modem error - starting RESET!\n"));
       /*modem AWOL, initiate reset and terminate Modem SM*/

      if (pAdiSm->CurrentAdiState!=STATE_DISABLE)						//
         ResetModemSM(pAdiSm);
    }

    /*call State Machine to process the TIMER_TICK event*/
    if ( pHw->AdiModemSm.InboundSyncPending  == FALSE &&
         pHw->AdiModemSm.InboundAsyncPending == FALSE )
    {
       ZAP(DEBUG_SM,DEBUG_INFO,(kTAB"Timer tick going to SM\n"));

       ModemSM(EVENT_TIMER_TICK, pHw, NULL);
    }
#ifdef ZAPS  // STATE_HARD_RESET_INITIATE
    else
    {
       ZAP(DEBUG_SM,DEBUG_ATTN,(kTAB"Timer tick swallowed, inbound msg pending...\n"));
    }
#endif

    /*if no timer is set inside State Machine, set default timeout*/
    if ( !timer_pending(&pAdiSm->timerSM) )
    {
       ZAP(DEBUG_SM,DEBUG_INFO,(kTAB"Default timeout value set (%x)\n", TRANS_TIME_HEARTBEAT));
       SetTimerInterval(pHw, TRANS_TIME_HEARTBEAT);
    }

    ZAP(DEBUG_SM, DEBUG_EXIT, ("---Watchdog timer exit\n"));
}

/***************************************************************************************/
/*										       */
/*  SetTimerInterval								       */	
/*										       */
/*  This utility sets next timer interval, as requested by caller.		       */	
/*  Caller provides delay in millisecond units, conversion done inside.		       */
/*  If current timer hasn't expire yet, it is cancelled at once.		       */
/*  New timer is created *automatically* and its Id is placed in pGlobals.	       */	
/*										       */	
/*  WARNING!									       */	
/*  There are no provisions to cope w/ <SetInterruptTimer> failure. 		       */	
/*  MAC OS manuals say, it will always succeed, so in them we trust!		       */
/*										       */	
/*  COMPATIBILITY								       */
/*  This call is "Carbon"-incompatible, good for MAC OS 8 and 9 only.		       */
/*  For MAC OS X, rewrite will be an absolute must.				       */	
/***************************************************************************************/
int SetTimerInterval(Hardware* pHw, int reqDelay)
{
    AdiMSM   *pAdiSm = &(pHw->AdiModemSm);
    int curtime = jiffies;
    int delay   = MSEC_TO_JIFFIES(reqDelay);
    int exptime;

    ZAP(DEBUG_SM, DEBUG_ENTRY, ("+++SetTimerInterval(%d)\n", reqDelay));

    /*If the timer is currently running, delete it*/
    if ( timer_pending(&pAdiSm->timerSM) )
       del_timer(&pAdiSm->timerSM);

    pAdiSm->CurrentExpirationTime = curtime;
    if ( reqDelay == 0 )
       goto SetTimer_exit;

    /*calculate the minimal delay requested*/
    if ( curtime < pAdiSm->SwapPageRequiredTime )
    {
       int newDelay;
       newDelay = pAdiSm->SwapPageRequiredTime - curtime;
       if ( delay < newDelay )
          delay = newDelay;
    }

    exptime = curtime + delay;
    pAdiSm->timerSM.expires = exptime;
    pAdiSm->CurrentExpirationTime = exptime;
    
    /*we expect this system call to succeed always*/
    add_timer(&pAdiSm->timerSM);

SetTimer_exit:    
    ZAP(DEBUG_SM, DEBUG_EXIT, ("---SetTimerInterval\n"));
    return USB_ST_NOERROR;
}


/***************************************************************************************/
/*  AdjustTimerInterval								       */
/*										       */
/*  This utility checks and sets next timer interval if necessary,		       */
/*  to a value not less than the parameter supplied.				       */
/*										       */
/*  There are two variables in iour structure we deal with:			       */
/*    SwapPageRequierdTime  - it's shadow for the time when swap would be over;        */ 
/*                            its value only grows, it *never* gets back!	       */ 	
/*    CurrentExpirationTime - it's shadow for system timer expiration value;	       */
/*                            it may be cancelled and set again later.		       */
/*  Both vars contain *absolute* time in duration units, since system startup.         */
/*										       */
/*  The service routine InitTime() returns current duration since startup,	       */	
/*  in *absolute* duration units.						       */
/*										       */
/*  The argument <reqDelay> is in relative duration units;			       */	
/*      absoluteExpirationTime = absoluteCurrentTime + requiredDelay		       */	
/*										       */
/*  The reason for using two separate vars is because two clients use timer.	       */	
/*  The basic client is SM - it sets the timeout interval after which 		       */
/*  the timer triggers sending out next CMV to modem for execution.		       */
/*										       */	
/*  It is expected that modem would either reply to CMV, or request a swap,	       */
/*  but *never*ever* do both things at once in two consequitive replies.	       */
/***************************************************************************************/
int AdjustTimerInterval(Hardware* pHw, int reqDelay)
{
    if ( reqDelay > 0 )
    {
       AdiMSM   *pAdiSm = &(pHw->AdiModemSm);
       int setTime = MSEC_TO_JIFFIES(reqDelay) + jiffies;
    
       if ( pAdiSm->SwapPageRequiredTime >= setTime )
          goto AdjustTimer_exit;
                                  
       pAdiSm->SwapPageRequiredTime = setTime;
    
       if ( pAdiSm->CurrentExpirationTime > setTime ) 
          goto AdjustTimer_exit;

       if ( timer_pending(&pAdiSm->timerSM) )
          return SetTimerInterval(pHw, reqDelay);
    }
AdjustTimer_exit:
    return USB_ST_NOERROR;
}

//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//

#include "Sm.c"

//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//\\//

