/**********************************************************************************/
/* Copyright (c) 2001, Analog Devices Inc., All Rights Reserved			  */
/*										  */	
/* Mpoa.c									  */
/*										  */
/* RFC2684(1483) "Multi-Protocol Encapsulation Over ATM AAL5" support module      */
/*										  */
/* This file is part of the "ADI USB ADSL Driver for Linux".			  */
/*										  */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it      */
/* and/or modify it under the terms of the GNU General Public License as 	  */
/* published by the Free Software Foundation; either version 2 of the License,    */
/* or (at your option) any later version.					  */
/*										  */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be     */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		  */
/* GNU General Public License for more details.					  */
/*										  */
/* You should have received a copy of the GNU General Public License		  */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA      */
/**********************************************************************************/

#include "Adiutil.h"

/**********************************************************************************/
/* Local prototypes								  */
/**********************************************************************************/
static Boolean MpoaFilterReceivedPacket(Hardware *pHw, ETHERNET_MAC_HEADER *pMac);

/**********************************************************************************/
/* These static data structures are the encapsulation headers we need for Mpoa    */
/**********************************************************************************/

#define IS_LLCBRIDGEDETHNOFCS(p) (\
    (*((UInt32 *)p)     == 0x0003AAAA) && \
    (*((UInt32 *)(p+4)) == 0x0700C280) && \
    (*((UInt32 *)(p+6)) == 0x00000700))

static LLC_HEADER MpoaHdr_LlcEncapBridgedEthNoFcs =
    { {0xAA,0xAA,0x03},
      {0x00,0x80,0xC2},
      {0x00,0x07},
      {0x00,0x00}
    };

#define IS_LLCBRIDGEDETHFCS(p) (\
    (*((UInt32 *)p)     == 0x0003AAAA) && \
    (*((UInt32 *)(p+4)) == 0x0100C280) && \
    (*((UInt32 *)(p+6)) == 0x00000100))

#ifdef LATER_LATER_LATER // this isn't used right now
static LLC_HEADER MpoaHdr_LlcEncapBridgedEthFcs =
    { {0xAA,0xAA,0x03},
      {0x00,0x80,0xC2},
      {0x00,0x01},
      {0x00,0x00}
    };
#endif

#define IS_LLCROUTEDIP(p) (\
    (*((UInt32 *)p)     == 0x0003AAAA) && \
    (*((UInt32 *)(p+4)) == 0x00080000))

static LLC_HEADER MpoaHdr_LlcEncapRoutedIp =
    { {0xAA,0xAA,0x03},
      {0x00,0x00,0x00},
      {0x08,0x00}
    };

#define IS_VCBRIDGEDETH(p) (\
    (*((UInt16 *)p)     == 0x0000))

static UInt8 MpoaHdr_VcEncapBridgedEth[ MPOA_VC_ENCAP_BRIDGED_ETH_SIZE] =
    { 0x00,0x00 };

static UInt8 MpoaHdr_LlcEncapPPPoA[ MPOA_LLC_ENCAP_PPPOA_SIZE ] = { 0xFE, 0xFE, 0x03, 0xCF };

#define IS_LLCPPPOA(p) (\
    (*((UInt32 *)p) == 0xFEFE03CF))

static UInt8 MacBroadcast[ ETH_LENGTH_OF_ADDRESS ] =
    { 0xFF,0xFF,0xFF,0xFF,0xFF,0xFF };

#define EQUAL_MACHDR(a1,a2) (\
    (*((UInt32 *)a1) == *((UInt32 *)a2)) && \
    (*((UInt32 *)(a1+2)) == *((UInt32 *)(a2+2))))


/**********************************************************************************/
/* MpoaInitialize								  */
/* Prepares for MPOA processing by setting up the proper encapsulation		  */
/**********************************************************************************/
void MpoaInitialize( Hardware *pHw, UInt32 MpoaMode )
{
    UInt16 HdrSize=0;
    UInt8 *pHdr=NULL;

    ZAP(DEBUG_MPOA, DEBUG_ENTRY, ("MpoaInitialize\n"));

    /*Setup the encapsulation header and function pointer based on the mode specified*/
    switch (MpoaMode)
    {
       case MPOA_MODE_BRIDGED_ETH_LLC:
          ZAP(DEBUG_MPOA, DEBUG_INFO, ("MpoaInitialize - MPOA mode set at: MPOA_MODE_BRIDGED_ETH_LLC\n"));
          pHdr    = (UInt8 *)&MpoaHdr_LlcEncapBridgedEthNoFcs;
          HdrSize = MPOA_LLC_ENCAP_BRIDGED_ETH_SIZE;
          break;
       case MPOA_MODE_BRIDGED_ETH_VC:
          ZAP(DEBUG_MPOA, DEBUG_INFO, ("MpoaInitialize - MPOA mode set at: MPOA_MODE_BRIDGED_ETH_VC\n"));
          pHdr    = (UInt8 *)&MpoaHdr_VcEncapBridgedEth;
          HdrSize = MPOA_VC_ENCAP_BRIDGED_ETH_SIZE;
          break;
       case MPOA_MODE_ROUTED_IP_LLC:
          ZAP(DEBUG_MPOA, DEBUG_INFO, ("MpoaInitialize - MPOA mode set at: MPOA_MODE_ROUTED_IP_LLC\n"));
          pHdr    = (UInt8 *)&MpoaHdr_LlcEncapRoutedIp;
          HdrSize = MPOA_LLC_ENCAP_ROUTED_IP_SIZE;
          break;
       case MPOA_MODE_ROUTED_IP_VC:
          ZAP(DEBUG_MPOA, DEBUG_INFO, ("MpoaInitialize - MPOA mode set at: MPOA_MODE_ROUTED_IP_VC\n"));
          /* Since there is no ethernet header and the encapsulation is already*/
          /* a part of the VC by definition, no additional header info needs   */
          /* stuffed into packets when in this mode			       */	
          pHdr    = NULL;
          HdrSize = 0;
          break;
       case MPOA_MODE_PPPOA_LLC:
          ZAP(DEBUG_MPOA, DEBUG_ATTN, ("MpoaInitialize - MPOA mode set at: MPOA_MODE_PPPOA_LLC\n"));
          pHdr    = (UInt8 *)&MpoaHdr_LlcEncapPPPoA;
          HdrSize = MPOA_LLC_ENCAP_PPPOA_SIZE;
          break;
       case MPOA_MODE_PPPOA_VC:
          ZAP(DEBUG_MPOA, DEBUG_ATTN, ("MpoaInitialize - MPOA mode set at: MPOA_MODE_PPPOA_VC\n"));
          pHdr    = NULL;
          HdrSize = 0;
          break;
       /**************************************************************************/
    }

    /**********************************************************************************/
    /* The NDIS code has a distinction between "creating" and "activating" the        */
    /* virtual channel. That code is only necessary when dealing with higher-level    */
    /* software that understands and manages virtual channels (like NDIS 5). 	      */	
    /* Since this is the Mac ethernet driver, we know that nobody above us is going   */
    /* to talk ATM, so we combined both functions into one, UniEstablishVc.           */
    /**********************************************************************************/
    UniEstablishVc(pHw, pHdr, HdrSize);

    pHw->MpoaMode = MpoaMode;

    ZAP(DEBUG_MPOA, DEBUG_EXIT, ("MpoaInitialize\n"));
}

/**********************************************************************************/
/* MpoaProcessReceivedPdu							  */
/*										  */
/* This routine will be called by the receive data routine by the "user"	  */	
/* of this driver. I thought it was best to delay as much processing so		  */
/* we could exit the data receiver handler as soon as possible			  */
/**********************************************************************************/
void MpoaProcessReceivedPdu(Hardware *pHw, ETHERNET_PACKET_BUFFER *pBuf)
{
    ETHERNET_MAC_HEADER *pMac;

    ZAP(DEBUG_MPOA, DEBUG_ENTRY, ("+++MpoaProcessReceivedPdu\n"));

    switch (pHw->MpoaMode)
    {
       case MPOA_MODE_BRIDGED_ETH_LLC:
          if (IS_LLCBRIDGEDETHNOFCS(pBuf->GB.pData))
          {
             /* We've got an LLC encapsulated ethernet packet without FCS,*/
             /* since our "user" is ethernet, skip over the LLC header    */
             pBuf->EncapSkipSize = MPOA_LLC_ENCAP_BRIDGED_ETH_SIZE;
          }
          else
          {
             if (IS_LLCBRIDGEDETHFCS(pBuf->GB.pData))
             {
                /* We've got an LLC encapsulated ethernet packet with FCS,*/
                /* since our "user" is ethernet, skip over the LLC header */
                pBuf->EncapSkipSize = MPOA_LLC_ENCAP_BRIDGED_ETH_SIZE;
             }
             else
             {
#ifdef ZAPS_VERBOSE
                UInt32 i;
                /* The header for this Pdu did not match our stored LLC header,*/
                /* so we cannot with good faith accept this Pdu		       */	 
                for (i=0; i<10; i++)
                    ZAP(DEBUG_MPOA, DEBUG_INFO,
                        (kTAB"MPOA header byte %x = 0x%x\n", i, pBuf->GB.pData[i]));
#endif
                ZAP(DEBUG_MPOA, DEBUG_ERRORS, (kTAB"MpoaProcessReceivedPdu: invalid LLC header\n"));
                pBuf->DropThisPacket = TRUE;
             }
          } 
          break;
       case MPOA_MODE_BRIDGED_ETH_VC:
          if (IS_VCBRIDGEDETH(pBuf->GB.pData))
          {
             /* We've got a VC encapsulated ethernet packet,	     */	
             /* since our "user" is ethernet, skip over the VC header*/
             pBuf->EncapSkipSize = MPOA_VC_ENCAP_BRIDGED_ETH_SIZE;
          }
          else
          {
             /* The header for this Pdu did not match our stored VC header,*/
             /* so we cannot with good faith accept this Pdu		   */
             ZAP(DEBUG_MPOA, DEBUG_ERRORS, (kTAB"MpoaProcessReceivedPdu: invalid VC header\n"));
             pBuf->DropThisPacket = TRUE;
          }
          break;
       case MPOA_MODE_PPPOA_LLC:
          if (IS_LLCPPPOA(pBuf->GB.pData))
          {
             /* We've got an LLC encapsulated PPPoA packet,      */
             /* since our "user" is PPP, skip over the LLC header*/
             pBuf->EncapSkipSize = MPOA_LLC_ENCAP_PPPOA_SIZE;
          }
          else
          {
             /* We're in LLC PPPOA mode, but this packet does not have an LLC*/
             /* header, so we cannot with good faith accept this Pdu         */
             ZAP(DEBUG_MPOA, DEBUG_ERRORS, ("MpoaProcessReceivedPdu: invalid LLC header\n"));
             pBuf->DropThisPacket = TRUE;
          }
          break;
       case MPOA_MODE_PPPOA_VC:
          /* In PPPOA_VC mode, there is no MPOA header, so we don't do skip anything*/
          pBuf->EncapSkipSize = 0;
          break;
       /****************************************************************************/

		   case MPOA_MODE_ROUTED_IP_LLC:	
			if (IS_LLCROUTEDIP(pBuf->GB.pData))
			{
				// We've got an LLC encapsulated routed IP packet, skip over the LLC header
				ZAP(DEBUG_MPOA, DEBUG_ERRORS, ("MpoaProcessReceivedPdu: We've got an LLC encapsulated routed IP packet, skip over the LLC header\n"));
				pBuf->EncapSkipSize = MPOA_LLC_ENCAP_ROUTED_IP_SIZE;
			}
			else
			{
				// We're in LLC routed IP mode, but this packet does not have an LLC
				// header, so we cannot with good faith accept this Pdu
				ZAP(DEBUG_MPOA, DEBUG_ERRORS, ("MpoaProcessReceivedPdu: invalid LLC header\n"));
				pBuf->DropThisPacket = TRUE;
			}
			break;
               
		   case MPOA_MODE_ROUTED_IP_VC:		
			// In ROUTED_IP_VC mode, there is no MPOA header, so we don't skip anything
			pBuf->EncapSkipSize = 0;
			break;
       	   default:
          ZAP(DEBUG_MPOA, DEBUG_ERRORS, (kTAB"MpoaProcessReceivedPdu: unsupported MpoaMode\n"));
          pBuf->DropThisPacket = TRUE;
    }


    /****************************************************************************/
    /* We've determined any header size that we need to skip and verified that  */
    /* we have a valid header. If we determined that this Pdu should be dropped,*/
    /* we shouldn't do any more processing					*/ 	
    /****************************************************************************/
    if (!pBuf->DropThisPacket)
    {
       pMac = (ETHERNET_MAC_HEADER *)(pBuf->GB.pData + pBuf->EncapSkipSize);

       switch (pHw->MpoaMode)
       {
 		 case MPOA_MODE_ROUTED_IP_LLC:	
  		 case MPOA_MODE_ROUTED_IP_VC:	
			ZAP(DEBUG_FFD, DEBUG_ERRORS, ("MpoaProcessReceivedPdu: Routed: NO Packet filtering!!\n"));
		  break;
         case MPOA_MODE_PPPOA_LLC:
	     case MPOA_MODE_PPPOA_VC:
             break;
         default:
             pBuf->DropThisPacket = MpoaFilterReceivedPacket(pHw, pMac);
             if (pBuf->DropThisPacket)
                pHw->Statistics[STAT_PAKTS_FILTERED]++;
             break;
       }
       /**************************************************************/
    }

    ZAP(DEBUG_MPOA, DEBUG_EXIT, ("---MpoaProcessReceivedPdu\n"));
}

/**********************************************************************************/
/* MpoaFilterReceivedPacket							  */
/*										  */
/* Checks the MAC header relative to our current filter mode. Returns		  */
/* TRUE if this packet should be filtered (not processed).			  */
/**********************************************************************************/
static Boolean MpoaFilterReceivedPacket(Hardware *pHw, ETHERNET_MAC_HEADER *pMac)
{
    struct net_device *ether = pHw->pLinuxNet;
    Boolean FilterIt = FALSE;

    ZAP(DEBUG_MPOA, DEBUG_ENTRY, ("+++MpoaFilterReceivedPacket: DstMAC= "));
#ifdef ZAPS
    {
        UInt32 i;
        for (i=0; i<ETH_LENGTH_OF_ADDRESS; i++)
            ZAP(DEBUG_MPOA, DEBUG_DATA, ("%x", pMac->DstAddress[i]));
    }
#endif
    ZAP(DEBUG_MPOA, DEBUG_INFO, ("\n"));

    /* If we're in promiscuous mode - then all packets go through!*/
    if (ether->flags & IFF_PROMISC)
    {
       ZAP(DEBUG_MPOA, DEBUG_INFO, (kTAB"IFF_PROMISC set.\n"));
       FilterIt = FALSE;
    }
    else
    {
       /* Lets determine if this is a Directed, Broadcast, or Multicast packet*/
       if (pMac->DstAddress[0] & 0x01)
       {
          /* This is either a Broadcast or Multicast packet*/
          if (EQUAL_MACHDR(pMac->DstAddress, MacBroadcast))
          {
             /* This is a broadcast packet*/
             ZAP(DEBUG_MPOA, DEBUG_INFO, (kTAB"Broadcast packet.\n"));
             FilterIt = FALSE;
          }
          else
          {
             /* This is a multicast packet*/
             if (ether->flags & IFF_ALLMULTI)
             {
                ZAP(DEBUG_MPOA, DEBUG_INFO, (kTAB"IFF_ALLMULTI set.\n"));
                FilterIt = FALSE;
             }
             else
             {
                ZAP(DEBUG_MPOA, DEBUG_INFO, (kTAB"Multicast packet.\n"));
                /* Now we have to compare the DstAddress with our list*/
                /* of acceptable multicast addresses                  */
                if (ether->mc_count > 0)
                {
                   int i;
                   struct dev_mc_list *mc;

                   /* Start out by assuming that we need to filter this*/
                   /* packet, since we will encounter a match only when*/
                   /* this is a valid packet to go through (not filter)*/
                   FilterIt = TRUE;

                   /* This loop works because z will evaluate to 0 or NULL*/
                   /* when z->next is NULL, which will stop the loop      */
                   for (i = 0, mc = ether->mc_list;
                        mc && i < ether->mc_count;
                        i++, mc = mc->next) 
                   {
                       if (EQUAL_MACHDR(pMac->DstAddress, mc->dmi_addr))
                       {
                          FilterIt = FALSE;
                          break; /* We found it - get out of the loop*/
                       }
                   }
                }
                else
                {
                   FilterIt = TRUE;
                }
             }
          }
       }
       else
       {
          /* This is a directed packet*/
          if (EQUAL_MACHDR(pMac->DstAddress, pHw->MAC))
          {
             ZAP(DEBUG_MPOA, DEBUG_INFO, (kTAB"Directed packet match.\n"));
             /* This packet is directed to us*/
             FilterIt = FALSE;
          }
          else
          {
             ZAP(DEBUG_MPOA, DEBUG_INFO, (kTAB"Directed packet not a match.\n"));
             /* This packet is not directed to us */
             FilterIt = TRUE;
          }
       }
    }

#ifdef ZAPS
    if (FilterIt)
    {
       ZAP(DEBUG_MPOA, DEBUG_EXIT, ("---MpoaFilterReceivedPacket TRUE\n"));
    }
    else
    {
       ZAP(DEBUG_MPOA, DEBUG_EXIT, ("---MpoaFilterReceivedPacket FALSE\n"));
    }
#endif
    return FilterIt;
}


void EthernetArpReply(Hardware *pHw, UInt8 *pBuf, UInt32 PacketSize)
{
    ETHERNET_MAC_HEADER *pRequestHdr;
    ETHERNET_MAC_HEADER *pReplyHdr;
    ARP_PACKET *pArp;
    UInt32 SrcIP, DstIP;

    ZAP(DEBUG_ROUTEIP, DEBUG_ENTRY, ("+++EthernetArpReply\n"));

    pRequestHdr = (ETHERNET_MAC_HEADER *)pBuf;
    pReplyHdr   = (ETHERNET_MAC_HEADER *)pHw->pPacket;

    // First, copy the original ARP request into our reply packet
    COPY_MEM(pBuf, pHw->pPacket, PacketSize);

    pArp  = (ARP_PACKET *)(pHw->pPacket + sizeof(ETHERNET_MAC_HEADER));
    SrcIP = *(UInt32 *)(pArp->Spa);
    DstIP = *(UInt32 *)(pArp->Tpa);

    ZAP(DEBUG_ROUTEIP, DEBUG_INFO, ("ARP Src IP: "));
    ZAP(DEBUG_ROUTEIP, DEBUG_INFO, ("%d.%d.%d.%d", pArp->Spa[0], pArp->Spa[1],  pArp->Spa[2],  pArp->Spa[3]));
    ZAP(DEBUG_ROUTEIP, DEBUG_INFO, (" Dst IP: "));
    ZAP(DEBUG_ROUTEIP, DEBUG_INFO, ("%d.%d.%d.%d", pArp->Tpa[0], pArp->Tpa[1],  pArp->Tpa[2],  pArp->Tpa[3]));
    ZAP(DEBUG_ROUTEIP, DEBUG_INFO, ("\nARP IPs, UInt32 versions - Src: %x, Dst: %x\n", SrcIP, DstIP));

    // If this is a "gratuitous ARP request", don't reply
    if ((SrcIP == DstIP) || (SrcIP == 0x0000))
    {
        ZAP(DEBUG_ROUTEIP, DEBUG_INFO, ("Gratuitous ARP, no reply generated.\n"));
        goto EthArpReplyExit;
    }

    // Build our ARP reply
    pArp->Op = ARP_OP_REPLY;
    // Swap source and dest IP addresses
    *(UInt32 *)pArp->Spa = DstIP;
    *(UInt32 *)pArp->Tpa = SrcIP;
    // Set the dest hardware address to the original source address
    COPY_MEM(pArp->Sha, pArp->Tha, MAC_SIZE);
    // Set source hardware address to the original dest IP address
    *(UInt32 *)(pArp->Sha+2) = DstIP;
    // Do the same in the ethernet header
    COPY_MEM(pReplyHdr->SrcAddress, pReplyHdr->DstAddress, MAC_SIZE);
    *(UInt32 *)(pReplyHdr->SrcAddress+2) = DstIP;
    
    // The packet is ready ... set the flag and tell OT 
    pHw->FakePacketSize  = PacketSize;
    pHw->FakePacketReady = TRUE;


	{	
		struct sk_buff *skb;
		if (!(skb = dev_alloc_skb(ETH_DATA_LEN+14+2)))
		{
			ZAP(DEBUG_FFD, DEBUG_ERRORS, (kTAB"Error allocating skb!\n"));
			goto EthArpReplyExit;
		}
		skb->dev = pHw->pLinuxNet;
		ZAP(DEBUG_FFD, DEBUG_ENTRY, ("EthernetArpReply: ROUTED : CALLING eth_copy_and_sum\n"));

		eth_copy_and_sum(skb, pHw->pPacket, PacketSize + 2 , 0);

		skb_put(skb, PacketSize + 2);
		skb->protocol = eth_type_trans(skb, pHw->pLinuxNet);
		netif_rx(skb);

	}  

EthArpReplyExit:
    ZAP(DEBUG_ROUTEIP, DEBUG_EXIT, ("---EthernetArpReply\n"));
}

