/**********************************************************************************/
/* MsgProtocol.c								  */
/*										  */
/* Contains the functions that make up the Message Protocol subsystem.		  */
/* This subsystem is responsable for encoding / decoding messages		  */	
/* sending/receiving messages and providing accessors that allow		  */
/* the caller to abort transactions						  */
/*										  */
/* This module reflects the combination of functionality taken			  */	
/* from the buffer.c and mp_int.c files in the Windows NDIS code.		  */
/*										  */
/* This file is part of the "ADI USB ADSL Driver for Linux".			  */	
/*										  */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it      */
/* and/or modify it under the terms of the GNU General Public License as 	  */
/* published by the Free Software Foundation; either version 2 of the License,    */
/* or (at your option) any later version.					  */
/*										  */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be     */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		  */
/* GNU General Public License for more details.					  */
/*										  */
/* You should have received a copy of the GNU General Public License		  */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	  */
/**********************************************************************************/

#include "Adiutil.h"
#include "../AdiUsbAdslDriver.h"

/**********************************************************************************/
/* Local prototypes								  */
/**********************************************************************************/
Boolean MsgCheckSymbolicAddress(char *saddr);
void GetConfigValues(Hardware *pHw);

/**********************************************************************************/
/* Data tables for CMVs								  */
/**********************************************************************************/
#define PREAMBLE                0x535C
#define RECEIVER_MODEM          0x0000
#define SENDER_HOST             0x0010
#define SUBTYPE_MEM_REQWRITE    0x0100
#define SUBTYPE_MEM_REQREAD     0x0000
#define TYPE_MEMACCESS          0x1000

#define WRITE_TO_MODEM (RECEIVER_MODEM + SENDER_HOST + SUBTYPE_MEM_REQWRITE + TYPE_MEMACCESS)
#define READ_FRM_MODEM (RECEIVER_MODEM + SENDER_HOST + SUBTYPE_MEM_REQREAD  + TYPE_MEMACCESS) 

/* for MsgSeq_SoftReset */
CmvMsg MsgCNTL0_RSET = { PREAMBLE, WRITE_TO_MODEM, 0xCDEF, 'C','N','T','L', 0, 0 };

/* for MsgSeq_EnaFR */
CmvMsg MsgOPTN0EnaFR = { PREAMBLE, WRITE_TO_MODEM, 0x1111, 'O','P','T','N', 0, 0x82800044 };

/* for MsgSeq_Retrainer */
CmvMsg MsgCNTL0_RSVD = { PREAMBLE, WRITE_TO_MODEM, 0xEEEE, 'C','N','T','L', 0, 1 };
CmvMsg MsgCNTL0_INIT = { PREAMBLE, WRITE_TO_MODEM, 0xEEEE, 'C','N','T','L', 0, 2 };
CmvMsg MsgINFO8a     = { PREAMBLE, READ_FRM_MODEM, 0xEEEE, 'I','N','F','O', 8, 0 };
CmvMsg MsgINFO10     = { PREAMBLE, READ_FRM_MODEM, 0xEEEE, 'I','N','F','O',10, 0 };
CmvMsg MsgOPTN0      = { PREAMBLE, WRITE_TO_MODEM, 0x1111, 'O','P','T','N', 0, 0x80020066 };
CmvMsg MsgOPTN2      = { PREAMBLE, WRITE_TO_MODEM, 0x2222, 'O','P','T','N', 2, 0 };
CmvMsg MsgOPTN4      = { PREAMBLE, WRITE_TO_MODEM, 0x3333, 'O','P','T','N', 4, 0 };
CmvMsg MsgOPTN7      = { PREAMBLE, WRITE_TO_MODEM, 0x2222, 'O','P','T','N', 7, 0 };
CmvMsg MsgOPTN3      = { PREAMBLE, WRITE_TO_MODEM, 0x2222, 'O','P','T','N', 3, 0 };
CmvMsg MsgOPTN5      = { PREAMBLE, WRITE_TO_MODEM, 0x2222, 'O','P','T','N', 5, 0 };
CmvMsg MsgOPTN6      = { PREAMBLE, WRITE_TO_MODEM, 0x2222, 'O','P','T','N', 6, 0 };
CmvMsg MsgOPTN15     = { PREAMBLE, WRITE_TO_MODEM, 0x2222, 'O','P','T','N',15, 0 };
CmvMsg MsgADPT2      = { PREAMBLE, WRITE_TO_MODEM, 0x1234, 'A','D','P','T', 2, 0 };
CmvMsg MsgINFO9      = { PREAMBLE, WRITE_TO_MODEM, 0x1234, 'I','N','F','O', 9, 0 };
CmvMsg MsgMASK8      = { PREAMBLE, WRITE_TO_MODEM, 0x1234, 'M','A','S','K', 8, 0 };
CmvMsg MsgMASK9      = { PREAMBLE, WRITE_TO_MODEM, 0x1234, 'M','A','S','K', 9, 0 };
CmvMsg MsgPSDM0      = { PREAMBLE, WRITE_TO_MODEM, 0x1234, 'P','S','D','M', 0, 0 };
CmvMsg MsgPFCL1      = { PREAMBLE, WRITE_TO_MODEM, 0x89AB, 'P','F','C','L', 1, 0 };

/* for MsgSeq_OpStat */
CmvMsg MsgSTAT0a     = { PREAMBLE, READ_FRM_MODEM, 0x0123, 'S','T','A','T', 0, 0 };
CmvMsg MsgDIAG2      = { PREAMBLE, READ_FRM_MODEM, 0x1111, 'D','I','A','G', 2, 0 };
CmvMsg MsgDIAG22     = { PREAMBLE, READ_FRM_MODEM, 0x2222, 'D','I','A','G',22, 0 };
CmvMsg MsgDIAG23     = { PREAMBLE, READ_FRM_MODEM, 0x3333, 'D','I','A','G',23, 0 };
CmvMsg MsgDIAG25     = { PREAMBLE, READ_FRM_MODEM, 0x4444, 'D','I','A','G',25, 0 };
CmvMsg MsgDIAG51     = { PREAMBLE, READ_FRM_MODEM, 0x5555, 'D','I','A','G',51, 0 };
CmvMsg MsgDIAG52     = { PREAMBLE, READ_FRM_MODEM, 0x6666, 'D','I','A','G',52, 0 };
CmvMsg MsgDIAG53     = { PREAMBLE, READ_FRM_MODEM, 0x7777, 'D','I','A','G',53, 0 };
CmvMsg MsgDIAG54     = { PREAMBLE, READ_FRM_MODEM, 0x8888, 'D','I','A','G',54, 0 };
CmvMsg MsgRATE0      = { PREAMBLE, READ_FRM_MODEM, 0x9999, 'R','A','T','E', 0, 0 };
CmvMsg MsgDIAG3      = { PREAMBLE, READ_FRM_MODEM, 0xAAAA, 'D','I','A','G', 3, 0 };
CmvMsg MsgDIAG47     = { PREAMBLE, READ_FRM_MODEM, 0xBBBB, 'D','I','A','G',47, 0 };
CmvMsg MsgDIAG49     = { PREAMBLE, READ_FRM_MODEM, 0xCCCC, 'D','I','A','G',49, 0 };
CmvMsg MsgINFOA      = { PREAMBLE, READ_FRM_MODEM, 0xDDDD, 'I','N','F','O',10, 0 };
CmvMsg MsgPSDM1      = { PREAMBLE, READ_FRM_MODEM, 0xEEEE, 'P','S','D','M', 1, 0 };
CmvMsg MsgINFO8b     = { PREAMBLE, READ_FRM_MODEM, 0xEEEE, 'I','N','F','O', 8, 0 };
CmvMsg MsgINFO14     = { PREAMBLE, READ_FRM_MODEM, 0xEEEE, 'I','N','F','O',14, 0 };

/* for MsgSeq_Stat */
CmvMsg MsgSTAT0b     = { PREAMBLE, READ_FRM_MODEM, 0x7654, 'S','T','A','T', 0, 0 };

/* for MSgSeq_ModemEna */
//FLASH_ACC + ENA_MODEMREBOOT_RQ + SENDER_HOST + RECEIVER_MODEM = 0x4110
CmvMsg MsgMODEMENA   = { PREAMBLE, 0x4110, 0xDEAD, 'I','N','F','O', 0, 0 };

UInt32 MsgInitDone = FALSE;

#ifdef LINUX

#define ADDMSG(Seq,Msg) \
    Seq.Msgs[Seq.MsgCount++] = &Msg; \
    if (!MsgInitDone) \
    { \
        z = Msg.SymAddrHiHi; \
        Msg.SymAddrHiHi = Msg.SymAddrHiLo; \
        Msg.SymAddrHiLo = z; \
        z = Msg.SymAddrLoHi; \
        Msg.SymAddrLoHi = Msg.SymAddrLoLo; \
        Msg.SymAddrLoLo = z; \
        Msg.Data = (((Msg.Data & 0xFFFF0000) >> 16) | ((Msg.Data & 0x0000FFFF) << 16)); \
    }

#define ADDMSG_IFVAR_NOTZERO(Seq,Msg,Var) \
    if (Var!=0) \
    { \
        Msg.Data = Var; \
        if (MsgInitDone) \
        { \
          Msg.Data = (((Msg.Data & 0xFFFF0000) >> 16) | ((Msg.Data & 0x0000FFFF) << 16)); \
        } \
        ADDMSG(Seq,Msg); \
    }
#define ADDMSG_FROM_CONFIG(Seq,Msg,Var) \
    Msg.Data = Var; \
    if (MsgInitDone) \
    { \
       Msg.Data = (((Msg.Data & 0xFFFF0000) >> 16) | ((Msg.Data & 0x0000FFFF) << 16)); \
    } \
    ADDMSG(Seq,Msg);
#define ADDMSG_FROM_TEXTFILE(Seq,Msg) \
    if (MsgInitDone) \
    { \
       z = Msg.SymAddrHiHi; \
       Msg.SymAddrHiHi = Msg.SymAddrHiLo; \
       Msg.SymAddrHiLo = z; \
       z = Msg.SymAddrLoHi; \
       Msg.SymAddrLoHi = Msg.SymAddrLoLo; \
       Msg.SymAddrLoLo = z; \
       Msg.Data = (((Msg.Data & 0xFFFF0000) >> 16) | ((Msg.Data & 0x0000FFFF) << 16)); \
    } \
    ADDMSG(Seq,Msg);
#else

#define ADDMSG(Seq,Msg) \
    Seq.Msgs[Seq.MsgCount++] = &Msg; \
    EndianConvert((UInt16 *)&Msg, CMV_DATA_WORDS);

#define ADDMSG_IFVAR_NOTZERO(Seq,Msg,Var) \
    if (Var!=0) \
    { \
        Msg.Data = Var; ****/\
        Seq.Msgs[Seq.MsgCount++] = &Msg; \
        EndianConvert((UInt16 *)&Msg, CMV_DATA_WORDS); \
    }

#define ADDMSG_FROM_CONFIG(Seq,Msg,Var) \
     Msg.Data = Var; \
     Seq.Msgs[Seq.MsgCount++] = &Msg; \
     EndianConvert((UInt16 *)&Msg, CMV_DATA_WORDS);
#define ADDMSG_FROM_TEXTFILE(Seq,Msg) \
     Seq.Msgs[Seq.MsgCount++] = &Msg; \
     EndianConvert((UInt16 *)&Msg, CMV_DATA_WORDS);

#endif /*ifdef LINUX*/

/**********************************************************************************/
/*										  */
/*  NAME: MsgDecode								  */
/*            									  */
/*  ABSTRACT: 									  */
/*      Decodes the input message into the output message			  */
/*										  */
/*  RETURNS: 									  */
/*      Success(true) or Failure(false)						  */
/*              								  */
/*  ASSUMPTIONS: 								  */
/*      Must be called after the message protocol subsystem			  */	
/*      has been intialised.							  */
/*										  */
/**********************************************************************************/
UInt32 MsgDecode(mp_message *Msg, UInt16 MsgRaw[8])
{
    unsigned int ret_code = MP_DECODE_OK;
    unsigned int mp_decode_stage;
    unsigned short temp;

    /* start at beginning of encoding */
    mp_decode_stage = MP_STAGE_PREAMBLE;

    /* encode the message one stage at a time */
    while ( mp_decode_stage < END_DECODE )
    {
       switch ( mp_decode_stage )
       {
          case MP_STAGE_PREAMBLE:
          {
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_FUNCTION_TYPE;
            
             /* extract the preamble */
             Msg->preamble = HSTOLE(MsgRaw[MP_RAW_PREAMBLE]);
            
             /* check the preamble */
             if ( Msg->preamble != MSG_PREAMBLE )
             {
                ret_code |= MP_DECODE_PREAMBLE_ERROR;
                ret_code |= MP_DECODE_ERROR;
             }
          }
             break;
          case MP_STAGE_FUNCTION_TYPE:
          {
             temp = HSTOLE(MsgRaw[MP_RAW_FUNCTION]);
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_FUNCTION_SUBTYPE;
            
             /* extract the type field */
             Msg->type = ( ( MP_FUNCTION_TYPE_MASK & temp ) 
                         >> MP_FUNCTION_TYPE_LSB );
            
             /* check the function subfield */
             switch ( Msg->type )
             {
                case MP_FUNCTION_TYPE_MEMACCESS:
                case MP_FUNCTION_TYPE_FLASHACC:
                case MP_FUNCTION_TYPE_ADSLDIRECTIVE:
                {
                   /* legal function types  */
                }
                   break;
                case MP_FUNCTION_TYPE_MSGDECERR:
                case MP_FUNCTION_TYPE_MSGACCERR:
                case MP_FUNCTION_TYPE_FLASHACCERR:
                {
                   /* legal function types  */
                }
                   break;
                default:
                {
                   /* illegal function type - set return code */
                   ret_code |= MP_DECODE_TYPE_ERROR;
                   ret_code |= MP_DECODE_ERROR;
                }
                   break;
             }
          }
             break;
          case MP_STAGE_FUNCTION_SUBTYPE:
          {
             temp = HSTOLE(MsgRaw[MP_RAW_FUNCTION]);
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_FUNCTION_SENDERID;
            
             /* extract the sub_type field */
             Msg->sub_type = ( ( MP_FUNCTION_SUBTYPE_MASK & temp ) >> MP_FUNCTION_SUBTYPE_LSB );
            
             /* check that the function type was correct, if it */
             /* wasn't, we can't check the sub - type field */
             if ( ( ret_code & MP_DECODE_TYPE_ERROR ) != 0 )
             {
                /* type field was in error, don't check sub - type */
             }
             else
             {
                switch ( Msg->type )
                {
                   case MP_FUNCTION_TYPE_MEMACCESS:
                   {
                      /* check the function subfield */
                      switch ( Msg->sub_type )
                      {
                         case SUBTYPE_MEMACCESS_REPLYREAD:
                         case SUBTYPE_MEMACCESS_REPLYWRITE:
                         {
                            /* legal function subtypes - and into field */
                         }
                            break;
                         default:
                         {
                            /* illegal function type - set return code */
                            ret_code |= MP_DECODE_SUBTYPE_ERROR;
                            ret_code |= MP_DECODE_ERROR;
                         }
                            break;
                      }
                   }
                      break;
                   case MP_FUNCTION_TYPE_FLASHACC:
                   {
                      /* check the function subfield */
                      switch ( Msg->sub_type )
                      {
                         case SUBTYPE_FLASHACC_ENKERNELREBOOTREPLY:
                         case SUBTYPE_FLASHACC_ENMODEMREBOOTREPLY:
                         case SUBTYPE_FLASHACC_DECOMPMODEMAPPLREPLY:
                         case SUBTYPE_FLASHACC_STREAMEDDATAWRREPLY:
                         case SUBTYPE_FLASHACC_STREAMEDDATAWRBLOCKREPLY:
                         case SUBTYPE_FLASHACC_READREPLY:
                         {
                            /* legal function subtypes - and into field */
                         }
                            break;
                         default:
                         {
                            /* illegal function type - set return code */
                            ret_code |= MP_DECODE_SUBTYPE_ERROR;
                            ret_code |= MP_DECODE_ERROR;
                         }
                            break;
                      }
                   }
                      break;
                   case MP_FUNCTION_TYPE_FLASHACCERR:
                   {
                      /* check the function subfield */
                      switch ( Msg->sub_type )
                      {
                         case SUBTYPE_FLASHACCERR_DECOMPRESSERR:
                         case SUBTYPE_FLASHACCERR_READCOMPERR:
                         {
                            /* legal function subtypes - and into field */
                         }
                            break;
                         default:
                         {
                            /* illegal function type - set return code */
                            ret_code |= MP_DECODE_SUBTYPE_ERROR;
                            ret_code |= MP_DECODE_ERROR;
                         }
                            break;
                      }
                   }
                      break;
                   case MP_FUNCTION_TYPE_ADSLDIRECTIVE:
                   {
                      /* check the function subfield */
                      switch ( Msg->sub_type )
                      {
                         case SUBTYPE_ADSLDIRECTIVE_KERNELREADY:
                         case SUBTYPE_ADSLDIRECTIVE_MODEMREADY:
                         case SUBTYPE_ADSLDIRECTIVE_MODEMCRCERROR:
                         {
                            /* legal function subtypes - and into field */
                         }
                            break;
                         default:
                         {
                            /* illegal function type - set return code */
                            ret_code |= MP_DECODE_SUBTYPE_ERROR;
                            ret_code |= MP_DECODE_ERROR;
                         }
                            break;
                      }
                   }
                      break;
                   default:
                   {
                      /* type is invalid and passed last test - can't happen */
                      /* got a serious error here */
                      ZAP(DEBUG_MSG, DEBUG_ERRORS, (kTAB"SW Internal ERROR 'msg->type'\n"));
                   }
                      break;
                }
             }
          }
             break;
          case MP_STAGE_FUNCTION_SENDERID:
          {
             temp = HSTOLE(MsgRaw[MP_RAW_FUNCTION]);
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_FUNCTION_RECEIVERID;
           
             /* extract the sender_id */
             Msg->sender_id = ( ( MP_FUNCTION_SENDERID_MASK & temp ) >> MP_FUNCTION_SENDERID_LSB );
           
             /* check the function subfield */
             switch ( Msg->sender_id )
             {
                case SUBTYPE_ID_ADSL:
                {
                   /* legal function sender_id - and into field */
                }
                   break;
                default:
                {
                   /* illegal function type - set return code */
                   ret_code |= MP_DECODE_SENDERID_ERROR;
                   ret_code |= MP_DECODE_ERROR;
                }
                   break;
             }
          }
             break;
          case MP_STAGE_FUNCTION_RECEIVERID:
          {
             temp = HSTOLE(MsgRaw[MP_RAW_FUNCTION]);
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_INDEX;
           
             /* extract the sender_id */
             Msg->receiver_id = ( ( MP_FUNCTION_RECEIVER_MASK & temp ) >> MP_FUNCTION_RECEIVER_LSB );
            
             /* check the function subfield */
             switch ( Msg->receiver_id )
             {
                case SUBTYPE_ID_ME:
                {
                   /* legal function receiver_id - and into field */
                }
                   break;
                default:
                {
                   /* illegal function type - set return code */
                   ret_code = MP_DECODE_ERROR;
                   /* stop processing the message */
                   mp_decode_stage = END_DECODE;
                }
                   break;
             }
             if ( Msg->type == MP_FUNCTION_TYPE_ADSLDIRECTIVE )
             {
                Msg->index = 0;
                Msg->fourcc = 0;
                Msg->offset_addr = 0;
                Msg->data = 0xDEADBEEF;
                mp_decode_stage = END_DECODE;
             }
          }
             break;
          case MP_STAGE_INDEX:
          {
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_SYMBOLIC_ADDR;
          
             /* insert the index */
             Msg->index = HSTOLE(MsgRaw[MP_RAW_INDEX]);
          }
             break;
          case MP_STAGE_SYMBOLIC_ADDR:
          {
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_OFFSET_ADDR;
        
             if ( Msg->type != MP_FUNCTION_TYPE_ADSLDIRECTIVE )
             {
                /* extract the symbolic address */

                temp = HSTOLE(MsgRaw[MP_RAW_SYMB_ADDR_HIGH]);
                Msg->symb_addr[0] = ( unsigned char )( ( unsigned short )( temp & 0xff00 ) >> 8 );
                Msg->symb_addr[1] = ( unsigned char )( ( unsigned short )( temp & 0x00ff ) );
                temp = HSTOLE(MsgRaw[MP_RAW_SYMB_ADDR_LOW]);
                Msg->symb_addr[2] = ( unsigned char )( ( unsigned short )( temp & 0xff00 ) >> 8 );
                Msg->symb_addr[3] = ( unsigned char )( ( unsigned short )( temp & 0x00ff ) );
           
                if ( MsgCheckSymbolicAddress( (char *)(&Msg->symb_addr[0]) ) == true )
                {
                   /* symbolic address is valid */
                }
                else
                {
                   /* symbolic address invalid - set return code */
                   ret_code |= MP_DECODE_SYMB_ADDR;
                   ret_code |= MP_DECODE_ERROR;
                }
             }
             else
             {
                /* can't check the symbolic address for ADSL */
                /* directive messages */
             }
          }
             break;
          case MP_STAGE_OFFSET_ADDR:
          {
             /* set mp_decode_stage to process next word */
             mp_decode_stage = MP_STAGE_DATA;
           
             /* not going to check this - just insert regardless */
             //Msg->offset_addr = HSTOLE(MsgRaw[MP_RAW_OFFSET_ADDR]);
             Msg->offset_addr = MsgRaw[MP_RAW_OFFSET_ADDR];
          }
             break;
          case MP_STAGE_DATA:
          {
             /* set mp_decode_stage to process next word */
             mp_decode_stage = END_DECODE;
           
             /* can't check this - just insert regardless */
             Msg->data = HSTOLE(MsgRaw[MP_RAW_DATA_HIGH]) * 0x10000 
                                + HSTOLE(MsgRaw[MP_RAW_DATA_LOW]);
          }
             break;
       }
    }

    if ( ret_code != MP_DECODE_OK )
       ZAP(DEBUG_MSG, DEBUG_ERRORS, ("\tMsgDecode error %x.\n",ret_code));

    /* all done */
    return ret_code;
}

/**********************************************************************************/
/*										  */
/*  NAME: MsgSend								  */
/*            									  */
/*  ABSTRACT: 									  */
/*      Sends the supplied message						  */
/*										  */
/*  RETURNS: 									  */
/*      Success(true) or Failure(false)						  */	
/*              								  */
/*  ASSUMPTIONS: 								  */
/*      Must be called after the message protocol subsystem			  */	
/*      has been intialised.							  */
/*										  */
/**********************************************************************************/
void MsgSend(Hardware *pHw, UInt16 *Packet)
{
    
    volatile int  OutboundPending;		
    spin_lock(&pHw->ApiLock);			
    OutboundPending = pHw->AdiModemSm.OutboundPending;						 //	
    if ( (OutboundPending==FALSE) && (pHw->AdiApiRWPending==FALSE) )				 //
      pHw->AdiModemSm.OutboundPending = TRUE;   						 //
    spin_unlock(&pHw->ApiLock);									 //
    

    ZAP(DEBUG_MSG, DEBUG_ENTRY, ("+++MsgSend\n"));

    
    // if no message pending, its safe to send another one 
    if ( (OutboundPending == FALSE ) && (pHw->AdiApiRWPending==FALSE) )
    {
        WriteDspMsg(pHw, kDSP_MP_TX_START, (UInt16 *)Packet);
    }
    else  
    {   											//
       memcpy((char *)pHw->Queue.CmvRaw,(char *)Packet,2*kCMV_DATA_WORDS);			//
       pHw->Queue.Pending = TRUE;								//
       pHw->Queue.Type = 0;									//
       ZAP(DEBUG_MSG, DEBUG_ERRORS, (kTAB"TxPending = TRUE, cannot send new message!\n"));	//
    }												//

    ZAP(DEBUG_MSG, DEBUG_EXIT,  ("---MsgSend\n"));
}


void UserMsgSend(Hardware *pHw, UInt16 *Packet)
{
    
    volatile int  OutboundPending;
    spin_lock(&pHw->ApiLock);
    OutboundPending = pHw->AdiModemSm.OutboundPending;
    if (OutboundPending==FALSE)
      pHw->AdiModemSm.OutboundPending = TRUE;   
    spin_unlock(&pHw->ApiLock);
    

    ZAP(DEBUG_MSG, DEBUG_ENTRY, ("+++MsgSend\n"));
    // if no message pending, its safe to send another one 
    if ( OutboundPending == FALSE )
    {
        pHw->AdiApiRWPending = TRUE;
        WriteDspMsg(pHw, kDSP_MP_TX_START, (UInt16 *)Packet);
        //printk(KERN_EMERG "User Msg Sent %x %x %x %x %x %x\n",Packet[0],Packet[1],Packet[2],Packet[3],Packet[4],Packet[5]);
    }
    else
    {   
       memcpy((char *)pHw->Queue.CmvRaw,(char *)Packet,2*kCMV_DATA_WORDS);
       pHw->Queue.Pending = TRUE;
       pHw->Queue.Type = 1;
       //printk(KERN_EMERG "User Msg Queued\n");
       ZAP(DEBUG_MSG, DEBUG_ERRORS, (kTAB"TxPending = TRUE, cannot send new message!\n"));
    }

    ZAP(DEBUG_MSG, DEBUG_EXIT,  ("---MsgSend\n"));
} // MsgSend


/**********************************************************************************/
/*  NAME: GetConfigValues							  */
/*            									  */	
/*  ABSTRACT: 									  */
/*      Prepares the Options, Adapters and Flags for SM.			  */	
/*      It will set default values, and possibly read some 			  */
/*      (optional) configuration file to fine tune them.			  */
/*										  */
/*  RETURNS: 									  */
/*      void									  */
/*              								  */
/*  ASSUMPTIONS: 								  */
/*										  */
/*      That <pHw> is valid and points to the globals block			  */	
/*										  */
/**********************************************************************************/
void GetConfigValues(Hardware *pHw)
{
    AdiMSM *pSm = &(pHw->AdiModemSm);

    ZAP(DEBUG_MSG, DEBUG_ENTRY, ("+++GetConfigValues\n"));

    /* Initialize defaults*/
    pSm->ADPT0 = 0;
    pSm->ADPT1 = 0;
    pSm->ADPT2 = 0;
    pSm->INFO9 = 0;
    pSm->CNTL0 = 0;
    pSm->OPTN0 = 0;
    pSm->OPTN2 = 0;
    pSm->OPTN4 = 0;
    pSm->OPTN7 = 0;
    pSm->OPTN3 = 0;
    pSm->OPTN5 = 0;
    pSm->OPTN6 = 0;
    pSm->PFCL1 = 0;
    pSm->MASK8 = 0;
    pSm->MASK9 = 0;
    pSm->INFO8 = 0;
    pSm->PSDM0 = 0;
    pSm->FLAG0 = 0;

    if (pHw->pDriverOptions)
    {
       /* Get the option values from the hardware struct*/
       pSm->OPTN0 = pHw->pDriverOptions[ConfigOptn0].Value;
       pSm->OPTN2 = pHw->pDriverOptions[ConfigOptn2].Value;
       pSm->OPTN3 = pHw->pDriverOptions[ConfigOptn3].Value;
       pSm->OPTN4 = pHw->pDriverOptions[ConfigOptn4].Value;
       pSm->OPTN5 = pHw->pDriverOptions[ConfigOptn5].Value;
       pSm->OPTN6 = pHw->pDriverOptions[ConfigOptn6].Value;
       pSm->OPTN7 = pHw->pDriverOptions[ConfigOptn7].Value;
       pSm->OPTN15 = pHw->pDriverOptions[ConfigOptn15].Value;
       pSm->LineType = pHw->pDriverOptions[ConfigLinetype].Value;
    }
    else
    {
       pSm->OPTN0    = 0x80020066;
       pSm->OPTN2    = 0x23700000;
       pSm->OPTN3    = 0x00000000;
       pSm->OPTN4    = 0x00000000;
       pSm->OPTN5    = 0x00000000;
       pSm->OPTN6    = 0x00000000;
       pSm->OPTN7    = 0x02CD8044;
       pSm->OPTN15   = 0x00000000;
       pSm->LineType = 1;
    }
     
    ZAP(DEBUG_MSG, DEBUG_EXIT, ("---GetConfigValues\n"));
}

/**********************************************************************************/
/*  NAME: MsgInitialize								  */
/*            									  */
/*  ABSTRACT: 									  */	
/*      Initializes the Message Protocol Subsystem				  */
/*      Encodes of all the messages that are required for training                */
/*      resetting etc. These encoded messages are stored away in		  */
/*      the global structure, from which they can be sent as desired.		  */
/*										  */
/*  RETURNS: 									  */
/*      Success(true) or Failure(false)						  */
/*              								  */
/*  ASSUMPTIONS: 								  */
/*      Must be called after the various OPTN0.. flags are set -  		  */
/*      this will be done at some stage in the USB init sequence		  */
/*										  */
/*      That <pHw> is valid and points to the globals block			  */
/*										  */
/**********************************************************************************/
void MsgInitialize(Hardware *pHw)
{
    AdiMSM *pSm = &(pHw->AdiModemSm);
    UInt8  z;
    int ii=0;

    ZAP(DEBUG_MSG, DEBUG_ENTRY, ("+++MsgInitialize\n"));

    GetConfigValues(pHw);

    if ((pSm->LineType == 10) && (pHw->pDriverCMVs[0].Data!=0xF9F9F9F9)) /*Read from CMV text file*/
    {
       pSm->MsgSeq_Retrainer.MsgCount = 0;

       while (TRUE)
       {
          if (pHw->pDriverCMVs [ii].Data == 0xF9F9F9F9)
          {
             break;
          }
          ADDMSG_FROM_TEXTFILE( pSm->MsgSeq_Retrainer, pHw->pDriverCMVs[ii]);
          ii+=1;
       }
    } 
    else /*read from configuration file*/
    {
       pSm->MsgSeq_Retrainer.MsgCount = 0;
       ADDMSG(pSm->MsgSeq_Retrainer, MsgCNTL0_RSVD); /*CNTL0 1 - to have enough time sending retrain CMVs*/ 

       ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgINFO8a,pSm->INFO8);
       ADDMSG(pSm->MsgSeq_Retrainer, MsgINFO10);
       if (pSm->OPTN0 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN0, pSm->OPTN0);
       }
       if (pSm->OPTN2 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN2, pSm->OPTN2);
       }
       if (pSm->OPTN4 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN4, pSm->OPTN4);
       }
       if (pSm->OPTN7 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN7, pSm->OPTN7);
       }
       if (pSm->OPTN3 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN3, pSm->OPTN3);
       }
       if (pSm->OPTN5 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN5, pSm->OPTN5);
       }
       if (pSm->OPTN6 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN6, pSm->OPTN6);
       }
       if (pSm->OPTN15 != 0xF9F8F7F6)
       {
          ADDMSG_FROM_CONFIG(pSm->MsgSeq_Retrainer, MsgOPTN15,pSm->OPTN15);
       }
       ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgADPT2, pSm->ADPT2);
       ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgINFO9, pSm->INFO9);
       ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgMASK8, pSm->MASK8);
       ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgMASK9, pSm->MASK9);
       ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgPSDM0, pSm->PSDM0);
       if (pSm->LineType == 0) // G.Lite
          ADDMSG_IFVAR_NOTZERO(pSm->MsgSeq_Retrainer, MsgPFCL1, pSm->PFCL1);

       ADDMSG(pSm->MsgSeq_Retrainer, MsgCNTL0_INIT);
       
    }

    /*Get our SoftReset message sequence initialized - sequence taken from NDIS code*/
    pSm->MsgSeq_SoftReset.MsgCount = 0;
    ADDMSG( pSm->MsgSeq_SoftReset, MsgCNTL0_RSET);

    /*Get our Stat message sequence initialized - sequence taken from NDIS code*/
    pSm->MsgSeq_Stat.MsgCount = 0;
    ADDMSG( pSm->MsgSeq_Stat, MsgSTAT0b);

    /*Get our OpStat message sequence initialized - sequence taken from NDIS code*/
    pSm->MsgSeq_OpStat.MsgCount = 0;
    ADDMSG( pSm->MsgSeq_OpStat, MsgSTAT0a);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG2);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG22);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG23);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG25);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG51);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG52);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG53);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG54);
    ADDMSG( pSm->MsgSeq_OpStat, MsgRATE0);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG3);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG47);
    ADDMSG( pSm->MsgSeq_OpStat, MsgDIAG49);
    ADDMSG( pSm->MsgSeq_OpStat, MsgINFOA);
    ADDMSG( pSm->MsgSeq_OpStat, MsgPSDM1);
    ADDMSG( pSm->MsgSeq_OpStat, MsgINFO8b);
    ADDMSG( pSm->MsgSeq_OpStat, MsgINFO14);
    /*add easy stat at the end of the chain - it is already endianized - do not use macro!*/
    pSm->MsgSeq_OpStat.Msgs[pSm->MsgSeq_OpStat.MsgCount++] = &MsgSTAT0b;

    /*Get our ModemEna message sequence initialized - CMV taken from NDIS code*/
    pSm->MsgSeq_ModemEna.MsgCount = 0;
    ADDMSG( pSm->MsgSeq_ModemEna, MsgMODEMENA);

    /*Set Enable Fast Retrain options - CMV not tested! [IIDos]*/
    pSm->MsgSeq_EnaFR.MsgCount = 0;
    ADDMSG( pSm->MsgSeq_EnaFR, MsgOPTN0EnaFR);

    /*Init our message pending flags*/
    pSm->InboundAsyncPending = FALSE;
    pSm->InboundSyncPending  = FALSE;
    pSm->OutboundPending     = FALSE;

    MsgInitDone = TRUE;
    
    ZAP(DEBUG_MSG, DEBUG_EXIT,  ("---MsgInitialize\n"));
}

static char symb_addresses[NUM_CMVS][4] =
{
    {'A', 'D', 'P', 'T'},
    {'C', 'N', 'T', 'L'},
    {'C', 'O', 'D', 'E'},
    {'D', 'I', 'A', 'G'},
    {'D', 'O', 'P', 'T'},
    {'F', 'L', 'A', 'G'},
    {'I', 'N', 'F', 'O'},
    {'I', 'N', 'T', 'L'},
    {'O', 'P', 'T', 'N'},
    {'P', 'F', 'C', 'L'},
    {'P', 'F', 'R', 'X'},
    {'P', 'F', 'T', 'X'},
    {'P', 'S', 'D', 'M'},
    {'R', 'A', 'T', 'E'},
    {'R', 'X', 'D', 'A'},
    {'S', 'T', 'A', 'T'},
    {'T', 'E', 'S', 'T'},
    {'T', 'O', 'N', 'E'},
    {'T', 'X', 'D', 'A'},
    {'U', 'O', 'P', 'T'}
};

#define SEARCHING   0x0
#define PASSED      0x1
#define FAILED      0x2

/**********************************************************************************/
/*										  */	
/*  NAME: MsgCheckSymbolicAddress						  */
/*            									  */	
/*  ABSTRACT: 									  */	
/*      Looks up a symbolic address						  */	
/*										  */
/*  RETURNS: 									  */
/*      Success(true) or Failure(false)						  */	
/*              								  */	
/*  ASSUMPTIONS: 								  */
/*      None.									  */
/*										  */
/**********************************************************************************/
Boolean MsgCheckSymbolicAddress(char *saddr)
{
    Boolean ret_code = false;
    unsigned int search_code = SEARCHING;
    unsigned int cmv = 0;
    unsigned int letter;

    /* search the reference array for a match */
    while ( ( cmv < NUM_CMVS ) && ( search_code == SEARCHING ) )
    {
       /* try to match the CMV to the current reference array CMV */
       letter = 0;
       while ( letter < 4 && ( saddr[letter] == symb_addresses[cmv][letter] ) )
       {
          letter++;
       }

       switch ( letter )
       {
          case 0:
          {
             /* no match on first char - check that first char of CMV being */
             /* checked is not before the first char of this entry in */
             /* the alphabet, if it is we can stop looking as entries in the */
             /* reference array are in ascending order */
             if (saddr[0] < symb_addresses[cmv][0])
             {
                /* give up looking, already too far into alphabet */
                search_code = FAILED;
                ret_code = false;
             }
             else
             {
                /* not too far yet, keep looking */
             }
          }
             break;
          case 4:
          {
             /* found a match */
             search_code = PASSED;
             ret_code = true;
          }
             break;
          default:
          {
             /* partial match, keep looking */
          }
             break;
       }
       cmv++;
    }

    ret_code = true;	//FFD090903_1 ( we do not check the cmv names )
    return (ret_code);
}

