/**********************************************************************************/
/*										  */ 
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved  		  */	
/* Oam.c 									  */
/*										  */
/* This file is part of the "ADI USB ADSL Driver for Linux".			  */	
/*										  */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it      */
/* and/or modify it under the terms of the GNU General Public License as 	  */
/* published by the Free Software Foundation; either version 2 of the License,    */
/* or (at your option) any later version.					  */
/*										  */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be     */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		  */
/* GNU General Public License for more details.					  */
/*										  */
/* You should have received a copy of the GNU General Public License		  */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	  */
/**********************************************************************************/


//---------------------------------------------------------------------------
// Operations And Maintenance ATM support module
//
// Some abbreviations ...
//
//  CC  = Continuity Check
// AIS  = Alarm Indication Signal
// RDI  = Remote Defect Indication 
//
//---------------------------------------------------------------------------

#include "Adiutil.h"

Boolean CheckForAndProcessOAMCell(Hardware *pHw, UInt8 *pCell);
static Boolean CheckForAndProcessOAMCCChange(Hardware *pHw, UInt8 *pCell);
static Boolean CheckForAndProcessOAMResponse(Hardware *pHw, UInt8 *pCell, UInt32 CellType);
static UInt16 ByteCalcCRC10(UInt8 *pData, UInt32 Size);
static void USBSendAmoeba(Hardware *pHw, UInt8 *pCell, Boolean MoreThanOneCell);
static void AmoebaCompletion(struct urb *urb);
static void ArmOAMTimer(Hardware *pHw);
int SetOAMTimerInterval(Hardware* pHw, int reqDelay);
void StopOAMTimer(Hardware *pHw);
void WriteToUSB (void);

Boolean CheckForAndProcessOAMCell(Hardware *pHw, UInt8 *pCell)
{
    UInt32 OAMType = OAM_NONE;
    UInt32 PTI, VPI, VCI, VpiVci,CellHeader;

    ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++CheckForAndProcessOAMCell\n"));

	CellHeader = BETOHL(*((UInt32 *)pCell)); 
	VpiVci     = (CellHeader >> 4) & ATM_CELL_REMOVE_GFC_MASK; 

    PTI = (pCell[ATM_CELL_HEADER_PT_OFFSET] & 0x0E) >> 1;
    VPI = (VpiVci & 0x00FF0000) >> 16;
    VCI = (VpiVci & 0x0000FFFF);

    ZAP(DEBUG_OAM, DEBUG_INFO, ("OAM cell: VPI=%#X, VCI=%#X, PTI=%#X\n", VPI, VCI, PTI));        

    if ( (PTI == 0x04) &&
         (VCI != 0x00) && (VCI != 0x03) &&
         (VCI != 0x04) && (VCI != 0x06) && (VCI != 0x07) )
    {
        ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"OAM F5 segment cell received.\n"));
        OAMType = OAM_F5_SEGMENT;
    }

    // Check for F5 end-to-end cell
    if ( (PTI == 0x05) &&
         (VCI != 0x00) && (VCI != 0x03) &&
         (VCI != 0x04) && (VCI != 0x06) && (VCI != 0x07) )
    {
        ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"OAM F5 end-to-end cell received.\n"));
        OAMType = OAM_F5_ENDTOEND;
    }

    // Check for F4 segment cell
    if (VCI == 0x03)
    {
        ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"OAM F4 segment cell received.\n"));
        OAMType = OAM_F4_SEGMENT;
    }

    // And finally, check for F4 end-to-end cell
    if (VCI == 0x04)
    {
        ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"OAM F4 end-to-end cell received.\n"));
        OAMType = OAM_F4_ENDTOEND;
    }



    // If this is not an OAM cell, we're done
    if (OAMType == OAM_NONE)
        goto OAM_Done;

    ZAP(DEBUG_OAM, DEBUG_INFO, ("OAM cell: VPI=%#X, VCI=%#X, PTI=%#X\n", VPI, VCI, PTI));

    // Check for and process CC activate and deactivate
    if (!CheckForAndProcessOAMCCChange(pHw, pCell))
    {
        // This is not a CC activate or deactivate cell, see if this is 
        // an OAM cell that needs a response (loopback, AIS, RDI)
        if (!CheckForAndProcessOAMResponse(pHw, pCell, OAMType))
        {
			// ????
        }
    }

OAM_Done:
    ZAP(DEBUG_OAM, DEBUG_EXIT, ("---CheckForAndProcessOAMCell\n"));
    memcpy(pHw->pOAMCell,pCell,ATM_CELL_SIZE);    
    memcpy(pHw->OAMCellHeader,pCell,ATM_CELL_HEADER_SIZE);

    if (OAMType == OAM_NONE)
        return FALSE;
    else
        return TRUE;
}


//---------------------------------------------------------------------------
// CheckForAndProcessOAMCCChange
//---------------------------------------------------------------------------

static Boolean CheckForAndProcessOAMCCChange(Hardware *pHw, UInt8 *pCell)
{
    Boolean Retval = FALSE;
    ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++CheckForAndProcessOAMCCChange\n"));

    if (pCell[OAM_TYPE_AND_FN] == OAM_ACTIVATION_CC)
    {
        Retval = TRUE;

        switch (pCell[OAM_MSGID_AND_DIR] & OAM_MASK_MSGID)
        {
            // As far as I know, we should only receive requests. The
            // confirm and denied messages only go to those endpoints
            // who actually originate CC activation changes - we don't.
            case OAM_ACTIVATE_REQUEST:
            {
                UInt8 OAM_Dir;

                ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"OAM CC activate request\n"));

                OAM_Dir = pCell[OAM_MSGID_AND_DIR] & OAM_MASK_DIR;
                if (OAM_Dir & OAM_MASK_DIR_AB)
                {
                    ZAP(DEBUG_OAM, DEBUG_INFO, ("AB direction.\n"));
                    pHw->OAMState_SendingCC = TRUE;

                }

                if (OAM_Dir & OAM_MASK_DIR_BA)
                {
                    ZAP(DEBUG_OAM, DEBUG_INFO, ("BA direction.\n"));
                    pHw->OAMState_ReceiveCC = TRUE;
                }

                // If OAM_Dir is non-zero, then it is valid
                if (OAM_Dir)
                {
                    ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"Preparing and sending OAM CC activate confirm.\n"));
                    // We have no real use for this cell after we're done
                    // here, so we'll just modify it in place and use it to send
                    // back the activate confirmation. Much of the cell format is
                    // the same as the request cell, so we don't have to change much.
                    pCell[OAM_MSGID_AND_DIR] = OAM_Dir | OAM_ACTIVATE_CONFIRMED;
                    
                    // Store a copy of the cell's header for later use
                    COPY_MEM(pCell, pHw->OAMCellHeader, ATM_CELL_HEADER_SIZE);
                    
                    CALC_AND_STORE_CRC10(pCell);

					USBSendAmoeba(pHw, pCell, FALSE);

                    // We need to be timing the time between receives or sends
                    // of CC cells, so start the timer if its not already ticking.
                    if (pHw->OAMState_TimerOn == FALSE)
                    {
                        ArmOAMTimer(pHw);
                    }
                }
                else
                {
                    ZAP(DEBUG_OAM, DEBUG_ERRORS, (kTAB"CC activation with no direction.\n"));
                }
            } break;

            case OAM_DEACTIVATE_REQUEST:
            {
                UInt8 OAM_Dir;

                ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"OAM CC deactivate request.\n"));

                OAM_Dir = pCell[OAM_MSGID_AND_DIR] & OAM_MASK_DIR;
                if (OAM_Dir & OAM_MASK_DIR_AB)
                {
                    ZAP(DEBUG_OAM, DEBUG_INFO, ("AB direction.\n"));
                    pHw->OAMState_SendingCC = FALSE;
                }

                if (OAM_Dir & OAM_MASK_DIR_BA)
                {
                    ZAP(DEBUG_OAM, DEBUG_INFO, ("BA direction.\n"));
                    pHw->OAMState_ReceiveCC = FALSE;
                    // When BA direction is deactivated, AIS state needs cleared
                    pHw->OAMState_AIS       = FALSE;
                    // Reset our received CC cell counter
                    pHw->OAMState_CCSink    = 0;
                }

                if (OAM_Dir)
                {
                    ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"Preparing and sending OAM CC deactivate confirm.\n"));

                    // If sending and receiving are now deactivated and the timer
                    // is currently running, shut it off.
                    if ((pHw->OAMState_TimerOn   == TRUE) &&
                        (pHw->OAMState_SendingCC == FALSE) &&
                        (pHw->OAMState_ReceiveCC == FALSE))
                    {
                        StopOAMTimer(pHw);
                    }

                    // We have no real use for this cell after we're done
                    // here, so we'll just modify it in place and use it to send
                    // back the activate confirmation. Much of the cell format is
                    // the same as the request cell, so we don't have to change much.
                    pCell[OAM_MSGID_AND_DIR] = OAM_Dir | OAM_DEACTIVATE_CONFIRMED;
                    
                    CALC_AND_STORE_CRC10(pCell);

					USBSendAmoeba(pHw, pCell, FALSE);
                }
                else
                {
                    ZAP(DEBUG_OAM, DEBUG_ERRORS, (kTAB"CC deactivation with no direction.\n"));
                }

            } break;

            case OAM_ACTIVATE_CONFIRMED:
            case OAM_ACTIVATE_REQUEST_DENIED:
            case OAM_DEACTIVATE_CONFIRMED:
            case OAM_DEACTIVATE_REQUEST_DENIED:
            {
                ZAP(DEBUG_OAM, DEBUG_ATTN, (kTAB"Warning - OAM activation change response received.\n"));
                goto CCChange_Done;
            } break;
        }
    }

CCChange_Done: 	
    ZAP(DEBUG_OAM, DEBUG_EXIT,  ("---CheckForAndProcessOAMCCChange\n"));
    return Retval;
}


//---------------------------------------------------------------------------
// CheckForAndProcessOAMResponse
//
// Loopback, AIS, and RDI OAM F5 cells need to have a response sent back.
//---------------------------------------------------------------------------

static Boolean CheckForAndProcessOAMResponse(Hardware *pHw, UInt8 *pCell, UInt32 CellType)
{
    Boolean Retval = FALSE;
    ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++CheckForAndProcessOAMResponse\n"));


    switch (pCell[OAM_TYPE_AND_FN])
    {
        case OAM_LOOPBACK:
        {
            if (pCell[OAM_LOOPBACK_INDICATOR] & OAM_LOOPBACK_REPLY_MASK)
            {
                ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"Loopback reply request received.\n"));
                
                // Mark this cell as a loopback response
                pCell[OAM_LOOPBACK_INDICATOR] = 0;

                // Tell lower code to send response cell
                Retval = TRUE;
            }
            else
            {
                // We should never get here, 'cause we have no current mechanism
                // for originating a loopback reply request.
                ZAP(DEBUG_OAM, DEBUG_ATTN, (kTAB"Loopback reply received.\n"));
            }
        } break;

        case OAM_AIS:
        {
            ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"AIS received.\n"));

            // Per ADI Windows code, ASI gets an RDI response
            pCell[OAM_TYPE_AND_FN] = OAM_RDI;

            // Tell lower code to send response cell
            Retval = TRUE;
        } break;
        
        case OAM_RDI:
        {
            // Per ADI Windows code, RDI gets a CC response
            pCell[OAM_TYPE_AND_FN] = OAM_CC;
        
            // Tell lower code to send response cell
            Retval = TRUE;
        } break;
        
        case OAM_CC:
        {
            ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"CC received.\n"));
			if ( pHw->OAMState_ReceiveCC == TRUE ) 
			{
			            pHw->OAMState_CCSink = 0;
			}
			else
			{
			            // Per ADI Windows code, CC gets an CC response
			            pCell[OAM_TYPE_AND_FN] = OAM_CC;

			            // Tell lower code to send response cell
			            Retval = TRUE;
			}
			
        } break;

        default:
        {
        } break;
		
    }

    // This will be TRUE if a cell needs to be sent
    if (Retval == TRUE)
    {
        CALC_AND_STORE_CRC10(pCell);
		USBSendAmoeba(pHw, pCell, FALSE);

    }
    
// Response_Done:
    ZAP(DEBUG_OAM, DEBUG_EXIT,  ("---CheckForAndProcessOAMResponse\n"));
    return Retval;
}


#define POLYNOMIAL 0x633

static UInt16 ByteCalcCRC10(UInt8 *pData, UInt32 Size)
{
    UInt32 i;
    UInt16 crc10_accum = 0;
    static Boolean AlreadyDone = FALSE;
    static UInt16 byte_crc10_table[256];

    // Generate byte values for use, needs to be done once
    if (!AlreadyDone) 
    {
        UInt32 k, j;
        UInt16 Gen_crc10_accum;

        for (k = 0;  k < 256;  k++)
        {
            Gen_crc10_accum = ((UInt16) k << 2);
            for (j = 0;  j < 8;  j++)
            {
                if ((Gen_crc10_accum <<= 1) & 0x400) 
                    Gen_crc10_accum ^= POLYNOMIAL;
            }
            byte_crc10_table[k] = Gen_crc10_accum;
        }   

        AlreadyDone = TRUE;   
    }   
    
    for (i = 0;  i < Size;  i++)
    {
        crc10_accum = ((crc10_accum << 8) & 0x3ff) ^ byte_crc10_table[( crc10_accum >> 2) & 0xff]
                                                   ^ *pData++;
    }

    return crc10_accum;
}

static void USBSendAmoeba(Hardware *pHw, UInt8 *pCell, Boolean MoreThanOneCell)
{

    int result;
    int cntr;	
    UInt32         uiTotalBytes = 0;
    UInt8 MyCell[ATM_CELL_SIZE*2];

    ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++USBSendAmoeba\n"));

   if ( MoreThanOneCell == TRUE )
	uiTotalBytes = ATM_CELL_SIZE*2;
   else
	uiTotalBytes = ATM_CELL_SIZE;

    memcpy(MyCell,pCell,uiTotalBytes);	

	
	for (cntr=0;cntr<uiTotalBytes;cntr++)
		ZAP(DEBUG_OAM, DEBUG_INFO, (" %#x ", MyCell[cntr]));
    ZAP(DEBUG_OAM, DEBUG_INFO,
        ("\n *** OAM **** Writing %#x bytes to USB\n", uiTotalBytes));
	

    FILL_BULK_URB(&pHw->urbOAMWrite, pHw->usbdev, pHw->pipeBulkDataOut,
                  pCell,uiTotalBytes , AmoebaCompletion, pHw);

    result = usb_submit_urb(&pHw->urbOAMWrite);

    ZAP(DEBUG_OAM, DEBUG_EXIT, ("---USBSendAmoeba { usb_submit_urb result=%d }\n",result));

}

//---------------------------------------------------------------------------
// AmoebaCompletion
//---------------------------------------------------------------------------

static void AmoebaCompletion(struct urb *urb)
{

    Hardware *pHw = urb->context;
//    struct net_device *ether;


    ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++ AmoebaCompletion\n"));

    if ((pHw == NULL) || (!pHw->IsOpen))
    {
       ZAP(DEBUG_OAM, DEBUG_ERRORS,
           (kTAB"NULL urb->context or IsOpen=FALSE in WriteCompletion!\n"));
       goto WriteCompletion_exit;
    }

WriteCompletion_exit:
    ZAP(DEBUG_OAM, DEBUG_EXIT, ("--- AmoebaCompletion\n"));

}


//---------------------------------------------------------------------------
// ArmOAMTimer
//---------------------------------------------------------------------------

static void ArmOAMTimer(Hardware *pHw)
{
   ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++ArmOAMTimer\n"));
   pHw->OAMState_TimerOn = TRUE;
   SetOAMTimerInterval(pHw,OAM_TIMER_VALUE);
   ZAP(DEBUG_OAM, DEBUG_EXIT, ("---ArmOAMTimer\n"));
}

//---------------------------------------------------------------------------
// StopOAMTimer
//---------------------------------------------------------------------------

void StopOAMTimer(Hardware *pHw)
{

    ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++StopOAMTimer\n"));

    pHw->OAMState_TimerOn = FALSE;

    del_timer(&pHw->OAMTimer);

    ZAP(DEBUG_OAM, DEBUG_EXIT, ("---StopOAMTimer\n"));
}

//---------------------------------------------------------------------------
// OAMTimer
//
// Called every 1 second.
//---------------------------------------------------------------------------

void OAMTimerFunction(unsigned long ptr)
{
    Hardware *pHw = (Hardware *)ptr;

    UInt8 *pCell;
    
	ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++ OAMTimerFunction\n"));


    if (!(pHw->AdiModemSm.CurrentAdiState & STATE_OPERATIONAL))
    {
        ZAP(DEBUG_OAM, DEBUG_ATTN, (kTAB" *** Called while not at OPER state.\n"));
        goto OAMTimer_Done;
    }

    //pCell = (UInt8 *)&pHw->OAMCell;
   pCell = (UInt8 *)pHw->pOAMCell;


    // If CC sending is enabled, then we need to send a CC
    if (pHw->OAMState_SendingCC == TRUE)
    {
        ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"====>>>>> Sending CC...\n"));

	
        COPY_MEM(pHw->OAMCellHeader, pCell, ATM_CELL_HEADER_SIZE);

        pCell[OAM_TYPE_AND_FN] = OAM_CC;

        CALC_AND_STORE_CRC10(pCell);

		if (pHw->OAMState_ReceiveCC == FALSE)
			USBSendAmoeba(pHw, pCell, FALSE);
    }

    // If CC receive is enabled, inc counter and send RDI if necessary
    if (pHw->OAMState_ReceiveCC == TRUE)
    {
        if (++pHw->OAMState_CCSink > 4)
        {
            ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"Sending RDI...\n"));
            // Turn on AIS state
            pHw->OAMState_AIS = TRUE;
            
            // Keep our sink timer counter from rolling over
            pHw->OAMState_CCSink = 5;
    
			if (pHw->OAMState_SendingCC == TRUE)
			{				            
			            COPY_MEM(pHw->OAMCellHeader, (pCell + ATM_CELL_SIZE), ATM_CELL_HEADER_SIZE);
			    
			            pCell [ATM_CELL_SIZE + OAM_TYPE_AND_FN] = OAM_RDI;
			    
			            CALC_AND_STORE_CRC10((pCell + ATM_CELL_SIZE));
						USBSendAmoeba(pHw, pCell, TRUE);

			}
			else
			{
			            COPY_MEM(pHw->OAMCellHeader, pCell, ATM_CELL_HEADER_SIZE);
			    
			            pCell[OAM_TYPE_AND_FN] = OAM_RDI;
			    
			            CALC_AND_STORE_CRC10(pCell);

						USBSendAmoeba(pHw, pCell, FALSE);
			}
			//////////////////
        }
        else
        {
			if (pHw->OAMState_SendingCC == TRUE)
			{
				USBSendAmoeba(pHw, pCell, FALSE);
			}
        }
		////////////////
    }
	else
	{
	}
	
OAMTimer_Done:
    // Need to reschedule the timer to go again
    ArmOAMTimer(pHw);

    ZAP(DEBUG_OAM, DEBUG_EXIT, ("--- OAMTimerFunction\n"));
}


int SetOAMTimerInterval(Hardware* pHw, int reqDelay)
{
     ZAP(DEBUG_OAM, DEBUG_ENTRY, ("+++SetOAMTimerInterval(%d)\n", reqDelay));

    /*If the timer is currently running, delete it*/
    if ( timer_pending(&pHw->OAMTimer) )
       del_timer(&pHw->OAMTimer);

	pHw->OAMTimer.expires = jiffies + MSEC_TO_JIFFIES(reqDelay); 


    add_timer(&pHw->OAMTimer);

    ZAP(DEBUG_OAM, DEBUG_EXIT, ("---SetOAMTimerInterval\n"));
    return USB_ST_NOERROR;
}

