/***************************************************************************************/
/*										       */	
/* Copyright (c) 2001, Analog Devices Inc., All Rights Reserved			       */	
/*										       */
/* Sar.c									       */	
/*										       */
/* AAL5 Segmentation and Reassembly support routines				       */	
/*										       */
/* This file is part of the "ADI USB ADSL Driver for Linux".			       */	
/*										       */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it 	       */
/* and/or modify it under the terms of the GNU General Public License as 	       */
/* published by the Free Software Foundation; either version 2 of the License, 	       */ 	
/* or (at your option) any later version.					       */
/*										       */	
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be 	       */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	       */	
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		       */
/* GNU General Public License for more details.					       */	
/*										       */
/* You should have received a copy of the GNU General Public License		       */	
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software      */	
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	       */	
/***************************************************************************************/

#include "Adiutil.h"

/***************************************************************************************/
/* Aal5InitSegmenter								       */	
/*										       */	
/* Initializes the specified Sar segmenting state machine, preparing for segmentation  */
/*										       */	
/* Notes:									       */
/*       If the PDU is encapsulated, the total encapsulated length should be	       */
/*       passed in as the PduLength						       */
/***************************************************************************************/
void Aal5InitSegmenter( SarSegmenter *pSar, 
                        UInt32        PduLength, 
                        UInt8        *pOutputBuffer, 
                        UInt8        *pCellHeader )
{
    ZAP(DEBUG_SAR, DEBUG_ENTRY, ("+++Aal5InitSar\n"));

    pSar->pOutputBuffer          = pOutputBuffer;
    pSar->pCellHeader            = pCellHeader;
    pSar->RawPduLength           = PduLength;
    pSar->PaddingLength          = 0;
    pSar->BytesLeftInCurrentCell = 0;
    pSar->CellCount              = 0;
    pSar->RunningCRC             = CRC_INITIAL_VALUE;

    ZAP(DEBUG_SAR, DEBUG_EXIT, ("---Aal5InitSar\n"));
}

/***************************************************************************************/
/* Aal5Segment									       */
/*										       */	
/* Performs the actual segmentation part of AAL5. Basically just splits the input      */
/* data into ATM_CELL_PAYLOAD_SIZE sized packets, putting an ATM cell header at	       */
/* the beginning of each packet, to form an ATM_CELL_SIZE sized packet.		       */
/*										       */
/* Notes:									       */	
/*       Unlike the NDIS code, this does NOT finish the CPCS PDU with		       */
/*       padding and trailer bytes - Aal5WritePadAndTrailer will do this	       */	
/***************************************************************************************/
void Aal5Segment( SarSegmenter *pSar, 
                  UInt8        *pInputBuffer, 
                  UInt32        InputLength )
{
    UInt16 BytesToCopy;

    ZAP(DEBUG_SAR, DEBUG_ENTRY, ("+++Aal5Segment\n"));

    /*Lets start by first calculating the CRC for this input buffer*/
    pSar->RunningCRC = CrcCalculate(pInputBuffer, InputLength, pSar->RunningCRC);

    /*Loop until we exhaust the bytes in the input buffer*/
    while (InputLength)
    {
       if (0 == pSar->BytesLeftInCurrentCell)
       {
          /*This is the start of a new ATM cell, so copy the default cell header*/
          ZAP(DEBUG_SAR, DEBUG_MEMCOPY, (kTAB"src=%p, dst=%p, len=%x\n", pSar->pCellHeader, 
                                         pSar->pOutputBuffer, ATM_CELL_HEADER_SIZE));
          COPY_MEM(pSar->pCellHeader, pSar->pOutputBuffer, ATM_CELL_HEADER_SIZE);
          pSar->pOutputBuffer += ATM_CELL_HEADER_SIZE;
          /*Now that we've actually started the cell, set the bytes left accordingly*/
          pSar->BytesLeftInCurrentCell = ATM_CELL_PAYLOAD_SIZE;
          pSar->CellCount++;
       }

       /*We have two possibilities, either the data left in the input buffer will */
       /*be exhausted within the current cell, or it will require additional cells*/
       BytesToCopy = ADI_Min(pSar->BytesLeftInCurrentCell, InputLength);
       ZAP(DEBUG_SAR, DEBUG_MEMCOPY, (kTAB"src=%p, dst=%p, len=%x\n", pInputBuffer, 
                                      pSar->pOutputBuffer, BytesToCopy));
       COPY_MEM(pInputBuffer, pSar->pOutputBuffer, BytesToCopy);
       pSar->BytesLeftInCurrentCell -= BytesToCopy;
       pSar->pOutputBuffer          += BytesToCopy;
       pInputBuffer                 += BytesToCopy;
       InputLength                  -= BytesToCopy;
   }

   ZAP(DEBUG_SAR, DEBUG_EXIT, ("---Aal5Segment\n"));
}

/***************************************************************************************/
/* Aal5WritePadAndTrailer							       */	
/*										       */
/* Performs cell padding when necessary and writes the CPCS PDU trailer		       */	 	
/*										       */
/* Returns the total count of ATM cells created for the Pdu			       */	
/***************************************************************************************/
UInt16 Aal5WritePadAndTrailer( SarSegmenter *pSar, 
                               UInt8         CpcsUu, 
                               UInt8         Cpi )
{
    UInt16 PaddingSize;
    UInt8 *pCell;
    Aal5Trailer *pTrailer;

    ZAP(DEBUG_SAR, DEBUG_ENTRY, ("+++Aal5WritePadAndTrailer\n"));

    /* If we don't have room for the trailer in the current packet, then    */
    /* we need to pad out the rest of the current packet and start a new one*/
    if (pSar->BytesLeftInCurrentCell < ATM_AAL5_TRAILER_SIZE)
    {
       /* We might have 0 bytes left in the current cell, which would mean we*/
       /* still need to start a new cell for the trailer, but we don't need  */
       /* to pad the current cell					     */	
       if (pSar->BytesLeftInCurrentCell)
       {
          ZERO_MEM(pSar->pOutputBuffer, pSar->BytesLeftInCurrentCell);
          /* Account for the padding in the CRC*/
          pSar->RunningCRC     = CrcCalculate(pSar->pOutputBuffer, 
                                              pSar->BytesLeftInCurrentCell, 
                                              pSar->RunningCRC);
          pSar->pOutputBuffer += pSar->BytesLeftInCurrentCell;
       }

       /* Start the new cell that will contain padding+trailer*/
       ZAP(DEBUG_SAR, DEBUG_MEMCOPY, (kTAB"src=%p, dst=%p, len=%x\n", pSar->pCellHeader, 
                                      pSar->pOutputBuffer, ATM_CELL_HEADER_SIZE));
       COPY_MEM(pSar->pCellHeader, pSar->pOutputBuffer, ATM_CELL_HEADER_SIZE);
       pSar->pOutputBuffer         += ATM_CELL_HEADER_SIZE;
       pSar->BytesLeftInCurrentCell = ATM_CELL_PAYLOAD_SIZE;
       pSar->CellCount++;
    }

    /* We're now ready with the cell that will contain trailer and possible padding*/
    PaddingSize = pSar->BytesLeftInCurrentCell - ATM_AAL5_TRAILER_SIZE;
    if (PaddingSize)
    {
       ZERO_MEM(pSar->pOutputBuffer, PaddingSize);
       /* Account for the padding in the CRC*/
       pSar->RunningCRC              = CrcCalculate(pSar->pOutputBuffer, 
                                                    PaddingSize, 
                                                    pSar->RunningCRC);
       pSar->BytesLeftInCurrentCell -= PaddingSize;
       pSar->pOutputBuffer          += PaddingSize;
    }

    /* Set the "AAL_Indicate" bit in the Payload Type field of the current ATM cell*/
    pCell = pSar->pOutputBuffer - (ATM_CELL_SIZE-pSar->BytesLeftInCurrentCell);
    pCell[ATM_CELL_HEADER_PT_OFFSET] |= ATM_PT_AAL_INDICATE;

    /* Finally, write out the CPCS PDU trailer*/
    pTrailer            = (Aal5Trailer *)(pSar->pOutputBuffer);
    pTrailer->CpcsUu    = CpcsUu;
    pTrailer->Cpi       = Cpi;
    pTrailer->PduLength = HSTOBE(pSar->RawPduLength);
    /* Finish the CRC by calculating the CRC for the trailer*/
    pSar->RunningCRC              = CrcCalculate(pSar->pOutputBuffer, 
                                                 ATM_AAL5_TRAILER_SIZE-ATM_AAL5_CRC_SIZE, 
                                                 pSar->RunningCRC);
    pTrailer->Crc                 = HLTOBE(~pSar->RunningCRC);
    pSar->pOutputBuffer          += ATM_AAL5_TRAILER_SIZE;
    pSar->BytesLeftInCurrentCell -= ATM_AAL5_TRAILER_SIZE;
    /* Actually, BytesLeftInCurrentCell had better equal 0!*/
    ZAPIF((pSar->BytesLeftInCurrentCell), DEBUG_SAR, DEBUG_ATTN,
          (kTAB"Aal5WritePadAndTrailer finished with BytesLeftInCurrentCell != 0!\n"));

    ZAP(DEBUG_SAR, DEBUG_EXIT, ("---Aal5WritePadAndTrailer\n"));

    return pSar->CellCount;
}

/***************************************************************************************/
/* Aal5InitReassembler								       */
/*										       */	
/* Initializes the specified Sar reassembly state machine, preparing for reassembly    */ 
/***************************************************************************************/
void Aal5InitReassembler(SarReassembler *pSar, ETHERNET_PACKET_BUFFER *pBuf)
{
    ZAP(DEBUG_SAR, DEBUG_ENTRY, ("+++Aal5InitReassembler: pSar=%p, pBuf=%p\n", pSar, pBuf));
    pSar->pQueueEntry           = pBuf;
    pSar->pReassemblyBuffer     = pBuf->GB.pData;
    pSar->BytesLeftInReassembly = pSar->ReassemblyBufferSize = pBuf->GB.AllocedSize;
    pSar->PduLengthFromTrailer  = 0;
    pSar->RunningCRC            = CRC_INITIAL_VALUE;
    pSar->CellCount             = 0;
    ZAP(DEBUG_SAR, DEBUG_EXIT, ("---Aal5InitReassembler\n"));
}

/***************************************************************************************/
/* Aal5ReassembleNonFinalCell							       */
/*										       */
/* Simply copies the payload of the cell into the reassembly buffer. Since we	       */
/* don't do anything with the cell header, the function expects the pointer	       */	
/* to be pointing to the payload of the cell, not the beginning of the cell	       */	
/*										       */	
/* Notes:									       */	
/*       Assumes there is ample space in the reassembly buffer for the cell	       */
/***************************************************************************************/
void Aal5ReassembleNonFinalCell( SarReassembler *pSar, 
                                 UInt8          *pCellPayload )
{
    ZAP(DEBUG_SAR, DEBUG_ENTRY, ("+++Aal5ReassembleNonFinalCell:\n"));

    /* First, lets calculate the crc for the cell*/
    pSar->RunningCRC = CrcCalculate(pCellPayload, 
                                    ATM_CELL_PAYLOAD_SIZE, 
                                    pSar->RunningCRC);

    /* Copy the cell data into the reassembly buffer*/
    ZAP(DEBUG_SAR, DEBUG_MEMCOPY, (kTAB"src=%p, dst=%p, len=%x\n", pCellPayload, 
                                   pSar->pReassemblyBuffer, ATM_CELL_PAYLOAD_SIZE));
    COPY_MEM(pCellPayload, pSar->pReassemblyBuffer, ATM_CELL_PAYLOAD_SIZE);
    pSar->BytesLeftInReassembly -= ATM_CELL_PAYLOAD_SIZE;
    pSar->pReassemblyBuffer     += ATM_CELL_PAYLOAD_SIZE;

    ZAP(DEBUG_SAR, DEBUG_EXIT, ("---Aal5ReassembleNonFinalCell\n"));
}

/***************************************************************************************/
/* Aal5ReassembleFinalCell							       */	
/*										       */
/* If this final cell contains real Pdu data, copies that data into the 	       */
/* reassembly buffer. If not, determines if we copied some padding from the	       */
/* previous cell and backs up our reassembly pointer appropriately		       */	
/*										       */	
/* Notes:									       */	
/*       Assumes there is ample space in the reassembly buffer for the cell	       */	
/***************************************************************************************/
void Aal5ReassembleFinalCell( SarReassembler *pSar, 
                              UInt8          *pCellPayload )
{
    Aal5Trailer *pTrailer;
    UInt16 RealPduDataLeft, PaddingSize;

    ZAP(DEBUG_SAR, DEBUG_ENTRY, ("+++Aal5ReassembleFinalCell:\n"));

    pTrailer = (Aal5Trailer *)(pCellPayload + (ATM_CELL_PAYLOAD_SIZE-ATM_AAL5_TRAILER_SIZE));
    /*******************************************************************************/ 
    /* First, lets calculate the crc for the cell. Since this is the last cell,    */
    /* it will have the stored CRC as the last dword. Including the stored CRC     */
    /* in the CRC calculation should result in the final CRC (stored in RunningCRC)*/
    /* being a known value (CRC_RESIDUE from Crc.h).				   */
    /*******************************************************************************/ 
    pSar->RunningCRC = CrcCalculate(pCellPayload, 
                                    ATM_CELL_PAYLOAD_SIZE, 
                                    pSar->RunningCRC);

    /*******************************************************************************/ 
    /* Before we do any actual data copying, we have to figure out how much of     */
    /* this cell is actual Pdu data. It will be one of the following formats:      */
    /*      data+trailer						           */
    /*      data+padding+trailer						   */
    /*      padding+trailer							   */
    /*******************************************************************************/ 
    pSar->PduLengthFromTrailer = BETOHS(pTrailer->PduLength);
    RealPduDataLeft            = pSar->PduLengthFromTrailer % ATM_CELL_PAYLOAD_SIZE;

    /*******************************************************************************/ 
    /* Note that RealPduDataLeft will actually be the amount of real Pdu data      */
    /* in the previous cell if there wasn't enough space left in that cell for     */
    /* the trailer. That would mean this packet contains padding+trailer	   */
    /*******************************************************************************/ 
    if (RealPduDataLeft > (ATM_CELL_PAYLOAD_SIZE-ATM_AAL5_TRAILER_SIZE))
    {
       /* Yep, this cell is just padding+trailer, so last cell was last real data*/
       /* We need to backup our reassembly info to account for padding we copied */
       PaddingSize                  = ATM_CELL_PAYLOAD_SIZE - RealPduDataLeft;
       pSar->pReassemblyBuffer     -= PaddingSize;
       pSar->BytesLeftInReassembly += PaddingSize;
    }
    else
    {
       /* if RealPduDataLeft == 0, then the last cell contained 100% Pdu data, and*/ 
       /* it finished out the Pdu, so we don't have to do any backtracking or     */
       /* additional copying							  */ 
       if (RealPduDataLeft != 0)
       {
          /* This cell contains some valid Pdu data to copy into the reassembly buffer*/
          ZAP(DEBUG_SAR, DEBUG_MEMCOPY, (kTAB"src=%p, dst=%p, len=%x\n", pCellPayload, 
                                         pSar->pReassemblyBuffer, RealPduDataLeft));
          COPY_MEM(pCellPayload, pSar->pReassemblyBuffer, RealPduDataLeft);
          pSar->BytesLeftInReassembly -= RealPduDataLeft;
          pSar->pReassemblyBuffer     += RealPduDataLeft;
       }
    }

    ZAP(DEBUG_SAR, DEBUG_EXIT, ("---Aal5ReassembleFinalCell\n"));
}

