/***********************************************************************************/
/*										   */
/* Copyright (c) 2001, Analog Devices Inc., All Rights Reserved			   */	
/*										   */
/* Uni.c									   */
/*										   */
/* UNI (User Network Interface) support module (similar to AtmUni.c from NDIS)     */
/*										   */
/* This file is part of the "ADI USB ADSL Driver for Linux".			   */
/*										   */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it       */ 
/* and/or modify it under the terms of the GNU General Public License as           */
/* published by the Free Software Foundation; either version 2 of the License,     */
/* or (at your option) any later version.					   */
/*										   */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be      */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	   */	
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		   */
/* GNU General Public License for more details.					   */
/*										   */
/* You should have received a copy of the GNU General Public License		   */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software  */ 
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA       */
/***********************************************************************************/

#include "Adiutil.h"

extern Boolean CheckForAndProcessOAMCell(Hardware *pHw, UInt8 *pCell);

/***********************************************************************************/
/* UniEstablishVc								   */
/*										   */
/* Allocates memory for and initializes our Vc (virtual channel, virtual circuit)  */
/* for this session. Real ATM allows multiple Vc's per Uni, but it seems that      */
/* there are no circumstances at this point that will require more than one        */
/* Vc in this driver. So, we don't do all the extra logic of managing multiple     */
/* Vc's.									   */
/***********************************************************************************/
void UniEstablishVc(Hardware *pHw, UInt8 *pEncapHdr, UInt32 EncapSize)
{
    ZAP(DEBUG_UNI, DEBUG_ENTRY, ("+++UniEstablishVc:\n"));
    
    /* Initialize the various fields of the Vc*/
    pHw->Vc.pEncapHdr  = pEncapHdr;
    pHw->Vc.EncapSize  = EncapSize;
    if (pHw->pDriverOptions)
    {
       pHw->Vc.Vpi        = pHw->pDriverOptions[ConfigVPI].Value;
       pHw->Vc.Vci        = pHw->pDriverOptions[ConfigVCI].Value;
    }
    else
    {
       pHw->Vc.Vpi        = DEFAULT_VPI;
       pHw->Vc.Vci        = DEFAULT_VCI;
    }
    pHw->Vc.VpiVci     = (pHw->Vc.Vpi<<16) | pHw->Vc.Vci;
    pHw->Vc.MaxSduSize = (OUTGOING_DATA_SIZE / ATM_CELL_SIZE) * ATM_CELL_PAYLOAD_SIZE;
    pHw->Vc.ReassemblyInProgress = FALSE;

    /* NDIS4 code always sets these to 0*/
    pHw->Vc.CpcsUu = 0;
    pHw->Vc.Cpi    = 0;

    /* This came straight from the NDIS4 code*/
    pHw->Vc.CellHeader[0] = (UInt8)( ( 0 << 4) | ((pHw->Vc.Vpi & 0xF0) >> 4));
    pHw->Vc.CellHeader[1] = (UInt8)( (pHw->Vc.Vpi << 4) | ((pHw->Vc.Vci & 0xF000) >> 12));
    pHw->Vc.CellHeader[2] = (UInt8)( (pHw->Vc.Vci & 0x0FF0) >> 4);
    pHw->Vc.CellHeader[3] = (UInt8)( (pHw->Vc.Vci & 0x000F) << 4);
    pHw->Vc.CellHeader[4] = 0;

    ZAP(DEBUG_UNI, DEBUG_EXIT, ("---UniEstablishVc\n"));
    return;
}

/***********************************************************************************/
/* UniProcessOutboundPdu							   */
/*										   */
/* Takes an ethernet or raw IP packet and turns it into ATM data. Modeled after    */
/* AtmUniSend in the NDIS code. Where we differ from the NDIS code, comments       */
/* explain why. One major difference is that we will get an entire packet in one   */
/* buffer. The NDIS code gets multiple buffers that compose the entire packet.     */
/* Also, this routine doesn't actually send the data - it just processes the	   */
/* input data into and output buffer.						   */
/*										   */
/* This routine assumes that all parameters have been verified (it doesn't	   */	
/* check for NULL pointers or zero lengths).					   */
/*										   */
/* Assumes the following fields in the Vc are initialized:			   */
/*      EncapSize, pEncapHdr							   */
/*      MaxSduSize								   */
/*      CellHeader								   */
/*      CpcsUu, Cpi								   */
/***********************************************************************************/
int UniProcessOutboundPdu(Hardware *pHw, UInt8 *pPacket, UInt32 PacketSize)
{
    SarSegmenter Sar;
    UInt16 CellsInOutput;
    OUTGOING_BUFFER *pBuf;

    ZAP(DEBUG_UNI, DEBUG_ENTRY, ("+++UniProcessOutboundPdu:\n"));

    /* The NDIS code checks to make sure the VC is ACTIVE. However, since*/
    /* we don't deal with ACTIVE or INACTIVE VC's, we don't need to worry*/
    /* about that check							 */

    /* Make sure that the packet + encapsulation is within the max SDU size*/
    if ((PacketSize + pHw->Vc.EncapSize) > pHw->Vc.MaxSduSize)
    {
       ZAP(DEBUG_UNI, DEBUG_ERRORS, (kTAB"UniProcessOutboundPdu: Packet+Encap size > MaxSduSize\n"));
       pHw->Statistics[STAT_PAKTS_LOST_OSIZE]++;
       return -EMSGSIZE;
    }

    /* Prepare our segmentation state machine for segmentation by first grabbing*/
    /* a segmentation buffer, then initializing the state machine	        */
#ifdef LINUX
    pBuf = &pHw->SegmentationBuffer;
#else
    pBuf = (OUTGOING_BUFFER *)GetBufferFromQ(&pHw->IdleOutQ);
#endif
    if (pBuf == NULL)
    {
       ZAP(DEBUG_UNI, DEBUG_ERRORS, (kTAB"UniProcessOutboundPdu: NULL buffer from IdleOutQ\n"));
       return -ENOBUFS;
    }

    Aal5InitSegmenter(&Sar, (PacketSize+pHw->Vc.EncapSize),
                      pBuf->GB.pData, (UInt8 *)(&(pHw->Vc.CellHeader)));

    /* If there is some RFC2684 encapsulation enabled, put that in the output first*/
    if (pHw->Vc.EncapSize)
       Aal5Segment(&Sar, pHw->Vc.pEncapHdr, pHw->Vc.EncapSize);

    /* Now segment the real network packet into ATM cells*/
    Aal5Segment(&Sar, pPacket, PacketSize);

    /* Finish the CPCS PDU processing by writing the trailer and any padding*/
    CellsInOutput = Aal5WritePadAndTrailer(&Sar, pHw->Vc.CpcsUu, pHw->Vc.Cpi);
    pHw->Statistics[STAT_CELLS_TX] += CellsInOutput;

    /* Size of output buffer is ATM cell count * size of ATM cell*/
    pBuf->TotalBytes = CellsInOutput * ATM_CELL_SIZE;
#ifdef LINUX
    ZAP(DEBUG_UNI, DEBUG_INFO,
        (kTAB"UniProcessOutboundPdu: packet processed into ATM cells - ready to be sent!\n"));
#else
    PutBufferInQ((GENERIC_BUFFER *)pBuf, &pHw->LiveOutQ);
    ZAP(DEBUG_UNI, DEBUG_INFO, (kTAB"UniProcessOutboundPdu: buffer added to LiveOutQ\n"));
#endif

    ZAP(DEBUG_UNI, DEBUG_EXIT, ("---UniProcessOutboundPdu\n"));

    return 0; /* success in Linux*/
}

/***********************************************************************************/
/* UniLookupVc									   */
/*										   */
/* Finds the Vc that matches the specified ATM header VpiVci value. For now	   */
/* (and maybe forever), we only support on Vc, so this is simply a check to 	   */
/* make sure the VpiVci matches whats stored, then a return of NULL or the	   */
/* CurrentVc pointer.								   */
/***********************************************************************************/
ATMVC *UniLookupVc(Hardware *pHw, UInt32 VpiVci)
{
    ZAP(DEBUG_UNI, DEBUG_ENTRY, ("+++UniLookupVc\n"));

    if (pHw->Vc.VpiVci == VpiVci)
    {
       ZAP(DEBUG_UNI, DEBUG_EXIT, ("---UniLookupVc\n"));
       return &pHw->Vc;
    }
    else
    {
       ZAP(DEBUG_UNI, DEBUG_ERRORS,
           ("UniLookupVc: pCurrentVc->VpiVci (%x) does not match VpiVci (%x)in cell!\n",
                                      pHw->Vc.VpiVci, VpiVci));
       ZAP(DEBUG_UNI, DEBUG_EXIT, ("---UniLookupVc\n"));
       return NULL;
    }
}

/***********************************************************************************/
/* UniProcessInboundData							   */	
/*										   */
/* Processes the ATM data, reassembling into a CPCS PDU (which is really just	   */
/* an ethernet packet with the CPCS trailer). Once an entire Pdu has been 	   */
/* received, adds the completed ethernet packet buffer to the ReadWaitingQ 	   */
/* so the caller (the upper-level network stack components) can get the packet.    */
/***********************************************************************************/
int UniProcessInboundData(Hardware *pHw, UInt8 *pData, UInt32 DataLength) 
{
    UInt32 CellHeader, VpiVci;
    UInt16 ReassembledPduLength;
    Boolean IsLastCell;
    ATMVC *pVc;

    ZAP(DEBUG_UNI, DEBUG_ENTRY, ("+++UniProcessInboundData: pData=%p, DataLength=%x\n",
                                 pData, DataLength));

    /* Lets first make sure there is an integral number of ATM cells in the buffer*/
    if (DataLength % ATM_CELL_SIZE)
    {
       ZAP(DEBUG_UNI, DEBUG_ERRORS,
           (kTAB"UniProcessInboundData: non-integral number of cells in incoming data.\n"));
       ZAP(DEBUG_UNI, DEBUG_ERRORS,
           (kTAB"UniProcessInboundData: incoming data length = %x.\n", DataLength));
       return -EMSGSIZE;
    }

    /* Process each ATM cell in the input buffer*/
    while (DataLength)
    {
       /* The first DWORD of the ATM cell contains all but the HEC byte of the*/
       /* cell's ATM header - but we don't look at the HEC byte anyway        */
       CellHeader = BETOHL(*((UInt32 *)pData));
       IsLastCell = (Boolean)(CellHeader & ATM_PT_AAL_INDICATE);
       VpiVci     = (CellHeader >> 4) & ATM_CELL_REMOVE_GFC_MASK;
       pVc        = UniLookupVc( pHw, VpiVci );



ZAP(DEBUG_OAM, DEBUG_INFO, (kTAB"******** OAM=%d VCI=%d VPI=%d DataMode=%d VaildVCIVPI=%d EveryTime=%d.\n",\
							pHw->pDriverOptions[ConfigDefaultOAM].Value,\
							pHw->pDriverOptions[ConfigVCI].Value,\
							pHw->pDriverOptions[ConfigVPI].Value,\
							DataMode, VaildVCIVPI, EveryTime));
		if (CheckForAndProcessOAMCell(pHw, pData) )
		{
			// This cell was an OAM cell, so we don't want to include it in 
			// the reassembly. Therefore, skip the cell and continue on ...
			ZAP(DEBUG_UNI, DEBUG_INFO, (kTAB"*** OAM cell received and processed.\n"));
			DataLength -= ATM_CELL_SIZE;
			pData      += ATM_CELL_SIZE;
			pHw->Statistics[STAT_CELLS_OAM_RCVD]++;
			continue;
		}



       /* If we cannot find an open VC for this cell, drop the cell*/
       if (NULL == pVc)
       {
          ZAP(DEBUG_UNI, DEBUG_ERRORS,
              (kTAB"UniProcessInboundData: no VC for cell, VpiVci=%x\n", VpiVci));
          DataLength -= ATM_CELL_SIZE;
          pData      += ATM_CELL_SIZE;
          pHw->Statistics[STAT_CELLS_LOST_VPIVCI]++;
          continue; /* forces code to go back to while (DataLength) line*/
       }

       /* NDIS code now checks to make sure the VC is in the Active state    .*/
       /* Our VC is a PVC that is always active, so we don't do this right now*/
       /* Go ahead and skip over the header for this cell		      */
       DataLength -= ATM_CELL_HEADER_SIZE;
       pData      += ATM_CELL_HEADER_SIZE;

       /* If there is not already a reassembly happening on this VC, get the   */
       /* reassembly structure initialized so we can get started with this cell*/
       if (!pVc->ReassemblyInProgress)
       {
          ETHERNET_PACKET_BUFFER *pReassemblyBuffer;
#ifdef LINUX
          pReassemblyBuffer = &pHw->ReassemblyBuffer;
#else
          pReassemblyBuffer = (ETHERNET_PACKET_BUFFER *)GetBufferFromQ(&pHw->IdleWaitQ);
#endif
            
         /* If no reassembly buffer is available, drop the cell*/
         if (NULL == pReassemblyBuffer)
         {
            ZAP(DEBUG_UNI, DEBUG_ERRORS,
                (kTAB"UniProcessInboundData: no reassembly buffer for cell,\n"));
            DataLength -= ATM_CELL_PAYLOAD_SIZE;
            pData      += ATM_CELL_PAYLOAD_SIZE;
            continue; /* forces code to go back to while (DataLength) line*/
         }

         Aal5InitReassembler(&(pVc->Reassembler), pReassemblyBuffer);
         pVc->ReassemblyInProgress = TRUE;
       }

       /********************************************************************/
       /* If there isn't sufficient space in the reassembly buffer to hold */
       /* another cell payload, drop the entire PDU being reassembled.     */
       /* (This can happen if the last cell of a PDU is lost, causing two  */
       /* PDUs to get concatenated)					   */
       /********************************************************************/
       if (pVc->Reassembler.BytesLeftInReassembly < ATM_CELL_PAYLOAD_SIZE)
       { 
          ZAP(DEBUG_UNI, DEBUG_ERRORS,
              (kTAB"UniProcessInboundData: no space left in buffer for cell, dropping PDU,\n"));
          DataLength -= ATM_CELL_PAYLOAD_SIZE;
          pData      += ATM_CELL_PAYLOAD_SIZE;
          pVc->ReassemblyInProgress = FALSE;
          pHw->Statistics[STAT_CELLS_LOST_OTHER] += pVc->Reassembler.CellCount;
#ifndef LINUX
          /* Return this buffer to its rightful owner*/
          PutBufferInQ((GENERIC_BUFFER *)pVc->Reassembler.pQueueEntry, &pHw->IdleWaitQ);
#endif
          continue; /* forces code to go back to while (DataLength) line*/
       }

       /* Ok, we've eliminated all possible error conditions, we've made sure */
       /* the reassembly buffer is ready to receive this cell, so lets do it! */
       if (!IsLastCell)
       {
          Aal5ReassembleNonFinalCell(&(pVc->Reassembler), pData);
       }
       else
       {
          Aal5ReassembleFinalCell(&(pVc->Reassembler), pData);
       }

       pVc->Reassembler.CellCount++;
        
       /* Start looking at the beginning of the next cell*/
       DataLength -= ATM_CELL_PAYLOAD_SIZE;
       pData      += ATM_CELL_PAYLOAD_SIZE;

       /* If this cell was the last cell in the PDU, do PDU processing*/
       if (IsLastCell)
       {
          /* Calculate the number of bytes we reassembled into this PDU*/
          ReassembledPduLength = pVc->Reassembler.ReassemblyBufferSize -
                                 pVc->Reassembler.BytesLeftInReassembly;
          /**************************************************************************/ 
          /* The way the CRC is calculated should result in the final calculated CRC*/
          /* of the incoming PDU being a known value (because the CRC stored in the */
          /* trailer of the PDU is used in the calculation), so we can verify the   */
          /* CRC by just comparing the end result with the fixed value (CRC_RESIDUE)*/
          /**************************************************************************/ 

          if (pVc->Reassembler.RunningCRC != CRC_RESIDUE)
          {
             ZAP(DEBUG_UNI, DEBUG_ERRORS, (kTAB"UniProcessInboundData: CRC error, dropping PDU,\n"));
             pVc->ReassemblyInProgress = FALSE;
             pHw->Statistics[STAT_CELLS_LOST_CRC]++;
          }
          else
          {
             // CRC is ok, lets check the packet length
             if (ReassembledPduLength != pVc->Reassembler.PduLengthFromTrailer)
             {
                ZAP(DEBUG_UNI, DEBUG_ERRORS,
                    (kTAB"UniProcessInboundData: PDU length error, dropping PDU, Calculated = %x, stored = %x\n",
                    ReassembledPduLength, pVc->Reassembler.PduLengthFromTrailer));
                pVc->ReassemblyInProgress = FALSE;
                pHw->Statistics[STAT_PAKTS_LOST_OSIZE]++;
             }

             /* If we get here and ReassemblyInProgress is still true, then we can*/
             /* go ahead and mark this buffer as being ready to receive.          */
             if (pVc->ReassemblyInProgress)
             {
                /***********************************************************/
                /* There might be some MPOA (RFC1483/2684) processing that */
                /* needs to be done, but we can wait until the higher level*/
                /* code actually asks us for this packet. Reset skip size  */
                /* for now to make sure its at a known value		   */
                /***********************************************************/

                pHw->Statistics[STAT_CELLS_RX] += pVc->Reassembler.CellCount;
                pVc->Reassembler.pQueueEntry->PacketSize     = ReassembledPduLength;
                pVc->Reassembler.pQueueEntry->EncapSkipSize  = 0;
                pVc->Reassembler.pQueueEntry->DropThisPacket = FALSE;
                ZAP(DEBUG_UNI, DEBUG_INFO,
                    (kTAB"UniProcessInboundData: Pdu processed successfully, calling ReadSendItUp,\n"));
                pVc->ReassemblyInProgress = FALSE;
                ReadSendItUp(pHw, pVc->Reassembler.pQueueEntry);
             }
#ifndef LINUX
             else
             {
                /* Something caused us to need drop this PDU, so lets return*/
                /* this buffer back to the idle receive buffer queue        */ 
                PutBufferInQ((GENERIC_BUFFER *)pVc->Reassembler.pQueueEntry, &pHw->IdleWaitQ);
             }
#endif
          }
       }
    }

    ZAP(DEBUG_UNI, DEBUG_EXIT, ("---UniProcessInboundData\n"));
    return 0; /* linux success*/
}
