/*********************************************************************************/
/*-------------------------------------------------------------------------------*/
/*										 */
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved			 */	
/*										 */
/* AdiUsbAdslDriver.c								 */
/*										 */
/* Main code for Linux kernel driver.					         */
/*										 */
/*-------------------------------------------------------------------------------*/
/* This file is part of the "ADI USB ADSL Driver for Linux".   			 */
/*   										 */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it     */
/* and/or modify it under the terms of the GNU General Public License as         */
/* published by the Free Software Foundation; either version 2 of the License,   */  
/* or (at your option) any later version.   				         */
/*									         */		
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be    */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   		 */
/* GNU General Public License for more details.   				 */
/*   										 */ 	
/* You should have received a copy of the GNU General Public License   		 */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software*/   
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     */ 
/*********************************************************************************/

#include "Adiutil.h"
#include "AdiUsbAdslDriver.h"
#include "Options.h"
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
/*********************************************************************************/
/* Local prototypes								 */ 	
/*********************************************************************************/
static void *adi_probe(struct usb_device *usb, unsigned int ifnum, const struct usb_device_id *id);
static void adi_disconnect(struct usb_device *usb, void *ptr);
static void adi_irq(struct urb *urb);
static int adi_open(struct net_device *ether);
static int adi_close(struct net_device *ether);
static int adi_ioctl(struct net_device *ether, struct ifreq *rq, int cmd);
static struct net_device_stats *adi_stats(struct net_device *ether);
static int adi_start_xmit(struct sk_buff *skb, struct net_device *ether);
static void adi_set_multicast(struct net_device *ether);
static void adi_tx_timeout(struct net_device *ether);
static int adi_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data);
static int adi_user(struct usb_device *dev, unsigned int code, void *buf);

/*********************************************************************************/
/* Driver global data                                                            */
/*********************************************************************************/

/*Tell usb what VID/PID combinations we support*/

static const struct usb_device_id adi_ids[] = {
{  USB_DEVICE (kEAGLE_VID, kEAGLEI_PID_PREFIRMWARE) }, // before USB firmware upload
{  USB_DEVICE (kEAGLE_VID, kEAGLEI_PID_PSTFIRMWARE) }, // after  USB firmware upload

{  USB_DEVICE (kEAGLE_VID, kEAGLEII_PID_PREFIRMWARE) }, // before USB firmware upload
{  USB_DEVICE (kEAGLE_VID, kEAGLEII_PID_PSTFIRMWARE) }, // after  USB firmware upload

{  USB_DEVICE (kEAGLE_VID, kEAGLEIIC_PID_PREFIRMWARE) }, // before USB firmware upload
{  USB_DEVICE (kEAGLE_VID, kEAGLEIIC_PID_PSTFIRMWARE) }, // after  USB firmware upload

{ }
};

/*send this to USB to register the driver*/
static struct usb_driver adi_driver = {
    name:       "adiusbadsl",
    id_table:   adi_ids,
    probe:      adi_probe,
    disconnect: adi_disconnect,
    ioctl:      adi_user,
};

/*We only support one physical modem plugged in (who would plug in more than one?)*/
static Hardware modem;

/*********************************************************************************/
/* MODULE stuff									 */
/*********************************************************************************/
MODULE_AUTHOR ("Toronto Design Center (TDC)");
MODULE_DESCRIPTION ("Analog Devices Inc. USB ADSL Modem driver (ethernet over ATM)");
MODULE_DEVICE_TABLE (usb, adi_ids);

/*********************************************************************************/
/* Code that's part of the USB driver interface					 */
/*********************************************************************************/


int ApiDataInitialized=FALSE;   // Flag is used to prevent User API Call before all //
                                // Necessay Data is initialized                     //
API_LINK_INFO ApiIOCTLBuffer;	// IOCTL Buffer for CMV API Calls		    //
										    //
void BootTimerHandler(unsigned long ptr);  // This functions handles delays between //
					   // boot procedure sequence		    //
										    //
int booter_thread(void *data);  // This is the thread that Boot Machine is always   //
				// running in its context.			    //
										    //
/*********************************************************************************/
/* adi_init									 */
/*********************************************************************************/
static int __init adi_init (void)
{
    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_init\n"));

    
    ApiDataInitialized = FALSE;				//
    modem.AdiModemSm.CurrentAdiState = STATE_UNDEFINED;	//
    
    usb_register(&adi_driver);
    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_init\n"));
    return 0;
}

module_init (adi_init);


/*********************************************************************************/
/* adi_exit									 */
/*********************************************************************************/
static void __exit adi_exit (void)
{
    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_exit\n"));

    usb_deregister (&adi_driver);

    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_exit\n"));
}

module_exit (adi_exit);


/*********************************************************************************/
/* adi_probe									 */
/*										 */
/* After we've uploaded the usb firmware, this gets called three times, 	 */
/* once for each interface exposed by the modem. Just FYI. I need to try	 */	
/* to see if we can just tell usb that we want the whole device, so he		 */
/* doesn't search for drivers for the individual interfaces.			 */
/*********************************************************************************/
static void *adi_probe(struct usb_device *usb, unsigned int ifnum, const struct usb_device_id *id)
{
    UInt32 pid;
    void *pRet = NULL;
    int ret;
    int i;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_probe\n"));

    ZAP(DEBUG_INIT, DEBUG_INFO, (kTAB"vid (%#X) pid (%#X), ifnum (%d)\n",
        usb->descriptor.idVendor, usb->descriptor.idProduct, ifnum));

    pid = usb->descriptor.idProduct;
    modem.UnplugFlag = false; // 04152003

    if (usb_set_configuration(usb, usb->config[0].bConfigurationValue) < 0)
    {
       ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"set_configuration failed\n"));
       goto probe_exit;
    }

    /*Make sure everything in modem is initialized!*/
    memset(&modem, 0, sizeof(Hardware));

    modem.usbdev = usb;

    switch (pid)
    {
       case kEAGLEI_PID_PREFIRMWARE:     
          ret = loadUsbFirmware(&modem, kEAGLEI_PID_PREFIRMWARE); break;
 	  break;
       case kEAGLEII_PID_PREFIRMWARE:   
          ret = loadUsbFirmware(&modem, kEAGLEII_PID_PREFIRMWARE); break;
          break;
       case kEAGLEIIC_PID_PREFIRMWARE:   
          ret = loadUsbFirmware(&modem, kEAGLEIIC_PID_PREFIRMWARE); break;
          break;
       case kEAGLEI_PID_PSTFIRMWARE:	
       case kEAGLEII_PID_PSTFIRMWARE:
       case kEAGLEIIC_PID_PSTFIRMWARE:
       {
          struct usb_endpoint_descriptor *epint;
          struct usb_endpoint_descriptor *epindata;
          struct usb_endpoint_descriptor *epoutdata;
          struct usb_endpoint_descriptor *epoutidma;
          struct net_device *ether;

          /**********************************************************************/
          /* Some one-time init stuff					        */	
          /* For now we'll just choose the biggest pipe. This will insure       */
          /* that we'll be ok no matter what the train rate is. Unfortunately,  */
          /* the driver model doesn't fit well with delaying read pipe size     */
          /* figuring, etc., until after we're trained.				*/
          /**********************************************************************/
          modem.IsoPipeSize     = kFASTEST_ISO_RATE;
          modem.IsoFramesPerUrb = kFRAMES_PER_ISO_URB;

#ifdef USEBULK
		  modem.LowRateFlag = false; //Low bit Rate: 04042003
#else
		  modem.LowRateFlag = true; //Low bit Rate: 04042003
#endif
          
		  /**********************************************************************/
          /*We use constants for the endpoint addresses - but this is the       */
          /*easiest way for now - and most of the other device-specific         */
          /*usb drivers do the same. Eliminates us having to parse through      */
          /*the configuration descriptor, etc., to get these numbers            */
          /**********************************************************************/
          modem.pipeBulkIdmaOut = usb_sndbulkpipe(usb, kEP_BULK_IDMA_OUT);
          modem.pipeBulkDataOut = usb_sndbulkpipe(usb, kEP_BULK_DATA_OUT);
          modem.pipeBulkDataIn  = usb_rcvbulkpipe(usb, kEP_BULK_DATA_IN);
          modem.pipeIsoDataIn   = usb_rcvisocpipe(usb, kEP_ISOC_DATA_IN);
          modem.pipeIntIn       = usb_rcvintpipe(usb, kEP_INT_IN);

          /*Make sure we can get all the memory we need*/
          modem.pInterruptData = GET_KBUFFER(sizeof(CDC_NOTIFY));
          modem.pReadyData     = GET_KBUFFER(ETHERNET_PACKET_DATA_SIZE);
          modem.pOutgoingData  = GET_KBUFFER(OUTGOING_DATA_SIZE);
          for (i=0; i<INCOMING_Q_SIZE; i++)
          {
#ifdef USEBULK
              modem.pIncomingData[i] = GET_KBUFFER(INCOMING_DATA_SIZE);
#else
              modem.pUrbReadIso[i]   = usb_alloc_urb(modem.IsoFramesPerUrb);
              modem.pIncomingData[i] = GET_KBUFFER((modem.IsoPipeSize * modem.IsoFramesPerUrb));
#endif
              if (!modem.pIncomingData[i])
              {
                 ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer returned from GET_KBUFFER\n"));
                 goto probe_exit;
              }
          }

          /*If any of those buffers were not succesfully allocated, we*/
          /*cannot proceed succesfully                                */
          if ( (!modem.pInterruptData) ||
               (!modem.pReadyData)     ||
               (!modem.pOutgoingData) )
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer returned from GET_KBUFFER\n"));
             goto probe_exit;
          }

			
			   modem.pOAMCell = GET_KBUFFER(128);
			if (modem.pOAMCell == 0 )
			{
				ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer returned from GET_KBUFFER\n"));
			    goto probe_exit;
			}	
			modem.pPacket = GET_KBUFFER(ETHERNET_MAX_PACKET_SIZE);
			if (modem.pPacket == 0 )
			{
				ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer returned from GET_KBUFFER\n"));
				goto probe_exit;
			}	


          /*************************************************************************/
          /* These buffers are used to make the Uni code happy. It was written     */
          /* for MacOS9 - where we had to keep an internal queue of incoming       */
          /* and outgoing packets. Since we don't have to do that here on Linux,   */
          /* we essentially create 1-deep queues and slightly modified the         */
          /* Uni code so it just uses these buffers instead of pulling from queues.*/ 
          /*************************************************************************/
          modem.ReassemblyBuffer.GB.AllocedSize   = ETHERNET_PACKET_DATA_SIZE;
          modem.ReassemblyBuffer.GB.pData         = modem.pReadyData;
          modem.SegmentationBuffer.GB.AllocedSize = OUTGOING_DATA_SIZE;
          modem.SegmentationBuffer.GB.pData       = modem.pOutgoingData;

          /*Initialize the CTRL URB queue*/
          ret = AllocQueuedBuffers(&modem, CTRL_URB_Q_SIZE, sizeof(QUEUED_URB_BUFFER), 
                                   0, &modem.CtrlUrbFreeQ, FALSE);
          if (ret != USB_ST_NOERROR)
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"Ran out of memory in AllocQueuedBuffers\n"));
             goto probe_exit;
          }

          INIT_LIST_HEAD(&modem.CtrlUrbPendQ);

          /*************************************************************************/
          /* Initialize our kernel timers					   */
          /* WatchdogTimer is the "management entity", per the ADI documentation.  */
          /* It is responsible for initiating a regular status check of the modem, */
          /* and it is the main "driver" for the state machine.			   */
          /*************************************************************************/
          init_timer(&modem.AdiModemSm.timerSM);
          modem.AdiModemSm.timerSM.function = WatchdogTimer;
          modem.AdiModemSm.timerSM.data     = (unsigned long)&modem;

          /*************************************************************************/
          /* The UHCI usb driver does not allow us to queue control urbs.          */
          /* Therefore, to make my own asynchronous control urb queue, I           */
          /* have a timer that goes off after 1 second after I sent a control      */
          /* urb to check to see if another control urb needs submitted. 	   */
          /*************************************************************************/
          init_timer(&modem.CtrlUrbQueueTimer);
          modem.CtrlUrbQueueTimer.function  = CtrlUrbQueueWatcher;
          modem.CtrlUrbQueueTimer.data      = (unsigned long)&modem; 

          /*Other init stuff*/
          modem.AdiModemSm.HeartbeatCounter = 0; 
          modem.AdiModemSm.CurrentAdiState  = STATE_JUST_PLUGGED_IN;

          init_timer(&modem.OAMTimer);
          modem.OAMTimer.function  = OAMTimerFunction;
          modem.OAMTimer.data      = (unsigned long)&modem; 


#ifdef USEBULK
          /*****************************************************************************/
          /* Other drivers seem to be calling usb_set_interface, but we're not using an*/
          /* alternate interface on any of the interfaces, so we don't need to as far  */
          /* as I can understand. All we really need to do from a USB stack perspective*/
          /* is tell usb that we will handle all the interfaces for this device, so he */
          /* doesn't enumerate through all drivers for each interface.		       */	
          /*****************************************************************************/
#endif
          usb_driver_claim_interface(&adi_driver, &(usb->actconfig->interface[0]), &modem);
          usb_driver_claim_interface(&adi_driver, &(usb->actconfig->interface[1]), &modem);
          usb_driver_claim_interface(&adi_driver, &(usb->actconfig->interface[2]), &modem);
          /* Get information about our endpoints*/
          epint     = usb->actconfig->interface[0].altsetting[0].endpoint + 0;
          epoutidma = usb->actconfig->interface[1].altsetting[0].endpoint + 0;
          epoutdata = usb->actconfig->interface[1].altsetting[0].endpoint + 1; 
#ifdef USEBULK
          epindata  = usb->actconfig->interface[2].altsetting[0].endpoint + 0;
#else /*ifdef USEBULK*/
          /*******************************************************************************/
          /* Need to eventually choose the alternate interface based on the train ds rate*/
          /* But, like I said above, the driver model doesn't fit well into the	 	 */
          /* concept of delaying the ISO pipe startup					 */
          /*******************************************************************************/
          if (usb_set_interface(usb, kUSB_INTF_IN, kFASTEST_ISO_INTF) < 0)
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"usb_set_interface failed on iso alt 7\n"));
          }
          else
          {
             epindata=usb->actconfig->interface[kUSB_INTF_IN].altsetting[kFASTEST_ISO_INTF].endpoint+0; 
          }
#endif /*ifdef USEBULK else*/


         init_waitqueue_head(&modem.wq);                                  //
         init_waitqueue_head(&modem.wq2);                                 //
         init_waitqueue_head(&modem.wq4);                                 //
		 modem.AdiApiRWPending = FALSE;                                   //
         modem.APICmvError = FALSE;                                       //
         sema_init(&modem.sem,1);                                         // 
         modem.ApiLock = SPIN_LOCK_UNLOCKED;                              //
         spin_lock_init(&modem.ApiLock);                                  //
         modem.ApiTestReq=FALSE;                                          //
         modem.GPRegister = 0;                                            // 
         modem.InitTime = jiffies;					  //
									  //
         init_timer(&modem.BootTimer);					  //
         modem.BootTimer.function = BootTimerHandler;			  //
         modem.BootTimer.data     = (unsigned long)&modem;		  //
         modem.booter_thread_pid = kernel_thread(booter_thread,(void *)&modem,0);			  //
         ApiDataInitialized = TRUE;                                       //


          /*******************************************************************************/
          /* Prepare our interrupt handler so we can handle swap page requests           */
          /* during IDMA boot (and other device interrupts after that)                   */
          /*******************************************************************************/
          FILL_INT_URB(&modem.urbInt, usb, modem.pipeIntIn, modem.pInterruptData,
                       sizeof(CDC_NOTIFY), adi_irq, &modem, epint->bInterval);
          if (usb_submit_urb(&modem.urbInt))
             ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"submit of Int urb failed\n"));

          /*All is well, we need to get a kernel Ethernet interface*/
          ether = init_etherdev(NULL, 0);
          if (!ether)
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"Failed to initialize ethernet device\n"));
             goto probe_exit;
          }

          /*Initialize ATM sublayers*/
          CrcGenerateTable();

          /*Link to be able to get back to the net_device from the Hardware*/
          modem.pLinuxNet = ether;

          /*We've got an ethernet interface - tell the kernel aobut our exports*/
          ether->priv               = &modem;

          /*******************************************************************************/
          /*The following has been added to assign a name to our interface. This also    */
          /*solves the problem of having an ethernet card in system and connecting the   */
          /*ADIUSB modem and booting the system. Previously we used to loose the main    */
          /*ethernet card                                                                */
          /*******************************************************************************/
          strcpy(ether->name,"ADIModem");
          ether->open               = adi_open;
          ether->stop               = adi_close;
          ether->do_ioctl           = adi_ioctl;
          ether->get_stats          = adi_stats;
          ether->hard_start_xmit    = adi_start_xmit;
          ether->set_multicast_list = adi_set_multicast;
          ether->tx_timeout         = adi_tx_timeout;
          ether->watchdog_timeo     = kTRANSMIT_TIMEOUT;

          /*Leave mtu at default for now*/
          // ether->mtu = ???;
          ZAP(DEBUG_ENET, DEBUG_ATTN, (kTAB"Default MTU is %x\n", ether->mtu));

          GetMacAddress(&modem);
          memcpy(ether->dev_addr, modem.MAC, ETH_LENGTH_OF_ADDRESS);

          /*Setup our /proc interface*/
          create_proc_read_entry("adimodem", 0, NULL, adi_read_proc, &modem);

        }
          break;
    }

    /*******************************************************************************/
    /* Even though the rest of the drivers that claim multiple interfaces	   */
    /* just go ahead and return NULL from the initial probe - that causes	   */
    /* the usb stack to think the device isn't claimed, so it goes and tries	   */
    /* to load another driver. In our case, since the device advertises 	   */
    /* itself as an ethernet control model comm class device, the acm and	   */
    /* cdcether drivers get loaded. So, we'll return our private pointer	   */
    /* here too (in addition to the usb_driver_claim_interface above), to	   */
    /* keep usb from trying to load another driver				   */
    /*******************************************************************************/
    pRet = &modem;


probe_exit:
    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_probe\n"));
    return pRet;
}

/*******************************************************************************/
/* adi_disconnect							       */		
/*******************************************************************************/
static void adi_disconnect(struct usb_device *usb, void *ptr)
{
    Hardware *pHw = (Hardware *)ptr;
    UInt32 pid;
    int i;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++adi_disconnect\n"));
    ZAP(DEBUG_FFD, DEBUG_ENTRY, ("+++adi_disconnect\n"));
    /*We can't do anything if we don't have a valid Hardware pointer*/
    if (pHw == NULL)
    {
       ZAP(DEBUG_INIT, DEBUG_ERRORS, (kTAB"NULL pointer passed to disconnect\n"));
       ZAP(DEBUG_FFD, DEBUG_ERRORS, (kTAB"NULL pointer passed to disconnect\n"));
       goto dis_done;
    }

    pHw->UnplugFlag = true; //04152003
    wait_ms(10); // test 04152003	
    kill_proc(pHw->booter_thread_pid, SIGKILL, 1);
    
    pid = usb->descriptor.idProduct;

    /*We only do this stuff on the post-firmware*/ 
    if ( (pid != kEAGLEI_PID_PREFIRMWARE) && (pid != kEAGLEII_PID_PREFIRMWARE) && (pid != kEAGLEIIC_PID_PREFIRMWARE))
    {
       /*If timers are currently running, delete them*/
       if (timer_pending(&pHw->AdiModemSm.timerSM))
          del_timer(&pHw->AdiModemSm.timerSM);
       if (timer_pending(&pHw->CtrlUrbQueueTimer))
          del_timer(&pHw->CtrlUrbQueueTimer);
       
       if (timer_pending(&pHw->OAMTimer))
    	  del_timer(&pHw->OAMTimer);
       
       if (timer_pending(&pHw->BootTimer))
    	  del_timer(&pHw->BootTimer);

      /*Remove our network interface from the kernel*/
      if (pHw->pLinuxNet)
      {
         unregister_netdev(pHw->pLinuxNet);
         pHw->pLinuxNet = NULL;
      }

      /*Unlink pending URBs*/
      usb_unlink_urb(&pHw->urbInt);
        
      /*Free memory we alloced*/
      FREE_KBUFFER(pHw->pInterruptData);
      FREE_KBUFFER(pHw->pReadyData);
      FREE_VBUFFER(pHw->pDsp);
      	
      for (i=0; i<INCOMING_Q_SIZE; i++)
      {
          FREE_KBUFFER(pHw->pIncomingData[i]);
#ifndef USEBULK
          usb_free_urb(pHw->pUrbReadIso[i]);
#endif
      }


      FREE_KBUFFER(pHw->pOAMCell);   
	  FREE_KBUFFER(pHw->pPacket);   

      FreeDspData(pHw);

#ifdef DELAY_ISO_STARTUP
      if (pHw->IsReadPipeStarted == TRUE)
         StopReadPipe(pHw);
#endif

      DeAllocQueuedBuffers(&pHw->CtrlUrbFreeQ);
      DeAllocQueuedBuffers(&pHw->CtrlUrbPendQ);
        
      /*Tell usb that we no longer claim these interfaces as our property*/
      usb_driver_release_interface(&adi_driver, &(usb->actconfig->interface[0]));
      usb_driver_release_interface(&adi_driver, &(usb->actconfig->interface[1]));
      usb_driver_release_interface(&adi_driver, &(usb->actconfig->interface[2]));

      /*Remove our /proc/adimodem interface*/
      remove_proc_entry("adimodem", NULL);
    }

dis_done:
    ZAP(DEBUG_INIT, DEBUG_EXIT, ("---adi_disconnect\n"));
    ZAP(DEBUG_FFD, DEBUG_EXIT, ("---adi_disconnect\n"));
}

/*******************************************************************************/
/* adi_irq								       */	
/*******************************************************************************/
static void adi_irq(struct urb *urb)
{
    Hardware *pHw;

    ZAP(DEBUG_INTS, DEBUG_ENTRY, ("+++adi_irq\n"));

    /*We put the modem pointer as the context*/
    pHw = (Hardware *)urb->context;

    /*Lets check the status first, to make sure this was succesfull*/
    if (urb->status<0)
    {
       ZAP(DEBUG_INTS, DEBUG_ERRORS, (kTAB"failed adi_irq urb status = %#x\n", urb->status));
       /*These errors seem to be what we get when the device has been unplugged, so*/
       /*lets stop a bunch of these from happening by unlinking the urb now        */
       if ((urb->status == -EILSEQ) ||
           (urb->status == -ETIMEDOUT))
       {
          usb_unlink_urb(&pHw->urbInt);
       }
       goto irq_done;
    }

    /*All is well, we can process the data!*/
    if (pHw->pInterruptData->bmRequestType == 0x08) // device-to-host interrupt
    {
       DeviceInt *pInt;

       pInt = (DeviceInt *)(&(pHw->pInterruptData->Data[0])); 
       switch (pInt->Interrupt)
       {
          case kADI_INT_LOADSWAPPAGE:
          {
             UInt16 SwapData = HSTOLE(pInt->SwapIntInfo.SwapData);
             ZAP(DEBUG_INTS, DEBUG_INFO, (kTAB"interrupt = kADI_INT_LOADSWAPPAGE\n"));
             if (pHw->AdiModemSm.CurrentAdiState != STATE_HARD_RESET_INITIATE)
                IdmaUploadSwapPage(pHw, SwapData);
          }
             break;
          case kADI_INT_INCOMINGCMV:
          {
             ZAP(DEBUG_INTS, DEBUG_INFO, (kTAB"interrupt = kADI_INT_INCOMINGCMV\n"));

             ProcessIncomingCmv(pHw, pInt->CmvIntInfo.CmvData);
          }
             break;
          default:
             ZAP(DEBUG_INTS, DEBUG_INFO, (kTAB"interrupt = unsupported(%d)\n", pInt->Interrupt));
             break;
        }
    }

irq_done:
    ZAP(DEBUG_INTS, DEBUG_EXIT, ("---adi_irq\n"));
}

/*******************************************************************************/
/* Code that's part of the ethernet driver interface			       */	
/*******************************************************************************/

/*******************************************************************************/
/* adi_open								       */	
/*******************************************************************************/
static int adi_open(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;

    MOD_INC_USE_COUNT;
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_open\n"));

#ifdef DELAY_ISO_STARTUP
    /*This will queue up multiple BULK IN or ISO IN urbs*/
    /*Only start the read pipe if we've successfully trained*/
    if (pHw->AdiModemSm.CurrentAdiState & STATE_OPERATIONAL )
#endif
       StartReadPipe(pHw);

    /*Tell the kernel we're ready for network traffic flow*/
    netif_start_queue(ether);

    /*Tell ourselves that we're open for business!*/
    pHw->IsOpen = TRUE;

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_open\n"));
    return 0;
}

/*******************************************************************************/
/* adi_close								       */	
/*******************************************************************************/
static int adi_close(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;

    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_close\n"));

    /*Tell ourselves that we're closed for business!*/
    pHw->IsOpen = FALSE;

    /*Tell the kernel that we do not want any more network traffic*/
    netif_stop_queue(ether);

    /*Stop our incoming data urbs*/
#ifdef DELAY_ISO_STARTUP
    if (pHw->IsReadPipeStarted == TRUE)
#endif
       StopReadPipe(pHw);

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_close\n"));
    MOD_DEC_USE_COUNT;
    return 0;
}

/*******************************************************************************/
/* adi_ioctl 								       */
/*******************************************************************************/
static int adi_ioctl(struct net_device *ether, struct ifreq *rq, int cmd)
{
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_ioctl\n"));

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_ioctl\n"));
    /*This particular ioctl is accessed through the socket API*/
    /*For now, I don't know of any ioctls that we need to support*/
    return -EOPNOTSUPP;
}

/*******************************************************************************/
/* adi_stats 								       */
/*******************************************************************************/
static struct net_device_stats *adi_stats(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;

    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_stats\n"));

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_stats\n"));
    return &pHw->LinuxStats;
}

/*******************************************************************************/
/* adi_start_xmit 							       */	
/*******************************************************************************/
static int adi_start_xmit(struct sk_buff *skb, struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;
    ETHERNET_MAC_HEADER *pHdr = (ETHERNET_MAC_HEADER *)skb->data;	
    int result=0;


    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_start_xmit\n"));


#ifdef DELAY_ISO_STARTUP
    /*If the read pipe hasn't started, then we haven't trained yet*/
    if (pHw->IsReadPipeStarted == FALSE)
    {
       ZAP(DEBUG_ENET, DEBUG_ERRORS, (kTAB"adi_start_xmit, but read pipe not started.\n"));
       pHw->LinuxStats.tx_errors++;
       goto xmit_exit;
    }
#endif

    /*Tell the kernel to hold off until we're done sending this packet*/
    netif_stop_queue(ether);

    /*Send the packet through the SAR to get ATM cells*/

    switch (pHw->MpoaMode)
    {
       case MPOA_MODE_ROUTED_IP_LLC   :
       case MPOA_MODE_ROUTED_IP_VC    :

			if (pHdr->EtherType != ETHERTYPE_IP)
			{
				if (pHdr->EtherType == ETHERTYPE_ARP)
				{
					EthernetArpReply(pHw, skb->data, skb->len);
					//goto xmit_exit;

					/***********************************************************
					  Note that in the MAC Code we go to xmit_exit but here it
					  seems that we have to somehow return the send function
					  successfully; for now we countiue the send function after 
					  sending back the ARP Request.This might be a potential BUG
					************************************************************/
					result = UniProcessOutboundPdu(pHw, skb->data, skb->len);
				}
				else
				{
					// If the packet is not an ARP or IP, then its invalid, but
					// tell Linux that all is well
					ZAP(DEBUG_ENET, DEBUG_ERROR, ("ERROR: Invalid ethernet in packet.\n"));
					goto xmit_exit;
				} 
			}
			else
			{
				// Adjust the packet pointer and size so we
				// skip the ethernet header
				skb->data   += sizeof(ETHERNET_MAC_HEADER);
				skb->len -= sizeof(ETHERNET_MAC_HEADER);
				result = UniProcessOutboundPdu(pHw, skb->data, skb->len);
			}

		  break;
       case MPOA_MODE_BRIDGED_ETH_LLC :
       case MPOA_MODE_BRIDGED_ETH_VC  :
          result = UniProcessOutboundPdu(pHw, skb->data, skb->len);
          break;
       case MPOA_MODE_PPPOA_LLC       :
       case MPOA_MODE_PPPOA_VC        :
          result = UniProcessOutboundPdu(pHw, skb->data+16, skb->len-16);
          break;
    }

    if (result)
    {
       ZAP(DEBUG_ENET, DEBUG_ERRORS, (kTAB"Error %x from UniProcessOutboundPdu, not sending.\n", result));
       pHw->LinuxStats.tx_errors++;
       netif_start_queue(ether);
    }
    else
    {
        /*Uni says all is ok to proceed with sending the data*/
        if ((result = WriteAtmData(pHw)))
        {
           /*Something went wrong in the write*/
           ZAP(DEBUG_ENET, DEBUG_ERRORS, (kTAB"Error %x from WriteAtmData, not sending.\n", result));
           pHw->LinuxStats.tx_errors++;
           netif_start_queue(ether);
        }
        else
        {
           pHw->LinuxStats.tx_packets++;
           pHw->LinuxStats.tx_bytes += skb->len;
           /*Not sure what this is for - but everyone else does it!*/
           ether->trans_start = jiffies;
           /*******************************************************************************/
           /*You'll notice that in this case, we do not start the netif queue             */
           /* .... that's because we're only using one write URB, so we have		  */
           /* to wait until the completion handler for the write is called		  */
           /* before we can tell kernel that we can handle more network data.		  */
           /* A potential performance improvement would be to have an internal		  */
           /* URB queue for outgoing data, so we could (theoretically) start		  */
           /* the software-processing of a packet while another one is being sent.	  */	
           /*******************************************************************************/
        }
    }

xmit_exit:
    /*We've copied the data into our own internal buffers, so we can*/ 
    /*go ahead and free the kernel's buffer for this packet*/
    dev_kfree_skb(skb);
    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_start_xmit\n"));
    return 0;
}

/*******************************************************************************/
/* adi_set_multicast							       */ 
/*******************************************************************************/
static void adi_set_multicast(struct net_device *ether)
{
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_set_multicast\n"));

    /*******************************************************************************/
    /* Our ethernet side is really all software, so anything that		   */
    /* a normal ethernet card would do in hardware, we do here in the driver.	   */
    /* On a normal ethernet card, you would upload the filter address list	   */
    /* to the card - but we just loop through a fast software compare. 		   */
    /* Since we will just always check whats in the net_device *, we don't 	   */
    /* have to do anything here!						   */
    /*******************************************************************************/

    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_set_multicast\n"));
}


/*******************************************************************************/
/* adi_tx_timeout							       */	
/*******************************************************************************/
static void adi_tx_timeout(struct net_device *ether)
{
    Hardware *pHw = (Hardware *)ether->priv;
    ZAP(DEBUG_ENET, DEBUG_ENTRY, ("+++adi_tx_timeout\n"));

    ZAP(DEBUG_ENET, DEBUG_ERRORS, (kTAB"WARNING **** transmit timed out!\n"));

    pHw->urbWrite.transfer_flags |= USB_ASYNC_UNLINK;
    usb_unlink_urb(&pHw->urbWrite);

    pHw->urbOAMWrite.transfer_flags |= USB_ASYNC_UNLINK;
    usb_unlink_urb(&pHw->urbOAMWrite);


    ZAP(DEBUG_ENET, DEBUG_EXIT, ("---adi_tx_timeout\n"));
}

/*******************************************************************************/
/* adi_read_proc							       */	
/*******************************************************************************/
static int adi_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
    Hardware *pHw = (Hardware *)data;
    int size;
    char *p = page;

    /*If this output needs to be larger than 4K (PAGE_SIZE), we need to do this differently*/
    p += sprintf(p, "\nAnalog Devices USB ADSL Modem Status Display\n");
    p += sprintf(p, "-------------------------------------------------------------\n");
    p += sprintf(p, "Tx Rate  %#8.8x  Rx Rate  %#8.8x  Crc      %#8.8x\n", 
                 pHw->AdiModemSm.UpRate / 1024, 
                 pHw->AdiModemSm.DownRate / 1024, 
                 pHw->Statistics[STAT_CELLS_LOST_CRC]);

    p += sprintf(p, "FEC      %#8.8x  Margin   %#8.8x  Atten    %#8.8x\n",
                 pHw->AdiModemSm.stats_Uncorr_Blks,
                 pHw->AdiModemSm.stats_Cur_SNR & 0xFF,
                 pHw->AdiModemSm.stats_Cur_Atten & 0xFF);

    p += sprintf(p, "VID-CPE  %#8.8x  VID-CO   %#8.8x  HEC      %#8.8x\n",
                 pHw->AdiModemSm.INFO14,
                 pHw->AdiModemSm.INFO08,
                 pHw->AdiModemSm.DIAG03);

    p += sprintf(p, "VPI      %#8.8x  VCI      %#8.8x  Delin    ", 
                 pHw->Vc.Vpi, pHw->Vc.Vci);

    /*Delineation is the only one where we print a string instead of a number*/
    if (pHw->AdiModemSm.flags & 0x0C00)
       p += sprintf(p, "ERROR\n");
    else
       if (pHw->AdiModemSm.flags & 0x0030) 
          p += sprintf(p, "LOSS\n");
       else
          p += sprintf(p, "GOOD\n");

    p += sprintf(p, "Cells Rx %#8.8x  Cells Tx %#8.8x\n",
                 pHw->Statistics[STAT_CELLS_RX],
                 pHw->Statistics[STAT_CELLS_TX]);

    p += sprintf(p, "Pkts Rx  %#8.8x  Pkts Tx  %#8.8x\n",
                 pHw->Statistics[STAT_PAKTS_RX],
                 pHw->Statistics[STAT_PAKTS_TX]);

    switch (pHw->AdiModemSm.CurrentAdiState)
    {
       case STATE_UNDEFINED:
          p += sprintf(p, "Modem is unplugged from USB.\n");
          break;
       case STATE_JUST_PLUGGED_IN:
          p += sprintf(p, "Modem waiting for driver response.\n");
          break;
       case STATE_UNTRAIN:
       case STATE_UNTRAIN_TX:
       case STATE_UNTRAIN_RX:
       case STATE_INITIALIZING:
       case STATE_INITIALIZING_TX:
       case STATE_INITIALIZING_RX:
          p += sprintf(p, "Modem is initializing\n");
          break;
       case STATE_HARD_RESET_INITIATE:
       case STATE_HARD_RESET_END:
       case STATE_BOOT_WAIT:
          p += sprintf(p, "Modem is booting\n");
          break;
       case STATE_OPERATIONAL:
       case STATE_OPERATIONAL_TX:
       case STATE_OPERATIONAL_RX:
          p += sprintf(p, "Modem is operational\n");
          break;
       case STATE_STALLED_FOREVER:
          p += sprintf(p, "Modem cannot boot. Unplug other devices and retry.\n");
          break;
#ifdef ZAPS
       default:
          p += sprintf(p, "Unhandled state\n");
          Zap(("Timer: unhandled state 0x%X\n", pHw->AdiModemSm.CurrentAdiState));
          break;
#endif
    }

    p += sprintf(p, "\n");

    /*Figure out how much of the page we used*/
    size  = p - page;
    size -= off;

    if (size < count)
    {
       *eof = 1; 
       if (size <= 0)
          return 0;
    }
    else
       size = count;

    /*Fool caller into thinking we started where he told us to in the page*/
    *start = page + off;

    return size;
}

/*******************************************************************************/
/* adi_user								       */	
/*									       */	
/* This is actually the ioctl handler for ioctls coming from the usermode      */
/* helper application. The other ioctl routine, adi_ioctl, will receive        */
/* ioctls coming from network socket applications.			       */
/*******************************************************************************/
static int adi_user(struct usb_device *dev, unsigned int code, void *buf)
{
    Hardware *pHw = &modem;
    UInt8 *pBuf = NULL;
    struct adi_dsp_info dsp_info;
    ParamReadType Param;
    int retval;

    
     UserDspRWInf DspRWInfo;  // Dsp Memory Read/Write Data Structure   //
     volatile int State;      // Modem State                            //
    

    ZAP(DEBUG_MISC, DEBUG_ENTRY, ("+++adi_user(code=%d)\n", code));
    switch (code) 
    {
       case ADIUSBADSLDSP:
       {
          if (buf == NULL) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("buf in DSP from app is NULL!\n"));
             retval = -EFAULT;
             goto err_out;
          }

          if (__copy_from_user(&dsp_info, buf, sizeof(struct adi_dsp_info))) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("copy_from_user failed??\n"));
             retval = -EFAULT;
             goto err_out;
          }
            
          pHw->IDMAStart = dsp_info.IDMAStart;
          /*Make sure we free existing DSP buffers, this might be a reload*/
          FreeDspData(pHw);
          FREE_VBUFFER(pHw->pDsp);
          pBuf = GET_VBUFFER(dsp_info.DspSize);

          if (!pBuf) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, (kTAB"Not enough memory for entire DSP code.\n"))
             retval = -EFAULT;
             goto err_out;
          }

          if (__copy_from_user(pBuf, dsp_info.pDsp, dsp_info.DspSize)) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("copy_from_user failed???\n"));
             retval = -EFAULT;
             goto err_out;
          }

          pHw->pDsp = pBuf;
            
          /*Get our DSP code ready for IDMA booting*/
          if (!DspFoodProcessor(&modem))
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, (kTAB"Error in DspFoodProcessor\n"));
             retval = -EFAULT;
             goto err_out;
          }
            
          /*Ok, we're ready to kickoff IDMA boot!*/
          BootTheModem(&modem, FALSE);
       }
          break;
       case ADIUSBADSLRESET:
       {
          BootTheModem(&modem, FALSE);
       }
          break;
       case ADIUSBADSLCMVS:
       {
          /*We use the dsp_info struct as a rider to copy the CMVs*/
          if (buf == NULL) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("buf pointer to CMVs from app is NULL!\n"));
             retval = -EFAULT;
             goto err_out;
          }

          if (__copy_from_user(&dsp_info, buf, sizeof(struct adi_dsp_info))) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("copy_from_user failed??\n"));
             retval = -EFAULT;
             goto err_out;
          }

          if (__copy_from_user(drvCMVs, dsp_info.pDsp, sizeof(drvCMVs))) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("copy_from_user failed???\n"));
             retval = -EFAULT;
             goto err_out;
          }

          /*CMVs retrieved, set options pointer and init Msg and Mpoa*/
          pHw->pDriverCMVs = drvCMVs;
       }
           break;

       case ADIUSBADSLOPTIONS:
       {
          /*We use the dsp_info struct as a rider to copy the options data*/
          if (buf == NULL) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("buf pointer to options from app is NULL!\n"));
             retval = -EFAULT;
             goto err_out;
          }

          if (__copy_from_user(&dsp_info, buf, sizeof(struct adi_dsp_info))) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("copy_from_user failed??\n"));
             retval = -EFAULT;
             goto err_out;
          }

          if (__copy_from_user(drvOptions, dsp_info.pDsp, sizeof(drvOptions))) 
          {
             ZAP(DEBUG_MISC, DEBUG_ERRORS, ("copy_from_user failed???\n"));
             retval = -EFAULT;
             goto err_out;
          }

          /*Options retrieved, set options pointer and init Msg and Mpoa*/
          pHw->pDriverOptions = drvOptions;
          MsgInitialize(&modem);
          MpoaInitialize(&modem, pHw->pDriverOptions[ConfigEncapsulation].Value);
       }
           break;
   
        
        case ADIUSBADSLSOFTRST:								    //
           // Restarts modem after sending CNTL.0.0 CMV to restart DSP processor 	    //
           SoftRestart(pHw); 								    //
           break;									    //
											    //
        case ADIUSBADSLMDRAEDY:								    //
           // Just copy the ApiDataInitialized Flag to user space 			    //
          __copy_to_user((char *)buf,(char *)&ApiDataInitialized,sizeof(int));	 	    //
          break;									    //
											    //
        case ADIUSBADSLREADPAR:								    //
           // Just copy the parameters to user space structure				    //
           Param.VPI=pHw->pDriverOptions[ConfigVPI].Value;				    //
           Param.VCI=pHw->pDriverOptions[ConfigVCI].Value;				    //
           Param.Encapsulation=pHw->pDriverOptions[ConfigEncapsulation].Value;	            //
           __copy_to_user((char *)buf,(char *)&Param,sizeof(ParamReadType));	 	    //
           break;									    //
 											    //
											    //
        case ADIUSBADSLWRITEGP:								    //
           // Copy the General Purpose Register Value from User mode			    //	
           // to hardware specific data structure then call ModemSM  			    //
           // to let state machine handler know about General purpose			    //
           // register change								    //
           __copy_from_user((char *)&(pHw->GPRegister),(char *)buf,sizeof(int)); 	    //
           ModemSM(EVENT_TIMER_TICK,pHw,NULL); 						    //
           break;									    //
											    //
        case ADIUSBADSLREADGP:  							    //
           // Just copy from hardware specific data structure to user mode		    //
           // buffer									    //
           __copy_from_user((char *)buf,(char *)&(pHw->GPRegister),sizeof(int)); 	    //
           break;									    //
 											    //
        case ADIUSBADSLDSPMEMR:								    //
           // 1- Copy Dsp Read/Write data structure passed from user space 		    //
           // 2- Call ReadDspMem (in util.c) with address passed by user in		    //
           //    above structure							    //
           // 3- Copy result to user mode buffer					    //
           __copy_from_user((char *)&DspRWInfo,(char *)buf,sizeof(DspRWInfo));		    //
           ReadDspMem(pHw,DspRWInfo.Addr,&DspRWInfo.Value);				    //
           DspRWInfo.ReturnValue = TRUE;						    //
           __copy_to_user((char *)buf,(char *)&DspRWInfo,sizeof(DspRWInfo));		    //
           break;									    //
											    //
        case ADIUSBADSLDSPMEMW:								    //
           // 1- Copy Dsp Read/Write data structure passed from user space 		    //
           // 2- Call WriteDspMem (in util.c) with address passed by user in		    //
           //    above structure							    //	
           // 3- Copy result to user mode buffer 					    //
           __copy_from_user((char *)&DspRWInfo,(char *)buf,sizeof(DspRWInfo));		    //
           WriteDspMem(pHw,DspRWInfo.Addr,&DspRWInfo.Value);				    //
           DspRWInfo.ReturnValue = TRUE;						    //
           __copy_to_user((char *)buf,(char *)&DspRWInfo,sizeof(DspRWInfo));		    //
           break;									    //
           										    //
											    //
        case ADIUSBADSLTESTREQ: 							    //
           // 1- Make Api test request flag on. This causes modem go to test 		    //
           //    mode after getting swap page number 1 & "Modem Ready" asynchronous	    //
           //    message								    //
           // 2- Put an undefined value to AdiModemSm.ModemReplyExpected to prevent	    //
           //    Watchdog timer reset modem during modem is in test state		    //
           // 3- Set cuurent state to STATE_TEST and statr Idma boot process    	    //
           pHw->ApiTestReq=TRUE; 							    //
           pHw->AdiModemSm.ModemReplyExpected = 0xff;					    //
           pHw->AdiModemSm.CurrentAdiState = STATE_TEST_RX;				    //
           BootTheModem(&modem, TRUE);							    //
           break;									    //
											    //
        case ADIUSBADSLGETSTATE:							    //
           // just copy current state variable from hardware specific structure to 	    //
           // user mode buffer								    //
           State = pHw->AdiModemSm.CurrentAdiState;					    //
           __copy_to_user((char *)buf,(char *)&State,sizeof(int)); 			    //
           break;									    //
											    //		
        case ADIUSBADSLSTATE:								    //
           // Just copy current state from user mode buffer to hardware specific data	    //
           // structure and call ModemSM to let state machine handler know about	    //
           // state change								    //
           __copy_from_user((char *)&State,(char *)buf,sizeof(int)); 			    //
           pHw->AdiModemSm.CurrentAdiState = State;					    //
           ModemSM(EVENT_TIMER_TICK,pHw,NULL);						    //
           break;									    //
											    //
        case ADIUSBADSLMAC:								    //
           // Just copy MAC address from hardware specific data structure to user mode      //
           // buffer 									    //
           __copy_to_user((char *)buf, (char *)modem.MAC, 				    //
           sizeof(UInt8)*ETH_LENGTH_OF_ADDRESS); 					    //
           break; 									    //
											    //
        case ADIUSBADSLVER:								    //
           // Just copy constant driver version to user mode buffer 			    //
           __copy_to_user((char *)buf, (char *)driver_version, sizeof(driver_version));	    //
           break; 									    //
											    //
        case ADIUSBADSLCMV :								    //
           // This is the entry of all CMV related API Functions			    //
											    //
           // 1 - If data is not initialized (wait 1 second  to make sure) just get out and //
           //  return error code = 2	    						    //
           if ((ApiDataInitialized==FALSE) ||(MSEC_TO_JIFFIES(1000)>(jiffies-pHw->InitTime))//
               /*|| (modem.AdiModemSm.CurrentAdiState==STATE_JUST_PLUGGED_IN)*/	)	    //
           {										    //
             ApiIOCTLBuffer.ReturnStatus=2;						    //
             __copy_to_user((char *)buf,(char *)&ApiIOCTLBuffer,sizeof(ApiIOCTLBuffer));    //
             return 1;									    //
           } 										    //
											    //
           // 2- Wait on semaphore for other process to complete. This is to prevent	    //
           //    two or more processes send CMV command simultanously because modem	    //
           //    can handle only one CMV at a time. The second process waits until	    //
           //    first one finishes its task. That's why we don't have CMV_PENDING	    //
           //    error return value as in windows model.				    //
           down_interruptible(&pHw->sem);						    //
											    //
           // 3- Make sure this command is not sent less than 200 ms after last one	    //
           //    to let modem do it's own tasks and not to overload it. 		    //
           if ((jiffies-pHw->LastApiTime)<MSEC_TO_JIFFIES(TIME_BETWEEN_CMV))		    //
              wait_ms(TIME_BETWEEN_CMV);						    //				
											    //
											    //
           // 4- Check if CMV is acceptable at this time. CMVs are not acceptable untill    //
           //    modem asks for swap page number 1 and driver writes it in modem memory     //
           //    but to be more safe CMV is not acceptable from user during boot untill     //
           //    modem goes to STATE_INIATLIZE state 					    //
           if (!pHw->UserCmvApiAcceptable)  // Modem not in booting mode 		    //
           {										    //
              ApiIOCTLBuffer.ReturnStatus=1;						    //
           }										    //
           else										    //
           {										    //
             // 5- Copy IOCTL buffer from user space  					    //
             __copy_from_user((char *)&ApiIOCTLBuffer,(char *)buf,sizeof(ApiIOCTLBuffer));  //

             // 6- Send CMV Command to modem. If kernel mode CMV is running, the	    //
             //    user mode CMV would be queued.					    //
             UserMsgSend(pHw,(UInt16 *)ApiIOCTLBuffer.DataBuffer);			    //
											    //
             // 7- Sleep and wait for interrupt handler to wake up 			    //
             pHw->APICmvError = 1;							    //
             interruptible_sleep_on_timeout(&(pHw->wq),MSEC_TO_JIFFIES(200));		    //
											    //
             if (pHw->APICmvError)							    //
               // 8- Check if modem replies error situation (APICmvError 		    //
               //    is set in interrupt handler in Me.c)				    //
               ApiIOCTLBuffer.ReturnStatus=3;						    //
             else									    //
             {										    //
               // 9- Success! Copy data from modem to kernel IOCTL buffer		    //
               ApiIOCTLBuffer.ReturnStatus=0;						    //
               memcpy((char *)ApiIOCTLBuffer.DataBuffer,(char *)(pHw->ApiCmvRaw)     	    //
               ,2*kCMV_DATA_WORDS);							    //
               ApiIOCTLBuffer.DataWord = pHw->ApiCmvData;				    //
             }										    //
           }										    //
											    //
           // 10- Copy kernel mode IOCTL buffer to user mode buffer			    //
           __copy_to_user((char *)buf,(char *)&ApiIOCTLBuffer,sizeof(ApiIOCTLBuffer));	    //
   											    //
           // 11- Save time (to be checked in the next CMV for 200ms wait).		    //
           pHw->LastApiTime = jiffies;							    //
											    //
           // 12- Signal on semaphore to make other processes waiting free.		    //
           up(&pHw->sem);								    //
											    //	
           // 13- Sleep 5ms! This is for getting other process waiting on semaphore	    //
           //     get cpu and pass the semaphore. otherwise this process will continue	    //
           //     and will get the cpu if API functions are called in a loop.		    //
           wait_ms(5);							   		    //
											    //
           break; 									    //
          


       default:
       {
          retval = -ENOIOCTLCMD;
          goto err_out;
       }
           break;
    }

    retval = 1;

err_out:
    ZAP(DEBUG_MISC, DEBUG_EXIT, ("---adi_user\n"));
    return retval;
}
































