/*********************************************************************************/
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved			 */	
/*										 */
/* Pipes.c									 */
/*										 */
/* Code for dealing with read and write data pipes				 */
/*										 */
/* This file is part of the "ADI USB ADSL Driver for Linux".			 */	
/*										 */
/* "ADI USB ADSL Driver for Linux" is free software; you can redistribute it     */
/* and/or modify it under the terms of the GNU General Public License as 	 */
/* published by the Free Software Foundation; either version 2 of the License,   */
/* or (at your option) any later version.					 */	
/*										 */
/* "ADI USB ADSL Driver for Linux" is distributed in the hope that it will be    */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the		 */
/* GNU General Public License for more details.					 */
/*										 */
/* You should have received a copy of the GNU General Public License		 */
/* along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software*/
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     */
/*********************************************************************************/
#include "Adiutil.h"

/*********************************************************************************/
/* Local prototypes								 */
/*********************************************************************************/
static void ReadBulkCompletion(struct urb *urb);
static void ReadIsocCompletion(struct urb *urb);
static void WriteCompletion(struct urb *urb);

/**********************************************************************************/
/* StartReadPipe								  */	
/*										  */	
/* Prepare to receive incoming network data by queueing up multiple BULK IN urbs. */
/* We previously followed the same model as the other USB network drivers, in that*/
/* we only allowed one incoming block of network data at a time. However, this    */
/* resulted in lost data (incoming data would be lost because it wasn't read by   */
/* USB in time). So, we've now switched to the same model that we use in the      */
/* MacOS9 driver, We queue up multiple BULK IN requests, so that (theoretically), */
/* the DMA of an incoming packet could be happening while the CPU is doing the    */
/* reassembly of another packet.						  */
/**********************************************************************************/
void StartReadPipe(Hardware *pHw)
{
    int result;
    int i;

    if (pHw->AdiModemSm.CurrentAdiState & STATE_OPERATIONAL )
    {
       ZAP(DEBUG_READ, DEBUG_ENTRY, ("+++StartReadPipe\n"));

#ifdef USEBULK
       /*Start the BULK read pipe URBs*/
       for (i=0; i<INCOMING_Q_SIZE; i++)
       {
            FILL_BULK_URB(&pHw->urbRead[i], pHw->usbdev, pHw->pipeBulkDataIn, 
                          pHw->pIncomingData[i], INCOMING_DATA_SIZE, ReadBulkCompletion, pHw);
            pHw->urbRead[i].transfer_flags |= USB_QUEUE_BULK;
            if ((result = usb_submit_urb(&pHw->urbRead[i])))
               ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error %x on read URB submit.\n", result));
        }
#else
    /*****************************************************************************/
    /* Start the ISOCH read URBs ... first we go through the ring and initialize */
    /* all the data. Since we setup an URB ring, we only have to actually submit */
    /* the first URB in the ring, the rest will be submitted through the URB     */
    /* linking mechanism (via the ->next field)					 */
    /*****************************************************************************/
        for (i=0; i<INCOMING_Q_SIZE; i++)
        {
            int j;
            urb_t *pUrb;

            pUrb = pHw->pUrbReadIso[i];
            spin_lock_init(&pUrb->lock);
            /*Init the URB fields per the spec*/
            pUrb->dev                    = pHw->usbdev;
            pUrb->pipe                   = pHw->pipeIsoDataIn;
            pUrb->context                = pHw;
            pUrb->complete               = ReadIsocCompletion;
            pUrb->transfer_buffer        = pHw->pIncomingData[i];
            pUrb->transfer_buffer_length = pHw->IsoPipeSize * pHw->IsoFramesPerUrb;
            pUrb->number_of_packets      = pHw->IsoFramesPerUrb;
            pUrb->transfer_flags         = USB_ISO_ASAP;
            /*Setup the frame descriptors that will receive the data*/
            for (j=0; j<pHw->IsoFramesPerUrb; j++)
            {
                pUrb->iso_frame_desc[j].offset = j * pHw->IsoPipeSize;
                pUrb->iso_frame_desc[j].length = pHw->IsoPipeSize;
             }
             /***********************************************************************/
             /* Link this urb to the previous one. If we're looking at the first    */
             /* urb, link it to the last one. (If only one urb is being used, then  */
             /* it will link to itself.					       */	
             /***********************************************************************/
             if (i)
             {
                pHw->pUrbReadIso[ i-1 ]->next = pUrb;
                ZAP(DEBUG_READ, DEBUG_ATTN, (kTAB"URB %d(%p) links to URB %d(%p)\n",
                                        i-1, pHw->pUrbReadIso[ i-1 ], i, pUrb));
             }
             else
             {
                pHw->pUrbReadIso[ INCOMING_Q_SIZE-1 ]->next = pUrb;
                ZAP(DEBUG_READ, DEBUG_ATTN, (kTAB"URB %d(%p) links to URB %d(%p)\n",
               INCOMING_Q_SIZE-1, pHw->pUrbReadIso[ INCOMING_Q_SIZE-1 ], i, pUrb));
              }
        }

        for (i=0; i<INCOMING_Q_SIZE; i++)
        {
            /*Ok, data is initialized, fire off the ISO URB ring*/
            if ((result = usb_submit_urb(pHw->pUrbReadIso[i])))
               ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error %x on ISO read URB submit.\n", result));
        }
#endif

        ZAP(DEBUG_READ, DEBUG_EXIT, ("---StartReadPipe\n"));
    }
}

/**********************************************************************************/
/* ReadBulkCompletion								  */
/*										  */
/* Called when incoming BULK data is ready for our consumption. We check a few    */
/* conditions to make sure all is well with the state of things, then we	  */	
/* let the ATM code reassemble the cells into an ethernet packet. The last	  */
/* thing we'll do is queue up another BULK IN urb, to get the next packet.	  */	
/**********************************************************************************/
static void ReadBulkCompletion(struct urb *urb)
{
    Hardware *pHw = urb->context;
    struct net_device *ether;
    int result;

    ZAP(DEBUG_READ, DEBUG_ENTRY, ("+++ReadBulkCompletion\n"));

    /*This should never happen - but in a kernel driver, better safe than sorry!*/
    if (!pHw)
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"NULL pHw from urb->context!\n"));
       goto Read_exit;
    }

    /* If we're not open for business, ignore this callback*/
    if (!pHw->IsOpen)
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Read callback, but IsOpen = FALSE!\n"));
       goto Read_exit;
    }

    ether = pHw->pLinuxNet;
    if (!netif_device_present(ether))
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Somebody has killed our network interface.\n"));
       goto Read_exit;
    }

#ifdef SYNC_READS
    /***************************************************************************/
    /* We only allow one read to happen at a time. Since we only queue up one  */
    /* BULK IN urb, I don't know how we could get called again. But, we do need*/
    /* to make sure that we're not being reentered, just in case.	       */	 
    /***************************************************************************/
    if (pHw->IsReading)
    {
       pHw->LinuxStats.rx_errors++;
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Read callback, but IsReading = TRUE!\n"));
       goto Read_exit;
    }

    pHw->IsReading = TRUE;
#endif

    if (urb->status != USB_ST_NOERROR)
    {
       /* These errors seem to be what we get when the device has been unplugged, so*/
       /* lets stop a bunch of these from happening by unlinking the urb now        */
        if ((urb->status != -EILSEQ) &&
            (urb->status != -ETIMEDOUT))
        {
           ZAP(DEBUG_READ, DEBUG_ERRORS,
               (kTAB"Error status = %x in ReadBulkCompletion - requeuing\n", urb->status));
           goto ReadResubmit;        
        }
        else
        {
           ZAP(DEBUG_READ, DEBUG_ERRORS,
               (kTAB"Error status = %x in ReadBulkCompletion - not requeuing\n", urb->status));
           goto Read_exit;        
        }
    }

    /*If this was a zero-length read, we're done*/
    if (!urb->actual_length)
       goto ReadResubmit;

    ZAP(DEBUG_READ, DEBUG_ATTN, (kTAB"Read data length is %#x bytes.\n", urb->actual_length));

    /*Ok, we're ready to deal with the incoming data*/
    result = UniProcessInboundData(pHw, urb->transfer_buffer, urb->actual_length);
    if (result)
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error %x from UniProcessInboundData\n", result));

ReadResubmit:
    if (pHw->AdiModemSm.CurrentAdiState & STATE_OPERATIONAL )
    {
       FILL_BULK_URB(urb, pHw->usbdev, pHw->pipeBulkDataIn, 
                     urb->transfer_buffer, INCOMING_DATA_SIZE, ReadBulkCompletion, pHw);
       if ((result = usb_submit_urb(urb)))
          ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error %x on read URB submit.\n", result));
     }

#ifdef SYNC_READS
   pHw->IsReading = FALSE;
#endif

/*We will only jump to here on conditions where we don't want to queue another read*/
Read_exit:
    ZAP(DEBUG_READ, DEBUG_EXIT, ("---ReadBulkCompletion\n"));
    return;
}

/**********************************************************************************/
/* ReadIsoCompletion								  */
/*										  */
/* Called when incoming ISO data is ready for our consumption. We check a few     */
/* conditions to make sure all is well with the state of things, then we	  */
/* let the ATM code reassemble the cells into an ethernet packet. We don't	  */
/* have to re-submit the URB because it will get automatically done by the	  */
/* URB linking logic in the usb driver						  */
/**********************************************************************************/
static void ReadIsocCompletion(struct urb *urb)
{
    Hardware *pHw = urb->context;
    struct net_device *ether;
    int result, i;
    UInt8 *pBuf;

    ZAP(DEBUG_READ, DEBUG_ENTRY, ("+++ReadIsocCompletion\n"));

    /*This should never happen - but in a kernel driver, better safe than sorry!*/
    if (!pHw)
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"NULL pHw from urb->context!\n"));
       goto ReadIso_exit;
    }

    /*If we're not open for business, ignore this callback*/
    if (!pHw->IsOpen)
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Read callback, but IsOpen = FALSE!\n"));
       goto ReadIso_exit;
    }

    ether = pHw->pLinuxNet;
    if (!netif_device_present(ether))
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Somebody has killed our network interface.\n"));
       goto ReadIso_exit;
    }

    if (urb->status == -ENOENT)
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error status = %x in ReadIsocCompletion\n", urb->status));
       goto ReadIso_exit;
    }

    /*If this was a zero-length read, we're done*/
    if (!urb->actual_length)
       goto ReadIso_exit;

#ifdef ZAPS
    if (urb->status != USB_ST_NOERROR)
    {
       ZAP(DEBUG_READ, DEBUG_ERRORS,
           (kTAB"Error status = %x in ReadIsocCompletion - still processing %x bytes\n",
            urb->status, urb->actual_length));
    }
    else
    {
       ZAP(DEBUG_READ, DEBUG_INFO,
           (kTAB"Successful iso read - data length is %#x bytes.\n", urb->actual_length));
    }
#endif

    /*Ok, we're ready to deal with the incoming data*/
    pBuf = urb->transfer_buffer;
    for (i=0; i<urb->number_of_packets; i++)
    {
        if (!urb->iso_frame_desc[i].status)
        {
           result = UniProcessInboundData(pHw, pBuf+urb->iso_frame_desc[i].offset,
                                          urb->iso_frame_desc[i].actual_length);
          if (result)
             ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error %x from UniProcessInboundData\n", result));
        }
    }

ReadIso_exit:
    ZAP(DEBUG_READ, DEBUG_EXIT, ("---ReadIsocCompletion\n"));
    return;
}


/**********************************************************************************/
/* StopReadPipe									  */
/*										  */
/* Simple, really, but maybe we'll eventually have to do more.			  */	
/**********************************************************************************/
void StopReadPipe(Hardware *pHw)
{
    int i;

    ZAP(DEBUG_READ, DEBUG_ENTRY, ("+++StopReadPipe\n"));

    for (i=0; i<INCOMING_Q_SIZE; i++)
    {
#ifdef USEBULK
        usb_unlink_urb(&pHw->urbRead[i]);
#else
        usb_unlink_urb(pHw->pUrbReadIso[i]);
#endif
    }

    ZAP(DEBUG_READ, DEBUG_EXIT, ("---StopReadPipe\n"));
}

/**********************************************************************************/
/* ReadSendItUp									  */	
/*										  */
/* This is called by the Uni code once it has determined that a valid		  */
/* ethernet packet has been reassembled and is ready for processing		  */
/* by higher level network code. First we'll do things that a normal		  */
/* ethernet card would do in hardware (packet filtering, etc.), then		  */
/* we'll follow the standard procedure for sending data back up the stack.	  */
/**********************************************************************************/ 
void ReadSendItUp(Hardware *pHw, ETHERNET_PACKET_BUFFER *pPacket)
{

    /***Added to support PPPOA***/
    void * DataForUpperLayer=NULL;
    int RealSize=0;
    unsigned long TempSize=0;
    /****************************/

    ZAP(DEBUG_READ, DEBUG_ENTRY, ("+++ReadSendItUp\n"));

    MpoaProcessReceivedPdu(pHw, pPacket);
    if (pPacket->DropThisPacket != TRUE)
    {
       struct sk_buff *skb;
       switch (pHw->MpoaMode)
       {
          case MPOA_MODE_ROUTED_IP_LLC  :	
          case MPOA_MODE_ROUTED_IP_VC   :	

			ZAP(DEBUG_FFD, DEBUG_ENTRY, ("*** ReadSendItUp: ROUTED : setting Real size\n"));

			RealSize = pPacket->PacketSize - pPacket->EncapSkipSize + sizeof(ETHERNET_MAC_HEADER);
			TempSize=RealSize;
			DataForUpperLayer=GET_KBUFFER(ETH_DATA_LEN+sizeof(ETHERNET_MAC_HEADER));
			memcpy(DataForUpperLayer+0,pHw->MAC,ETH_LENGTH_OF_ADDRESS);
			memcpy(DataForUpperLayer+ETH_LENGTH_OF_ADDRESS,pHw->MAC,ETH_LENGTH_OF_ADDRESS);
			*((UInt8 *)DataForUpperLayer+11) += 1; // make the src & dst different
			memset(DataForUpperLayer+12,0x08,1); /*ICMP packet*/
			memset(DataForUpperLayer+13,0x00,1); /*ICMP packet*/
			memcpy(DataForUpperLayer+14,pHw->pReadyData+pPacket->EncapSkipSize,RealSize-sizeof(ETHERNET_MAC_HEADER));

			  break;
          case MPOA_MODE_BRIDGED_ETH_LLC:
          case MPOA_MODE_BRIDGED_ETH_VC :
             RealSize = pPacket->PacketSize - pPacket->EncapSkipSize;
             break;
          case MPOA_MODE_PPPOA_LLC      :
          case MPOA_MODE_PPPOA_VC       : 
             RealSize = pPacket->PacketSize - pPacket->EncapSkipSize + 14 + 2;
             TempSize=RealSize;
             DataForUpperLayer=GET_KBUFFER(ETH_DATA_LEN+14+2);
             memset(DataForUpperLayer+0,0x00,1);
             memset(DataForUpperLayer+1,0x01,1);
	     memset(DataForUpperLayer+2,0x02,1);
	     memset(DataForUpperLayer+3,0x03,1);
	     memset(DataForUpperLayer+4,0x04,1);
	     memset(DataForUpperLayer+5,0x05,1);
  	     memset(DataForUpperLayer+6,0xff,1);
	     memset(DataForUpperLayer+7,0xff,1);
	     memset(DataForUpperLayer+8,0xff,1);
	     memset(DataForUpperLayer+9,0xff,1);
	     memset(DataForUpperLayer+10,0xff,1);
	     memset(DataForUpperLayer+11,0xff,1);

	     memset(DataForUpperLayer+12,0x08,1); /*ICMP packet*/
	     memset(DataForUpperLayer+13,0x00,1); /*ICMP packet*/
	     memset(DataForUpperLayer+14,((TempSize&0xFF00)>>8),1);
	     memset(DataForUpperLayer+15,(TempSize&0xFF),1);
	     memcpy(DataForUpperLayer+16,pHw->pReadyData+pPacket->EncapSkipSize,RealSize-16);
             break;
        }

        /*Ok, we've got the green light to send the packet on up*/
        //if (!(skb = dev_alloc_skb(RealSize)))
        if (!(skb = dev_alloc_skb(ETH_DATA_LEN+14+2)))
        {
           ZAP(DEBUG_READ, DEBUG_ERRORS, (kTAB"Error allocating skb!\n"));
           goto SendItUp_exit;
        }

        skb->dev = pHw->pLinuxNet;

        switch (pHw->MpoaMode)
        {
           case MPOA_MODE_ROUTED_IP_LLC  :	
           case MPOA_MODE_ROUTED_IP_VC   :	
				ZAP(DEBUG_FFD, DEBUG_ENTRY, ("ReadSendItUp: ROUTED : CALLING eth_copy_and_sum\n"));
				eth_copy_and_sum(skb, DataForUpperLayer, RealSize, 0);
				FREE_KBUFFER(DataForUpperLayer);
			   break;
           case MPOA_MODE_BRIDGED_ETH_LLC:
           case MPOA_MODE_BRIDGED_ETH_VC :
              eth_copy_and_sum(skb, pHw->pReadyData+pPacket->EncapSkipSize, RealSize, 0);
              break;
           case MPOA_MODE_PPPOA_LLC      :
           case MPOA_MODE_PPPOA_VC       :
              eth_copy_and_sum(skb, DataForUpperLayer, RealSize, 0);
              FREE_KBUFFER(DataForUpperLayer);
              break;
        } 

        skb_put(skb, RealSize);
        skb->protocol = eth_type_trans(skb, pHw->pLinuxNet);
        netif_rx(skb);

        pHw->LinuxStats.rx_packets++;
        pHw->LinuxStats.rx_bytes += RealSize;
    }

SendItUp_exit:
    ZAP(DEBUG_READ, DEBUG_EXIT, ("---ReadSendItUp\n"));

    return;
}


/**********************************************************************************/ 
/* WriteAtmData									  */	
/*										  */
/* Another simple routine, just sends the ATM cells on their merry way.		  */
/**********************************************************************************/ 
int WriteAtmData(Hardware *pHw)
{
    int result;

    if (pHw->AdiModemSm.CurrentAdiState & STATE_OPERATIONAL )
    {
       ZAP(DEBUG_WRITE, DEBUG_INFO,
           ("*** Writing %#x bytes to USB\n", pHw->SegmentationBuffer.TotalBytes));

       FILL_BULK_URB(&pHw->urbWrite, pHw->usbdev, pHw->pipeBulkDataOut,
                     pHw->pOutgoingData, pHw->SegmentationBuffer.TotalBytes, 
                     WriteCompletion, pHw);
       result = usb_submit_urb(&pHw->urbWrite);
    }
    else
    {
       result=1; 
     }

     return result;
}


/**********************************************************************************/ 
/* WriteCompletion								  */
/*										  */
/* Called when the usb write of outgoing ATM cells is complete. 		  */	
/* Since we do not do any internal queueing of outgoing data, the		  */
/* main purpose for this routine is to wake up the kernel's outgoing		  */
/* network queue so we get another packet to send out (if anybody has one).	  */	
/**********************************************************************************/ 
static void WriteCompletion(struct urb *urb)
{
    Hardware *pHw = urb->context;
    struct net_device *ether;

    ZAP(DEBUG_WRITE, DEBUG_ENTRY, ("+++WriteCompletion\n"));

    if ((pHw == NULL) || (!pHw->IsOpen))
    {
       ZAP(DEBUG_WRITE, DEBUG_ERRORS,
           (kTAB"NULL urb->context or IsOpen=FALSE in WriteCompletion!\n"));
       goto WriteCompletion_exit;
    }

    ether = pHw->pLinuxNet;

    if (!netif_device_present(ether))
    {
       ZAP(DEBUG_WRITE, DEBUG_ERRORS, (kTAB"Network interface no longer present!\n"));
       goto WriteCompletion_exit;
    }

    if (urb->status)
       ZAP(DEBUG_WRITE, DEBUG_ERRORS, (kTAB"urb has status of %x\n", urb->status));

    ether->trans_start = jiffies;
    netif_wake_queue(ether);

WriteCompletion_exit:
    ZAP(DEBUG_WRITE, DEBUG_EXIT, ("---WriteCompletion\n"));
}
