/*********************************************************************************/
/*										 */
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved			 */
/*										 */
/* Config.c									 */
/*										 */
/* Various routines that implement a minimal ini-style configuration file system */
/*										 */
/*********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "Zaps.h"
#include "Stuff.h"
#include "Konst.h"
#include "Options.h"
#include "AdiUsbAdslDriver.h"

/*********************************************************************************/
/* Local prototypes								 */
/*********************************************************************************/

static Boolean ConfigGetEntryText(char *pOption, char *pValue, FILE *fp);
static Boolean ConfigGetEntryInt(char *pOption, UInt32 *pValue, FILE *fp);
static Boolean ConfigOptionExists(char *srcBuffer, char *Option);
static Boolean ConfigExtractValue(char *srcBuffer, char *Destination);
static Boolean ConfigTrim(char *str);
static UInt8 htoi(UInt8 h);
static void BufferToHex(char *buffer, UInt32 *Value);
static void Capitalize(char *src_buffer);

#define PREAMBLE                0x535C
#define RECEIVER_MODEM          0x0000
#define SENDER_HOST             0x0010
#define SUBTYPE_MEM_REQWRITE    0x0100
#define TYPE_MEMACCESS          0x1000
#define WRITE_TO_MODEM (RECEIVER_MODEM + SENDER_HOST + SUBTYPE_MEM_REQWRITE + TYPE_MEMACCESS)
/*************************************************************/

/*********************************************************************************/
/* ConfigGetEntryText								 */
/*										 */ 
/* Given a specified option, returns its value if it exists in the		 */
/* config file.									 */
/*********************************************************************************/
static Boolean ConfigGetEntryText(char *pOption, char *pValue, FILE *fp)
{
    Boolean rv = false;
    char buffer[256];

    /* First we need to see if this key already exists*/
    fseek(fp, 0, SEEK_SET);

    /* Search the entire file*/
    while ( feof(fp) == 0 )
    {
       /* Read a line from the file*/
       if ( fgets(buffer, 256, fp) )
       {
          /* Does this line have our option*/
          if ( ConfigOptionExists(buffer, pOption) )
          {
             if ( ConfigExtractValue(buffer, pValue) )
             {
                ConfigTrim(pValue);
                rv = true;
             }
             break;
          }
       }
    }
    return ( rv );
}

/*********************************************************************************/
/* ConfigOptionExists								 */
/*										 */
/* Returns TRUE if the specified Option exists in the config file		 */
/*********************************************************************************/
static Boolean ConfigOptionExists(char *srcBuffer, char *Option)
{
    Boolean rv = false;
    char buffer[256];
    char *pChar;

    /* We copy the buffer because wo modify it*/
    strcpy(buffer, srcBuffer);

    /* Look for the terminator*/
    pChar = strtok(buffer, "=");
    /* Is it there ?*/
    if ( pChar )
    {
       /* Copy the buffer*/
       strcpy(buffer, pChar);
       
       /* Remove Leading and trailing whitespace*/
       ConfigTrim(buffer);
        
       /* Yep, get the next piece*/
       if ( strcmp(buffer, Option) == 0 )
       {
          rv = true;
       }
    }
    return ( rv );
}

/*********************************************************************************/
/* ConfigExtractValue								 */
/*										 */
/* Given an 'option=value' item in a string, extracts the value portion		 */
/* and stores it in the Destination param					 */
/*********************************************************************************/
static Boolean ConfigExtractValue(char *srcBuffer, char *Destination)
{
    Boolean rv = false;
    char buffer[256];
    char *pChar;

    /* We copy the buffer because wo modify it*/
    strcpy(buffer, srcBuffer);

    /* Look for the terminator*/
    pChar = strtok(buffer, "=");
    /* Is it there ?*/
    if ( pChar )
    {
       /* Yep, get the next piece*/
       pChar = strtok(NULL, "\n\r");
       /* Is it there*/
       if ( pChar )
       {
          /* Copy the string*/
          strcpy(Destination, pChar);
          /* And trim*/
          ConfigTrim(Destination);
          /* Set rv*/
          rv = true;
       }
    }
    return ( rv );
}

/*********************************************************************************/
/* ConfigTrim									 */
/*										 */
/* Trims excess whitespace from the beginning and end of the specified string    */
/*********************************************************************************/
static Boolean ConfigTrim(char *str)
{
    Boolean rv = true;
    char *pChar = str + strlen(str) - 1;
    char buffer[256];
    int count = 0;
    
    /* remove trailing white space first*/
    while ( isspace(*pChar) )
    {
       *pChar-- = '\0';
    }
    
    /* remove leading whitespace*/
    pChar = str;
    while ( isspace(*pChar) )
    {
       count ++;
       pChar ++;
    }
    
    /* Copy over the source string without leading space*/
    strcpy(buffer, str + count);
    /* and then copy it back into our buffer*/
    strcpy(str, buffer); 

    return ( rv );
}

/*********************************************************************************/
/* ConfigGetOptions								 */
/*										 */
/* Reads the array of supplied DriverOptions values from the configuration file  */ 
/*********************************************************************************/
void ConfigGetOptions( void )
{
    int loop = 0;
    DriverOption *pOptions;
    FILE *fp;

    /* Open it*/
    if ((fp = fopen(kCONFIG_FILENAME, "rt")))
    {
       pOptions = drvOptions;

       ZAP(DEBUG_CONFIG, DEBUG_INFO, ("Opened it\n"));

       /* For each DriverOption*/
       for ( loop = 0 ; loop < NUM_DRV_OPTIONS ; loop ++ )
       {
           ZAP(DEBUG_CONFIG, DEBUG_INFO, ("Trying to read %s\n",pOptions[loop].Name ));

           /* Try and read the data */               
           if ( ! ConfigGetEntryInt(pOptions[loop].Name, &pOptions[loop].Value, fp) )
           {
              pOptions[loop].Value = 0xF9F8F7F6;
              ZAP(DEBUG_CONFIG, DEBUG_INFO, ("config file could not read %s\n", pOptions[loop].Name));
           }
           else
           {
              ZAP(DEBUG_CONFIG, DEBUG_INFO,
                  ("Option %s assigned value %#x\n", pOptions[loop].Name, pOptions[loop].Value));
           }
       }
       fclose(fp);
    }
    else
       ZAP(DEBUG_CONFIG, DEBUG_ERRORS, (kTAB"Config file %s does not exist\n", kCONFIG_FILENAME));
}   

/*********************************************************************************/
/* ConfigGetEntryInt								 */
/*										 */
/* Reads a value for the specified option from the config file			 */
/*********************************************************************************/
static Boolean ConfigGetEntryInt(char *pOption, UInt32 *pValue, FILE *fp)
{
    char buffer[256];
    Boolean rv = false;
   
    int i;
    char b[4];
    char TmpBuffer[20];
    int TmpLen=0;
    int ii=0;
    int jj=0;
    int kk=0;
    char Temp[40];

    for (ii=0;ii<20;ii++)
        TmpBuffer[ii]='0';    

    /* Use the text code*/
    if ( ConfigGetEntryText(pOption, buffer, fp) )
    {
       strcpy (Temp, pOption);
       Capitalize (Temp);
       if (strcmp(Temp, "LINETYPE")==0)
       {
          *pValue=atoi(buffer);
          rv = true;
          return rv;
       }
       
       TmpLen=strlen(buffer);
       for (ii=0;ii<8-TmpLen;ii++)
       {
           jj++;
           TmpBuffer[ii]='0';
       }
       kk=0;
       for (ii=jj;ii<8;ii++)
       {
           TmpBuffer[ii]=buffer[kk];
           kk++;
       }

       for (i=0; i<4; i++)
       {
           b[3-i] = (htoi(TmpBuffer[i*2])<<4) + (htoi(TmpBuffer[i*2+1]));
       }

       *pValue = *((UInt32 *)b);

       /* set rv*/
       rv = true;
    }
        
    return ( rv );      
}

/*********************************************************************************/
/* htoi										 */
/*										 */
/* Converts a hex char to an integer (my own wacky implementation).		 */
/* I don't know why this isn't part of a standard library or something,		 */
/* it seems like this is a quite-frequently-rewritten function.			 */
/*********************************************************************************/
static UInt8 htoi(UInt8 h)
{
    switch (toupper(h))
    {
       case '0':
       case '1':
       case '2':
       case '3':
       case '4':
       case '5':
       case '6':
       case '7':
       case '8':
       case '9':
          return h-'0';
          break;
       case 'A':
       case 'B':
       case 'C':
       case 'D':
       case 'E':
       case 'F':
          return h-'A'+10;
          break;
    }

    /* If we get here ... the character was invalid*/
    return 0;
}

static void Capitalize(char *src_buffer)
{
   int i,len;
   len = strlen(src_buffer);
   for (i=0;i<len;++i)
       src_buffer[i]= (char)toupper(src_buffer[i]);
}

/*********************************************************************************/
/* Turn a char * to a hex							 */
/*										 */
/*********************************************************************************/
static void BufferToHex(char *src_buffer, UInt32 *Value)
{
    int i;
    char b[4];
    char TmpBuffer[20];
    int TmpLen=0;
    int ii=0;
    int jj=0;
    int kk=0;
    char buffer[20];

    for (ii=0;ii<20;ii++)
        TmpBuffer[ii]='0';  
  
    if ((src_buffer[0]=='0') && ((src_buffer[1]=='x') || (src_buffer[1]=='X')))
    {
       strcpy(buffer,src_buffer+2);
    }
    else
    {
       strcpy(buffer,src_buffer);
    }

    TmpLen=strlen(buffer);
    for (ii=0;ii<8-TmpLen;ii++)
    {
        jj++;
        TmpBuffer[ii]='0';
    }
    kk=0;
    for (ii=jj;ii<8;ii++)
    {
        TmpBuffer[ii]=buffer[kk];
        kk++;
    }

    for (i=0; i<4; i++)
    {
        b[3-i] = (htoi(TmpBuffer[i*2])<<4) + (htoi(TmpBuffer[i*2+1]));
    }

    *Value = *((UInt32 *)b);
}

/**********************************************************************************/
/* GetCMVFromFile		    						  */
/*										  */
/* Reads the predefined CMVs from a text file for POTS or ISDN depends on type of */
/* modem									  */ 
/**********************************************************************************/
void GetCMVFromFile( void )
{
    Boolean RetVal=TRUE;
    FILE *fp=NULL;
    char Tmp_Buffer[256];
    char Src_Buffer[256];
    char Temp[3];
    char *pChar;
    int ii=0,jj=0;
    char TmpBuffer[20];
    CmvMsg *pCMVs;

    pCMVs = drvCMVs;

    for (ii=0;ii<20;ii++)
        TmpBuffer[ii]='0';    

    for (ii=0;ii<100;ii++)
    {
    	pCMVs [ii].Preamble = PREAMBLE;
        pCMVs [ii].Function = WRITE_TO_MODEM;
        pCMVs [ii].Index = 0x1234+ii;
        pCMVs [ii].SymAddrHiHi = 'O';
        pCMVs [ii].SymAddrHiLo = 'P';
        pCMVs [ii].SymAddrLoHi = 'T';
        pCMVs [ii].SymAddrLoLo = 'N';
        pCMVs [ii].Offset = 0;
        pCMVs [ii].Data = 0xF9F9F9F9;
    }

    switch (ModemType)
    {
       case MODEM_POTS  :
          if (!(fp = fopen(kPOTS_CMV_FILENAME, "rt")))
	  {
     	     RetVal = FALSE;
	     ZAP(DEBUG_CONFIG, DEBUG_ERRORS, (kTAB"CMV file %s does not exist\n", kPOTS_CMV_FILENAME));
          }
          break;
       case MODEM_ISDN :
          if (!(fp = fopen(kISDN_CMV_FILENAME, "rt")))
	  {
  	     RetVal = FALSE;
	     ZAP(DEBUG_CONFIG, DEBUG_ERRORS, (kTAB"CMV file %s does not exist\n", kISDN_CMV_FILENAME));
          }
          break;
    } 
    

    if (!RetVal)
    {
       return;
    }

    while (feof(fp) == 0)
    {
       if (!fgets(Src_Buffer, 256, fp))
       {
          continue;
       }
       strcpy(Tmp_Buffer, Src_Buffer);
       ConfigTrim(Tmp_Buffer);

       if (strncmp(Tmp_Buffer, ";", 1) == 0 )
       {
          continue;
       }     
       strncpy (Temp, Tmp_Buffer, 2);
       Capitalize(Temp);
       if (strncmp(Temp, "CW", 2) != 0 )
       {
          continue;
       }
     
       pChar = strtok(Tmp_Buffer, " ");
       if ( pChar )
       {
          pChar = strtok(NULL, " ");
          if ( pChar )
          {
             strcpy(Tmp_Buffer, pChar);
             ConfigTrim(Tmp_Buffer);
	     if (strlen(Tmp_Buffer) > 4)
             {
                continue;
	     }
	     else
             {
	        pCMVs [jj].SymAddrHiHi = toupper(Tmp_Buffer[0]);
	        pCMVs [jj].SymAddrHiLo = toupper(Tmp_Buffer[1]);
	        pCMVs [jj].SymAddrLoHi = toupper(Tmp_Buffer[2]);
	        pCMVs [jj].SymAddrLoLo = toupper(Tmp_Buffer[3]);

      	        strcpy(Tmp_Buffer, Src_Buffer);
                pChar = strtok(Tmp_Buffer, " ");
                if ( pChar )
                {
                   pChar = strtok(NULL, " ");
                   if ( pChar )
                   {
                      pChar = strtok(NULL, " ");
                      if ( pChar )
                      {
                         strcpy(Tmp_Buffer, pChar);
                         ConfigTrim(Tmp_Buffer);
                         pCMVs [jj].Offset = atoi(Tmp_Buffer);
      	                 strcpy(Tmp_Buffer, Src_Buffer);
                         pChar = strtok(Tmp_Buffer, " ");
                         if ( pChar )
                         {
		            pChar = strtok(NULL, " ");
                            if ( pChar )
		            {
		               pChar = strtok(NULL, " ");
                               if ( pChar )
			       {
                                  pChar = strtok(NULL, "\n\r ");
                                  if ( pChar )
                                  {
                                     strcpy(Tmp_Buffer, pChar);
                                     ConfigTrim(Tmp_Buffer);
                                     //BufferToHex(Tmp_Buffer, &pCMVs[jj].Data);
				     sscanf(Tmp_Buffer,"%x", &pCMVs[jj].Data); // FFD090903_2 : bug-fix
			             jj++;
                                  }
                                  else
                                  {
                                     continue;
                                  }  
                               }
                               else
                               {
                                  continue;
                               }
                            }
                            else
                            {
                               continue;
                            }
                         }
     		         else
		         {
                            continue;
                         }
                      }
                      else
                      {
                         continue;
                      }
                   }
		   else
                   {
                      continue;
                   }
                }
	        else
	        {
	           continue;
                }
             }
          }
          else
          {
             continue;
          }
       }
       else
       {
          continue;
       }
    }
    fclose(fp);
}

