/*****************************************************************************/
/*									     */
/* Copyright (c) 2002, Analog Devices Inc., All Rights Reserved		     */
/*									     */	
/* main.c								     */	
/*									     */
/* Main code for usermode process to manage Linux kernel driver.	     */	
/*									     */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>      /* open */
#include <sys/ioctl.h>  /* ioctl */
#include <unistd.h>     /* close */
#include <linux/usbdevice_fs.h>
#include <limits.h>     /* PATH_MAX */
#include <dirent.h>     /* DIR */

#include "Zaps.h"
#include "Stuff.h"
#include "Konst.h"
#include "AdiUsbAdslDriver.h"

static UInt32           DspSize;
static int              drv;        // file handle to usbdevfs

extern void ConfigGetOptions( void );
extern DriverOption drvOptions[];

extern CmvMsg drvCMVs[100];
extern void GetCMVFromFile(void);


/*****************************************************************************/
/* htoi									     */	
/*									     */
/* From NDIS code							     */	
/*****************************************************************************/
static UInt32 htoi(UInt8 *h, UInt32 l)
{

  UInt32 num, i;

  for (num = 0, i=0; i<l && h[i]!=0; i++)
  {
      num<<=4;
      if (h[i]<='9' && h[i]>='0')
          num += h[i] - '0';
      else if (h[i]<='F' && h[i]>='A')
              num += h[i] - 'A' + 10;
           else if (h[i]<='f' && h[i]>='a')
                   num += h[i] - 'a' + 10;
  }

  return num;
}

/*****************************************************************************/
/* ConvertSRecordsToBinary						     */
/*									     */	
/* Converts Motorolla SRecords read from .BNM files into a binary format     */
/* Assumes that pDst was zero'd at alloc time.				     */	
/*****************************************************************************/
static Boolean ConvertSRecordsToBinary( UInt8 *pDst, Firmware *pFw )
{
    UInt8 *pNext;
    UInt8 *pSrc;
    UInt32 i, len, ofs;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("-->ConvertSRecordsToBinary: pDst=%x, pFw=%x\n", (UInt32)pDst, (UInt32)pFw));

    pNext = pDst;
    pSrc  = pFw->pBits;

    /* Keep processing as long as there is data to be processed*/
    while ((UInt32)(pSrc-pFw->pBits)<pFw->Size)
    {
       if (*pSrc != 's' && *pSrc != 'S')
       {
          ZAP(DEBUG_INIT, DEBUG_ERRORS, ("SRecord does not begin with S!\n"));
          return FALSE;
       }

       pSrc++; // skip over the S

       if (*pSrc == '9')
          break; // this indicates the end of file

       if (*pSrc == '2')
       {
          pSrc++;
          len = htoi(pSrc, 2) - 4;
          pSrc+=2;
          ofs = htoi(pSrc, 6);
          pSrc+=6;
       }
       else
       {
          if (*pSrc == '3')
          {
             pSrc++;
             len = htoi(pSrc, 2) - 5;
             pSrc+=2;
             ofs = htoi(pSrc, 8);
             pSrc+=8;
          }
          else
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, ("SRecord with wrong format!\n"));
             return FALSE;
          }
       }
       for (i=0; i<len; i++)
           *(pDst+ofs+i) = (UInt8)htoi(pSrc+i*2, 2);
       pSrc += (len*2+4);
       /* Track furthest extent written to so driver can vmalloc exact size*/
       if (ofs+len > DspSize)
          DspSize = ofs+len;
    }

    ZAP(DEBUG_INIT, DEBUG_EXIT, ("<--ConvertSRecordsToBinary\n"));
    return TRUE;
}

/*****************************************************************************/
/* ProcessADSLFirmware							     */	
/*									     */
/* The DSP code is in motorolla S-record format. We need to process it into  */ 
/* blocks and pages so we can intelligently send it to the DSP.		     */
/*****************************************************************************/
static OSStatus ProcessADSLFirmware( Hardware *pHw )
{
    UInt32 i;
    UInt8 *pBuf;
    int ret_val;
    struct usbdevfs_ioctl ioctl_wrapper;
    struct adi_dsp_info dsp_info;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("-->ProcessADSLFirmware:\n"));

    /****************************************************************/
    /* If we get here, then we know all of the firwmare files	    */
    /* were downloaded successfully. We need to process the	    */
    /* ADSL firmware data into MAIN and SWAP page records	    */
    /* We'll start be allocating permanent storage for the DSP code */
    /*								    */
    /* NDIS code just allocates 1MB of memory, so we'll do that too.*/ 
    /* However, we could probably allocate a LOT less if we do some */
    /* instrumentation or initial examination of the .BNM files     */
    /****************************************************************/
    pBuf = (UInt8 *)GET_MEMORY(1024*1024);
    IFNULL_RETURN_MEMERR(pBuf, "Not enough memory for entire DSP code.\n");
    pHw->pDsp = pBuf;

    /* Before processing, read some IDMA info from the first buffer.*/
    if (*(pHw->ADSLFirmware[0].pBits+1) == '2') // does file start with "S2"
       pHw->IDMAStart = htoi(pHw->ADSLFirmware[0].pBits+4, 6);
    else
       if (*(pHw->ADSLFirmware[0].pBits+1) == '3') // or "S3"
          pHw->IDMAStart = htoi(pHw->ADSLFirmware[0].pBits+4, 8);

    ZAP(DEBUG_INIT, DEBUG_INFO, ("Begin DSP processing. IDMAStart=%x\n", pHw->IDMAStart));

    DspSize = 0;
    /*******************************************************************/ 
    /* Loop through the firmware data that was read in from .BNM files,*/
    /* converting each file to an internal binary format. As each file */
    /* is completely processed, we can free the memory alloced for the */
    /* file, since the data is now stored in the new binary buffer.    */
    /*******************************************************************/ 
    for (i=0; i<DSP_FIRMWARE_FILES; i++)
    {
        if (!ConvertSRecordsToBinary(pBuf, &(pHw->ADSLFirmware[i])))
        {
           ZAP(DEBUG_INIT, DEBUG_ERRORS, ("Failure in ConvertSRecordsToBinary!\n"));
           return ioErr;
        }
        FREE_BUFFER(pHw->ADSLFirmware[i].pBits);
        pHw->ADSLFirmware[i].pBits = NULL;
    }

    /**********************************************************************/
    /* Now the real processing begins ......				  */	
    /* Except that we've divided the processing and the kernel driver will*/
    /* take it from here.  Pass the binary data to the kernel driver      */
    /**********************************************************************/

    dsp_info.IDMAStart = pHw->IDMAStart;
    dsp_info.DspSize = DspSize;
    dsp_info.pDsp = pHw->pDsp;
    ioctl_wrapper.ifno = 1;
    ioctl_wrapper.ioctl_code = ADIUSBADSLDSP;
    ioctl_wrapper.data = &dsp_info;
    ret_val = ioctl(drv, USBDEVFS_IOCTL, &ioctl_wrapper);

    /* Free the big buffer we alloced above*/
    FREE_BUFFER(pBuf);

    if (ret_val < 0)
    {
       ZAP(DEBUG_INIT, DEBUG_ERRORS, ("Failed to d/l dsp code to driver, ret_val = %d\n", ret_val));
       return ioErr;
    }

    ZAP(DEBUG_INIT, DEBUG_INFO, ("End DSP processing. DspSize=%x\n", DspSize));

    return noErr;
}

/*****************************************************************************/
/* LoadFirmwareData							     */
/*									     */
/* Allocates space for and loads the firmware data from the specified file   */
/*****************************************************************************/
OSStatus LoadFirmwareData( const char *pFilename, Firmware *pFirmware )
{
    OSStatus ret = noErr; 
    long theSize, newSize;
    FILE *fp;

    /* First we need to get the file opened*/
    fp = fopen(pFilename, "rb");
    if ( fp )
    {
       /* File opened, now find out how big it is*/
       fseek(fp, 0L, SEEK_END);
       theSize = ftell(fp);
       fseek(fp, 0L, SEEK_SET);
       ZAP(DEBUG_INIT, DEBUG_INFO, ("File size for %s is %ld\n", pFilename, theSize));
                        
       if (theSize)
       {
          pFirmware->Size  = (UInt32)theSize;
          /* Allocate some memory for the file data*/
          pFirmware->pBits = (UInt8 *)GET_BUFFER(pFirmware->Size);
    
          if (pFirmware->pBits != NULL)
          {
             newSize = theSize;
             /* Finally, we can now read the file data!*/
             newSize = fread(pFirmware->pBits, 1, theSize, fp);
             /* We didn't get back the # of bytes we asked for!*/
             if (newSize != theSize)
             {
                ZAP(DEBUG_INIT, DEBUG_ERRORS,
                    ("\t\tRequested 0x%lX bytes from file, got 0x%lX bytes.", theSize, newSize));
                ret = ioErr;
             }
          }
          else
          {
             ret = ioErr;
             ZAP(DEBUG_INIT, DEBUG_ERRORS, ("Allocate for size %ld failed.\n", theSize));
          }
       }
       else 
       {
          ZAP(DEBUG_INIT, DEBUG_ERRORS, ("Get file size failed.\n"));
       }
       /* We opened the file, so we have to close it*/
       fclose(fp);
    }
    else
    {
       ret = ioErr;
       ZAP(DEBUG_INIT, DEBUG_ERRORS, ("fopen failed opening %s.\n", pFilename));    
    }

    return ret;
}

/*****************************************************************************/
/* LoadFirmwareFiles							     */	
/*									     */
/* Loads firwmare data into global storage				     */	
/*****************************************************************************/
OSStatus LoadFirmwareFiles( Hardware *pHw )
{
    OSStatus myErr=FALSE;
    UInt32 i;

    /* Make sure all pointers are set to NULL first*/
    for (i=0; i<DSP_FIRMWARE_FILES; i++)
        pHw->ADSLFirmware[i].pBits = NULL;
    
    switch (ModemType)
    {
       case MODEM_POTS:     
          myErr = LoadFirmwareData(CFILE_ADSL_POTS_FIRMWARE0, &(pHw->ADSLFirmware[0])); 
          if (!myErr)
          {
             myErr = LoadFirmwareData(CFILE_ADSL_POTS_FIRMWARE1, &(pHw->ADSLFirmware[1])); 
             if (!myErr)
             {
                myErr = LoadFirmwareData(CFILE_ADSL_POTS_FIRMWARE2, &(pHw->ADSLFirmware[2])); 
                if (!myErr)
                {
                   myErr = LoadFirmwareData(CFILE_ADSL_POTS_FIRMWARE3, &(pHw->ADSLFirmware[3])); 
                   if (!myErr)
                   {
                      myErr = LoadFirmwareData(CFILE_ADSL_POTS_FIRMWARE4, &(pHw->ADSLFirmware[4])); 
                      if (myErr)
                      {
                         ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_POTS_FIRMWARE4));
                      }
                      else
                      {
                         myErr = ProcessADSLFirmware( pHw );
                      }
                   }
                   else
                   {
                      ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_POTS_FIRMWARE3));
                   }
                }
                else
                {
                   ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_POTS_FIRMWARE2));
                }
             }
             else
             {
                ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_POTS_FIRMWARE1));
             }
          }
          else
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_POTS_FIRMWARE0));
          }
          break;
       case MODEM_ISDN:
          myErr = LoadFirmwareData(CFILE_ADSL_ISDN_FIRMWARE0, &(pHw->ADSLFirmware[0])); 
          if (!myErr)
          {
             myErr = LoadFirmwareData(CFILE_ADSL_ISDN_FIRMWARE1, &(pHw->ADSLFirmware[1])); 
             if (!myErr)
             {
                myErr = LoadFirmwareData(CFILE_ADSL_ISDN_FIRMWARE2, &(pHw->ADSLFirmware[2])); 
                if (!myErr)
                {
                   myErr = LoadFirmwareData(CFILE_ADSL_ISDN_FIRMWARE3, &(pHw->ADSLFirmware[3])); 
                   if (!myErr)
                   {
                      myErr = LoadFirmwareData(CFILE_ADSL_ISDN_FIRMWARE4, &(pHw->ADSLFirmware[4])); 
                      if (myErr)
                      {
                         ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_ISDN_FIRMWARE4));
                      }
                      else
                      {
                         myErr = ProcessADSLFirmware( pHw );
                      }
                   }
                   else
                   {
                      ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_ISDN_FIRMWARE3));
                   }
                }
                else
                {
                   ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_ISDN_FIRMWARE2));
                }
             }
             else
             {
                ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_ISDN_FIRMWARE1));
             }
          }
          else
          {
             ZAP(DEBUG_INIT, DEBUG_ERRORS, ("download of %s failed.\n", CFILE_ADSL_ISDN_FIRMWARE0));
          }
          break;
    }
    /************************************************************************/

    /* If there were any errors, free memory that was allocated*/
    if (myErr)
    {
       for (i=0; i<DSP_FIRMWARE_FILES; i++)
       {
           if (pHw->ADSLFirmware[i].pBits)
              FREE_BUFFER(pHw->ADSLFirmware[i].pBits);
       }
    }

    return myErr;
}

/*****************************************************************************/
/* ResetModem           					             */	
/*									     */
/* Resetes the modem							     */	
/*****************************************************************************/
static OSStatus ResetModem(void)
{
    int ret_val;
    struct usbdevfs_ioctl ioctl_wrapper;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++ResetModem\n"));

    ioctl_wrapper.ifno = 1;
    ioctl_wrapper.ioctl_code = ADIUSBADSLRESET;
    ioctl_wrapper.data = &ioctl_wrapper; //NULL;
    ret_val = ioctl(drv, USBDEVFS_IOCTL, &ioctl_wrapper);
   
    if (ret_val < 0)
    {
       ZAP(DEBUG_INIT, DEBUG_ERROR, ("Failed to reset driver, ret_val = %d\n", ret_val));
       return ioErr;
    }
    
    ZAP(DEBUG_INIT, DEBUG_INFO, ("---ResetModem \n"));
    
    return noErr;
}

/*****************************************************************************/
/* SendOptionsToDriver         					             */	
/*									     */
/* Sends options in adiusbadsl.conf to driver				     */	
/*****************************************************************************/
static OSStatus SendOptionsToDriver(void)
{
    int ret_val;
    struct usbdevfs_ioctl ioctl_wrapper;
    struct adi_dsp_info dsp_info;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++SendOptionsToDriver\n"));

    dsp_info.DspSize = 0;
    dsp_info.pDsp = (UInt8 *)&drvOptions;
    ioctl_wrapper.ifno = 1;
    ioctl_wrapper.ioctl_code = ADIUSBADSLOPTIONS;
    ioctl_wrapper.data = &dsp_info;
    ZAP(DEBUG_INIT, DEBUG_INFO, ("Sending pointer %p to ioctl\n", drvOptions));
    ret_val = ioctl(drv, USBDEVFS_IOCTL, &ioctl_wrapper);
   
    if (ret_val < 0)
    {
       ZAP(DEBUG_INIT, DEBUG_ERROR, ("Failed to send driver options, ret_val = %d\n", ret_val));
       return ioErr;
    }
    
    ZAP(DEBUG_INIT, DEBUG_INFO, ("---SendOptionsToDriver\n"));
    
    return noErr;
}

/*****************************************************************************/
/* SendCMVsToDriver         					             */	
/*									     */
/* Sends CMVs from either CMV text file to driver				     */	
/*****************************************************************************/
static OSStatus SendCMVsToDriver(void)
{
    int ret_val;
    struct usbdevfs_ioctl ioctl_wrapper;
    struct adi_dsp_info dsp_info;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++SendCMVsToDriver\n"));

    dsp_info.DspSize = 0;
    dsp_info.pDsp = (UInt8 *)&drvCMVs;
    ioctl_wrapper.ifno = 1;
    ioctl_wrapper.ioctl_code = ADIUSBADSLCMVS;
    ioctl_wrapper.data = &dsp_info;
    ZAP(DEBUG_INIT, DEBUG_INFO, ("Sending pointer %p to ioctl\n", drvCMVs));
    ret_val = ioctl(drv, USBDEVFS_IOCTL, &ioctl_wrapper);
   
    if (ret_val < 0)
    {
       ZAP(DEBUG_INIT, DEBUG_ERROR, ("Failed to send driver CMVs, ret_val = %d\n", ret_val));
       return ioErr;
    }
    
    ZAP(DEBUG_INIT, DEBUG_INFO, ("---SendCMVsToDriver\n"));
    
    return noErr;
}


/*****************************************************************************/
/* OpenDriver	         					             */	
/*									     */
/* Opens driver								     */	
/*****************************************************************************/
static Boolean OpenDriver( void ) 
{
    char rootname[PATH_MAX] = "";
    char busname[PATH_MAX] = "";
    char devname[PATH_MAX] = "";
    DIR *rootdir;
    DIR *busdir;
    struct dirent *rootentry;
    struct dirent *busentry;
    usb_device_descriptor descriptor;
    int bMatch = FALSE;
    int fd, ret;
    UInt32 TempbcdDevice;

    ZAP(DEBUG_INIT, DEBUG_ENTRY, ("+++OpenDriver\n"));

    sprintf(rootname, "/proc/bus/usb");

    /* open rootdir and iterate on contents */
    rootdir = opendir(rootname);
    while ((rootentry = readdir(rootdir)) != NULL)
    {
       /* ignore . and .. entries although it would work anyway */
       if (rootentry->d_name[0] == '.')
       {
          continue;
       }

       /* process each entry which is also a directory */
       sprintf(busname, "%s/%s", rootname, rootentry->d_name);
       busdir = opendir(busname);
       if (busdir == NULL)
       {
          continue;
       }

       while ((busentry = readdir(busdir)) != NULL)
       {
          /* This dir is a bus, and each entry a device */
          /* Open each device and check the descriptor */
          /* ignore . and .. entries although it would work anyway */
          if (busentry->d_name[0] == '.')
          {
             continue;
          }
          sprintf(devname, "%s/%s", busname, busentry->d_name);
          fd = open(devname, O_RDONLY);
          if (fd < 0)
          {
             continue;
          }

          ret = read(fd, &descriptor, sizeof(usb_device_descriptor));
          close(fd);
          if (ret < 0)
          {
             continue;
          }

          TempbcdDevice = descriptor.bcdDevice & 0x80;
          if (TempbcdDevice == 0x80)
	  {
             ModemType = MODEM_ISDN;
	  }
	  else
	  { 
             ModemType = MODEM_POTS;
	  }
          if ((descriptor.idVendor == kEAGLE_VID) &&
              ((descriptor.idProduct==kEAGLEI_PID_PSTFIRMWARE)||(descriptor.idProduct == kEAGLEII_PID_PSTFIRMWARE)||
               (descriptor.idProduct == kEAGLEIIC_PID_PSTFIRMWARE)))
          {
             /* terminate processing, we're done */
             bMatch = TRUE;
             break;
          }            
       }
       closedir(busdir);

       if (bMatch)
       {
           break;
       }
    }
    closedir(rootdir);

    /* see if we broke or blew through the end*/
    if (!bMatch) 
    {
       ZAP(DEBUG_INIT, DEBUG_ERRORS, ("Failed to find driver\n"));
       return FALSE;
    }

    drv = open(devname, O_RDWR);
    if (drv < 0)
    {
       ZAP(DEBUG_INIT, DEBUG_ERRORS, ("Failed to open driver\n"));
       return FALSE;
    }

    ZAP(DEBUG_INIT, DEBUG_ATTN, ("File handle=%x\n", drv));
    return TRUE;
}

int main(int argc, char **argv)
{
    Hardware Hw;
    int ret = noErr;

    if (!OpenDriver())
    {
       ret = ioErr;
       goto done;
    }

    GetCMVFromFile();
    

    /* consult options and conditionally do the following*/
    if (argc == 1)
    {
       ConfigGetOptions();
       SendCMVsToDriver();
       SendOptionsToDriver();
       ret = LoadFirmwareFiles(&Hw);
    }
    else
    {
       if (!strcmp("reset", argv[1]))
       {
           ret = ResetModem();
       }
       else
       {
          if (!strcmp("options", argv[1]))
          {
             /* Load the device options from /etc/analog/adiusbadsl.conf*/
             ConfigGetOptions();
  	     SendCMVsToDriver();
             SendOptionsToDriver();
          }
          else
          {
             ZAP(DEBUG_INIT, DEBUG_ATTN, ("Unknown option\n"));
             ret = ioErr;
          }
       }
    } 
    if (drv >= 0)
       close(drv);
done:
    {
	// execl("/sbin/ifconfig","ifconfig","ADIModem","10.10.10.11","netmask","255.255.255.0", NULL);
	
    }
    return ret;
}

